const getParamsUsed = (obj, dependencies) => {
    for (const property of Object.keys(obj)) {
        const val = obj[property];
        if (property === "parameter") {
            dependencies.push(val);
        }
        else if (typeof val === "object") {
            getParamsUsed(val, dependencies);
        }
    }
};
const constructDependencyGraph = (spec) => {
    var _a;
    // process spec to construct a dependency tree, enabling efficient re-computation of only entities that are affected by a change.
    const steps = [];
    // Process data section of spec
    for (const dataSpec of (spec.data || [])) {
        const dependencies = [];
        if ("source" in dataSpec) {
            dependencies.push(dataSpec.source);
        }
        if ("transform" in dataSpec) {
            for (const transform of dataSpec.transform) {
                if (transform.type === "lookup") {
                    dependencies.push(transform.from);
                }
            }
        }
        steps.push({
            name: dataSpec.name,
            type: "data",
            dependencies,
            spec: dataSpec
        });
    }
    // Process groupings
    for (const groupingSpec of (spec.groupings || [])) {
        steps.push({
            name: groupingSpec.name,
            type: "grouping",
            dependencies: [(_a = groupingSpec.data) !== null && _a !== void 0 ? _a : groupingSpec.network],
            spec: groupingSpec
        });
    }
    // Process networks
    for (const networkSpec of (spec.networks || [])) {
        let dependencies = [];
        if ("nodes" in networkSpec && "links" in networkSpec) {
            // a simple network construction
            dependencies = [networkSpec.nodes, networkSpec.links];
        }
        else if ("parts" in networkSpec) {
            // a complex network import
            for (const part of networkSpec.parts) {
                dependencies.push(part.data);
            }
        }
        else if ("source" in networkSpec) {
            dependencies.push(networkSpec.source);
        }
        steps.push({
            name: networkSpec.name,
            type: "network",
            dependencies,
            spec: networkSpec
        });
    }
    // Process orderings
    for (const orderingSpec of (spec.orderings || [])) {
        const dependencies = [];
        if ("data" in orderingSpec) {
            dependencies.push(orderingSpec.data);
        }
        if ("network" in orderingSpec) {
            dependencies.push(orderingSpec.network);
        }
        if (("method" in orderingSpec) && (typeof orderingSpec.method === "object") && ("parameter" in orderingSpec.method)) {
            dependencies.push(orderingSpec.method.parameter);
        }
        steps.push({
            name: orderingSpec.name,
            type: "ordering",
            dependencies,
            spec: orderingSpec
        });
    }
    // process maps
    for (const mapSpec of (spec.maps || [])) {
        steps.push({
            name: mapSpec.name,
            type: "map",
            dependencies: [],
            spec: mapSpec
        });
    }
    // process layouts
    for (const layoutSpec of (spec.layouts || [])) {
        const dependencies = [];
        if ("network" in layoutSpec) {
            dependencies.push(layoutSpec.network);
        }
        if ("data" in layoutSpec) {
            dependencies.push(layoutSpec.data);
        }
        if ("order" in layoutSpec) {
            dependencies.push(layoutSpec.order);
        }
        if ("scale" in layoutSpec) {
            dependencies.push(layoutSpec.scale);
        }
        if ("map" in layoutSpec) {
            dependencies.push(layoutSpec.map);
        }
        for (const field of ["x", "y", "radius", "angle"]) {
            if (field in layoutSpec) {
                if ("order" in layoutSpec[field]) {
                    dependencies.push(layoutSpec[field].order);
                }
                if ("field" in layoutSpec[field]) {
                    dependencies.push(layoutSpec[field].order);
                }
            }
        }
        getParamsUsed(layoutSpec, dependencies);
        steps.push({
            name: layoutSpec.name,
            type: "layout",
            dependencies,
            spec: layoutSpec
        });
    }
    // process scales
    for (const scaleSpec of (spec.scales || [])) {
        const dependencies = [];
        if ("data" in scaleSpec.domain) {
            dependencies.push(scaleSpec.domain.data.split(".")[0]);
        }
        steps.push({
            name: scaleSpec.name,
            type: "scale",
            dependencies: dependencies,
            spec: scaleSpec
        });
    }
    // process params
    for (const paramSpec of (spec.parameters || [])) {
        if ("source" in paramSpec) {
            const dependencies = [paramSpec.source];
            for (const step of paramSpec.transform) {
                if (step.type === "edgesBetween") {
                    dependencies.push(step.secondSelection);
                }
            }
            steps.push({
                name: paramSpec.name,
                type: "parameter",
                dependencies,
                spec: paramSpec
            });
        }
        else {
            steps.push({
                name: paramSpec.name,
                type: "parameter",
                dependencies: [],
                spec: paramSpec
            });
        }
    }
    // process tables
    for (const tableSpec of (spec.tables || [])) {
        const dependencies = [];
        if ("group" in tableSpec) {
            dependencies.push(tableSpec.group);
        }
        if ("order" in tableSpec.rowOrder) {
            dependencies.push(tableSpec.rowOrder.order);
        }
        if ("order" in tableSpec.colOrder) {
            dependencies.push(tableSpec.colOrder.order);
        }
        // TODO: other case of these if statements...
        steps.push({
            name: tableSpec.name,
            type: "table",
            dependencies,
            spec: tableSpec
        });
    }
    // process vis
    for (const visSpec of (spec.vis || [])) {
        const dependencies = getVisDependencies(visSpec);
        steps.push({
            name: visSpec.name,
            type: "vis",
            dependencies,
            spec: visSpec
        });
    }
    return steps;
};
// TODO: filter out objects (e.g. scales or layouts) that are "locally scoped" to within vis block
const getVisDependencies = (visSpec) => {
    let dependencies = [];
    // the data source
    // N.B. a map mark doesn't have one
    if ("entries" in visSpec) {
        dependencies.push(visSpec.entries);
    }
    if ("layout" in visSpec) {
        dependencies.push(visSpec.layout);
    }
    if ("ifInSelection" in visSpec) {
        dependencies.push(visSpec.ifInSelection);
    }
    for (const field of Object.keys(visSpec.mark || {})) {
        const val = visSpec.mark[field];
        if (typeof val === "object") {
            if ("scale" in val) {
                dependencies.push(val.scale);
            }
            if ("order" in val) {
                dependencies.push(val.order);
            }
            if ("condition" in val) {
                if ("inSelection" in val.condition) {
                    dependencies.push(val.condition.inSelection.split(".")[0]);
                }
            }
        }
    }
    // handle nesting in the vis block
    for (const subSpec of (visSpec.vis || [])) {
        dependencies = [...dependencies, ...getVisDependencies(subSpec)];
    }
    return dependencies;
};
const determineEvaluationOrder = (dependencyGraph, modifiedItems) => {
    // Return a list of affected items, in an order in which they can be re-evaluated
    // first identify everything that has been invalidated, by BFS
    let invalidated = dependencyGraph.filter(item => modifiedItems.includes(item.name));
    let children = [];
    do {
        const invalidatedNames = invalidated.map(item => item.name);
        children = dependencyGraph.filter(item => item.dependencies.some(depName => invalidatedNames.includes(depName)))
            .filter(item => !invalidated.includes(item));
        invalidated = [...invalidated, ...children];
    } while (children.length > 0);
    // then list in appropriate order
    let recompute = [];
    while (invalidated.length > 0) {
        const invalidatedNames = invalidated.map(item => item.name);
        const canRecompute = invalidated.filter(item => !item.dependencies.some(depName => invalidatedNames.includes(depName)));
        recompute = [...recompute, ...canRecompute];
        invalidated = invalidated.filter(item => !canRecompute.includes(item));
    }
    return recompute;
};
export { constructDependencyGraph, determineEvaluationOrder };
//# sourceMappingURL=dependencyGraph.js.map