var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { sceneFromJSON, CanvasRenderer, CanvasHandler, SVGHandler, SVGRenderer, SVGStringRenderer } from "vega-scenegraph";
import { getBoundsFromSpec, processFullSpec, updateSpec } from "./complete";
import { backgroundClickHandler, createSelectionOverlay, cursorEventHandler, mousedown, mousemove, mouseout, mouseover, mouseup } from "./viewer/interaction";
import { createInputs } from "./viewer/boundInputs";
import { createAnimationControls } from "./viewer/animationControls";
import Leaflet from "leaflet";
class GraphgraView {
    constructor(spec, options) {
        var _a, _b, _c, _d, _e, _f;
        this._origin = [0, 0];
        this._zoom = 1;
        this.leafletMaps = [];
        this.spec = spec;
        this.background = (_a = options.background) !== null && _a !== void 0 ? _a : "#ffffff";
        this.container = (_b = options.container) !== null && _b !== void 0 ? _b : "#view";
        this.logLevel = (_c = options.logLevel) !== null && _c !== void 0 ? _c : "none";
        this.renderer = (_d = options.renderer) !== null && _d !== void 0 ? _d : "svg"; // TODO: implement other renderers
        this.funcs = (_e = options.funcs) !== null && _e !== void 0 ? _e : {};
        this.paramCallbacks = (_f = options.paramCallbacks) !== null && _f !== void 0 ? _f : {};
        //this.tooltip = options.tooltip ??
    }
    render() {
        return __awaiter(this, void 0, void 0, function* () {
            // element is selector for the element on the page in which to add the embedded vie
            // spec is either a specification object, or the URL of a specification object
            let specObj;
            if (typeof (this.spec) === "string") {
                const specStringPromise = fetch(this.spec)
                    .then(response => response.json());
                specObj = yield specStringPromise;
            }
            else {
                specObj = this.spec;
            }
            // this computes the bounds of the region within which whatever is designed in the spec is rendered
            const bounds = getBoundsFromSpec(this.spec);
            const width = bounds.width + 2 * bounds.x;
            const height = bounds.height + 2 * bounds.y;
            ///
            const parentObj = document.querySelector(this.container);
            parentObj.innerHTML = "";
            const inputObj = document.createElement("div");
            parentObj.appendChild(inputObj);
            const elementObj = document.createElement("div");
            parentObj.appendChild(elementObj);
            if ("maps" in specObj && specObj.maps.length > 0) {
                const mapDiv = document.createElement("div");
                mapDiv.id = "mapDiv";
                mapDiv.style.position = "absolute";
                mapDiv.style.left = "0";
                mapDiv.style.height = "1000px";
                mapDiv.style.width = "1000px";
                mapDiv.style.zIndex = "0";
                elementObj.appendChild(mapDiv);
            }
            const vegaDiv = document.createElement("div");
            vegaDiv.style.position = "absolute";
            vegaDiv.style.left = "0";
            vegaDiv.style.zIndex = "2";
            elementObj.appendChild(vegaDiv);
            const tooltipDiv = document.createElement("div");
            tooltipDiv.style.position = "absolute";
            tooltipDiv.style.left = "0";
            tooltipDiv.style.zIndex = "3";
            tooltipDiv.style.opacity = "0";
            tooltipDiv.style.color = "black";
            tooltipDiv.style.backgroundColor = "white";
            tooltipDiv.style.fontFamily = "sans-serif";
            tooltipDiv.style.fontSize = "11px";
            tooltipDiv.style.padding = "8px";
            tooltipDiv.style.borderRadius = "3px";
            tooltipDiv.style.boxShadow = "2px 2px 4px rgba(0, 0, 0, 0.1)";
            tooltipDiv.style.border = "1px solid #d9d9d9";
            elementObj.appendChild(tooltipDiv);
            const updateZoom = (ev) => __awaiter(this, void 0, void 0, function* () {
                // The WheelEvent has a deltaMode attribute, which specifies unit for the delta values
                // https://www.w3.org/TR/uievents/#idl-wheelevent
                // It seems that Chrome provides values in pixels, whereas Firefox provides values in lines.
                const deltaMode = ev.deltaMode;
                const deltaY = ev.deltaY;
                let deltaPixels;
                if (deltaMode === 1) {
                    deltaPixels = deltaY * 16;
                }
                else if (deltaMode === 0) {
                    deltaPixels = deltaY;
                }
                else {
                    console.log(`Encountered wheel event with unexpected deltaMode (${deltaMode}))`);
                    deltaPixels = 0;
                }
                const currentZoom = this._zoom;
                this._zoom -= deltaPixels / 1000;
                // if there's a map, zoom by 0 or one levels at a time
                if (this.leafletMaps.length > 0) {
                    const log2 = x => Math.log(x) / Math.log(2);
                    this._zoom = Math.pow(2, Math.round(log2(this._zoom)));
                    // let zoomDelta = log2(this._zoom) - log2(currentZoom);
                    let zoomDelta = 0;
                    if (deltaPixels < -20) {
                        // zoom in
                        zoomDelta = 1;
                        this._zoom = 2 * currentZoom;
                        for (const map of this.leafletMaps) {
                            console.log(`zooming in by ${zoomDelta}`);
                            map.zoomIn(zoomDelta);
                        }
                        // we also need to pan the non-map view
                        // (zooming the map zooms in on the center, whereas changing the vega view scaling stretches it from the top-left corner towards the bottom-right)
                        // TODO: should we always do a pan on a normal zoom too?
                        this._origin = [
                            this._origin[0] - width / (4 * currentZoom),
                            this._origin[1] - height / (4 * currentZoom)
                        ];
                    }
                    else if (deltaPixels > 20) {
                        //zoom out
                        if (currentZoom === 1) {
                            // prevent zooming out beyond initial view
                            // cnvas can become smaller than map, and bad things happen
                            this._zoom = currentZoom;
                            return;
                        }
                        zoomDelta = -1;
                        this._zoom = currentZoom / 2;
                        for (const map of this.leafletMaps) {
                            console.log(`zooming in by ${zoomDelta}`);
                            map.zoomIn(zoomDelta);
                        }
                        // we also need to pan the non-map view
                        // (zooming the map zooms in on the center, whereas changing the vega view scaling stretches it from the top-left corner towards the bottom-right)
                        // TODO: should we always do a pan on a normal zoom too?
                        this._origin = [
                            this._origin[0] + width / (2 * currentZoom),
                            this._origin[1] + height / (2 * currentZoom)
                        ];
                    }
                }
                const res = yield updateSpec(this.spec, state, [], createMap, this.funcs);
                setDisplay(res.sceneJSON);
                // update size of the overlay SVG, so selection rectangles/lasso polygons remain visible
                const bbox = vegaDiv.getBoundingClientRect();
                selectionSVG.style.width = `${bbox.width}px`;
                selectionSVG.style.height = `${bbox.height}px`;
            });
            const setDisplay = (sceneJSON) => {
                var _a, _b;
                const scene = sceneFromJSON(sceneJSON);
                const r = renderToSVG(vegaDiv, scene, width, height, this.renderer, this._origin, this._zoom);
                if (this.renderer === "canvas") {
                    r.canvas();
                }
                else {
                    r.svg();
                }
                const Handler = (this.renderer === "canvas") ? CanvasHandler : SVGHandler;
                // This would be a cleaner solution than using mouseup/mousedown event handlers.
                // However, it doesn't currently work for tooltips on links/lines.
                //  `handleTooltip` in vega-scenegraph gets called with the `item` argument being the correct line, but then this is reset to null by the line:
                //  `item = resolveItem(item, event, this.canvas(), this._origin);`
                // This is because `resolveItem()` calls `pickLine()`, which checks the distance between the cursor's position and the line's endpoints, rather than to the closest point on the line.
                /*
                const tooltipFunc = (handler, ev, item, value) => {
          
                  // value replaces target.mark.items[0].tooltip ?
                    tooltipDiv.style.opacity = value ? "1" : "0";
          
                    tooltipDiv.style.left = `${ev.offsetX}px`;
                    tooltipDiv.style.top = `${ev.offsetY}px`;
                    tooltipDiv.innerHTML = value;
          
                  console.log("Tooltip func called:", { handler, event, item, value });
                }
          
                 */
                // override default tooltip function to prevent the browser displaying the `title` next to our custom tooltip div
                const tooltipFunc = (handler, ev, item, value) => null;
                const handler = new Handler(undefined, tooltipFunc)
                    .initialize(r.element())
                    .scene(scene);
                const interactionModifiers = {
                    "rect": [],
                    "lasso": [],
                    "dragElement": [],
                    "dragSelection": new Map(),
                    "pan": null,
                };
                if ("pan" in specObj) {
                    interactionModifiers.pan = specObj.pan;
                }
                for (const selection of (specObj.parameters || [])) {
                    if ("dragging" in selection) {
                        interactionModifiers.dragSelection.set(selection.name, selection.dragging);
                    }
                }
                let visToProcess = JSON.parse(JSON.stringify(specObj.vis));
                while (visToProcess.length > 0) {
                    const vis = visToProcess.pop();
                    visToProcess = [...visToProcess, ...(vis.vis || [])];
                    for (const action of (vis.actions || [])) {
                        if (action.interaction === "drag") {
                            interactionModifiers.dragElement.push({ modifierKeyRequired: action.modifierKeyRequired, modifierKeyAbsent: action.modifierKeyAbsent });
                        }
                        else if (action.interaction === "select" && action.using === "lasso") {
                            interactionModifiers.lasso.push({
                                modifierKeyRequired: action.modifierKeyRequired,
                                modifierKeyAbsent: action.modifierKeyAbsent
                            });
                        }
                        else if (action.interaction === "select" && action.using === "rect") {
                            interactionModifiers.rect.push({
                                modifierKeyRequired: action.modifierKeyRequired,
                                modifierKeyAbsent: action.modifierKeyAbsent
                            });
                        }
                    }
                }
                // Separately register event handler for each evnet type in CursorInteraction
                const offset = { x: (_a = specObj.x) !== null && _a !== void 0 ? _a : 25, y: (_b = specObj.y) !== null && _b !== void 0 ? _b : 25 };
                handler.on("click", (ev, target) => {
                    cursorEventHandler(ev, target, updateSelection2, interactionModifiers);
                    backgroundClickHandler(ev, target, specObj.parameters, updateSelection2);
                });
                handler.on("mouseenter", (ev, target) => cursorEventHandler(ev, target, updateSelection2, interactionModifiers));
                handler.on("mouseleave", (ev, target) => cursorEventHandler(ev, target, updateSelection2, interactionModifiers));
                handler.on("mouseover", (ev, target) => cursorEventHandler(ev, target, updateSelection2, interactionModifiers));
                handler.on("mouseout", (ev, target) => {
                    mouseout(ev, target, tooltipDiv);
                    cursorEventHandler(ev, target, updateSelection2, interactionModifiers);
                });
                handler.on("mousedown", (ev, target) => mousedown(ev, target, interactionState, sceneJSON, updateSelection2, interactionModifiers, this._zoom, this._origin, offset));
                handler.on("mousemove", (ev) => mousemove(ev, interactionState, interactionModifiers));
                handler.on("mouseup", (ev) => mouseup(ev, interactionState, sceneJSON, updateLayout, updateSelection2, updatePan, updateOrder, interactionModifiers, state, this._zoom, this._origin, offset));
                if (specObj.zoom) {
                    //handler.on("mousewheel", (ev) => updateZoom);
                    vegaDiv.onwheel = updateZoom;
                }
                // I think this is needed as we are not using the Vega view component (only vega-render)
                handler.on("mouseover", (ev, target) => mouseover(ev, target, tooltipDiv));
            };
            const createMap = (config, name, state, bounds) => __awaiter(this, void 0, void 0, function* () {
                const mapDiv = document.getElementById("mapDiv"); // TODO: what if multiple renderings, each with a map
                // avoid re-rendering map - a terrible hack
                // TODO: FIXME
                if (mapDiv.children.length > 0) {
                    return;
                }
                const thisMapDiv = document.createElement("div");
                thisMapDiv.style.position = "absolute";
                thisMapDiv.style.left = (bounds.x || 0) + "px";
                thisMapDiv.style.top = (bounds.y || 0) + "px";
                thisMapDiv.style.width = (bounds.width || 1000) + "px";
                thisMapDiv.style.height = (bounds.height || 1000) + "px";
                mapDiv.appendChild(thisMapDiv);
                const map = Leaflet.map(thisMapDiv).setView(config.center, config.zoomLevel);
                Leaflet.tileLayer(config.url, config.options)
                    .addTo(map);
                map.removeControl(map.zoomControl);
                mapDiv.style.zIndex = "0";
                state[name + "#leafletMap"] = map;
                this.leafletMaps.push(map);
            });
            const { sceneJSON, state } = yield processFullSpec(specObj, bounds, {}, createMap, this.funcs);
            // don't stomp on a sceneJOSN produced by createMap
            if (!this.sceneJSON) {
                this.sceneJSON = sceneJSON;
            }
            this.state = state;
            const paramNetwork = new Map();
            for (const param of (specObj.parameters || [])) {
                if ("network" in param) {
                    paramNetwork.set(param.name, param.network);
                }
            }
            const updatePan = (dragStartEv, dragEndEv) => __awaiter(this, void 0, void 0, function* () {
                const dx = dragEndEv.screenX - dragStartEv.screenX;
                const dy = dragEndEv.screenY - dragStartEv.screenY;
                this._origin = [
                    this._origin[0] + dx / (this._zoom * this._zoom),
                    this._origin[1] + dy / (this._zoom * this._zoom)
                ];
                for (const map of this.leafletMaps) {
                    map.panBy({ x: -dx / this._zoom, y: -dy / this._zoom });
                }
                const res = yield updateSpec(this.spec, state, [], createMap, this.funcs);
                setDisplay(res.sceneJSON);
            });
            const setParam = (paramName, newVal) => __awaiter(this, void 0, void 0, function* () {
                console.log(`UPDATED PARAM ${paramName} to ${newVal}`);
                state[paramName] = newVal;
                if (paramName in this.paramCallbacks) {
                    this.paramCallbacks[paramName](newVal);
                }
                const res = yield updateSpec(this.spec, state, [paramName], createMap, this.funcs);
                setDisplay(res.sceneJSON);
            });
            // function to externally set param values
            this.setParam = (paramName, newVal) => __awaiter(this, void 0, void 0, function* () {
                if (JSON.stringify(newVal) === JSON.stringify(state[paramName])) {
                    return;
                }
                setParam(paramName, newVal);
                if (paramName in this.paramInputElements) {
                    this.paramInputElements[paramName].value = newVal;
                }
            });
            const updateLayout = (layout_name, ids, shift) => __awaiter(this, void 0, void 0, function* () {
                console.log("UPDATING LAYOUT:", { layout_name, ids, shift });
                for (const id of ids) {
                    state[layout_name].get(id).x += shift.dx;
                    state[layout_name].get(id).y += shift.dy;
                }
                const res = yield updateSpec(this.spec, state, [layout_name], createMap, this.funcs);
                setDisplay(res.sceneJSON);
            });
            const editSelection = (selectionName, newSelectionIds, mode, networkName) => {
                const network = state[networkName];
                const newSelection = {
                    nodes: newSelectionIds.nodes.map(id => network.nodes.filter(n => n.id === id)[0]),
                    links: newSelectionIds.links.map(id => network.links.filter(l => l.id === id)[0]),
                    network: networkName
                };
                const selection = state[selectionName];
                if (mode === "replace") {
                    state[selectionName] = newSelection;
                }
                else if (mode === "remove") {
                    state[selectionName].nodes = selection.nodes.filter(n => !newSelection.nodes.includes[n]);
                    state[selectionName].links = selection.links.filter(n => !newSelection.links.includes[n]);
                }
                else if (mode === "add") {
                    state[selectionName].nodes = [...selection.nodes, ...newSelection.nodes];
                    state[selectionName].nodes = selection.nodes.filter((d, i) => (i === selection.nodes.indexOf(d)));
                    state[selectionName].links = [...selection.links, ...newSelection.links];
                    state[selectionName].links = selection.links.filter((d, i) => (i === selection.links.indexOf(d)));
                }
                else if (mode === "toggle") {
                    const newNodes = newSelection.nodes.filter(n => !selection.nodes.includes(n));
                    const remainingNodes = selection.nodes.filter(n => !newSelection.nodes.includes(n));
                    state[selectionName].nodes = [...remainingNodes, ...newNodes];
                    const newLinks = newSelection.links.filter(n => !selection.links.includes(n));
                    const remainingLinks = selection.links.filter(n => !newSelection.links.includes(n));
                    state[selectionName].links = [...remainingLinks, ...newLinks];
                }
                else {
                    // default is "replace"
                    state[selectionName].nodes = [...selection.nodes, ...newSelection.nodes];
                    state[selectionName].links = [...selection.links, ...newSelection.links];
                }
                if (selectionName in this.paramCallbacks) {
                    this.paramCallbacks[selectionName](state[selectionName]);
                }
            };
            const updateOrder = (order_name) => __awaiter(this, void 0, void 0, function* () {
                const res = yield updateSpec(this.spec, state, [order_name], createMap, this.funcs);
                setDisplay(res.sceneJSON);
            });
            const updateSelection2 = (newSelections, actions) => __awaiter(this, void 0, void 0, function* () {
                for (const [actionName, action] of actions) {
                    const newSelection = newSelections.get(actionName);
                    const mode = action.action || "replace";
                    //const networkName = action.network;
                    const networkName = paramNetwork.get(action.selection); // spec[action.selection].network;
                    editSelection(actionName, newSelection, mode, networkName);
                }
                const selectionNames = Array.from(newSelections.keys());
                const res = yield updateSpec(this.spec, state, selectionNames, createMap, this.funcs);
                setDisplay(res.sceneJSON);
            });
            this.paramInputElements = createInputs(state, inputObj, specObj.parameters || [], setParam);
            createAnimationControls(inputObj, specObj.parameters || [], setParam);
            const { selectionSVG, lasooPolygon, lasooPolygonOutline, lassoCircle, rectBrush } = createSelectionOverlay(elementObj, width, height);
            const interactionState = {
                dragStartEv: null,
                dragEndEv: null,
                draggedObj: null,
                draggedSelection: null,
                draggedTableHeading: null,
                panStartEv: null,
                draggedLayout: null,
                layoutName: null,
                lassoPoints: [],
                brush: { x1: null, x2: null, y1: null, y2: null },
                // DOM elements
                lasooPolygon,
                lasooPolygonOutline,
                lassoCircle,
                rectBrush
            };
            setDisplay(this.sceneJSON);
        });
    }
    renderHeadless(Renderer) {
        return __awaiter(this, void 0, void 0, function* () {
            // spec is either a specification object, or the URL of a specification object
            let specObj;
            if (typeof (this.spec) === "string") {
                const specStringPromise = fetch(this.spec)
                    .then(response => response.json());
                specObj = yield specStringPromise;
            }
            else {
                specObj = this.spec;
            }
            // this computes the bounds of the region within which whatever is designed in the spec is rendered
            const bounds = getBoundsFromSpec(this.spec);
            const width = bounds.width + 2 * bounds.x;
            const height = bounds.height + 2 * bounds.y;
            ///
            const { sceneJSON, state } = yield processFullSpec(specObj, bounds, {}, this.funcs); // FIXME: map creation won't work
            this.sceneJSON = sceneJSON;
            this.state = state;
            ////
            const scene = sceneFromJSON(sceneJSON);
            return new Renderer()
                .initialize(null, width, height)
                .renderAsync(scene);
        });
    }
    toSVG() {
        return __awaiter(this, void 0, void 0, function* () {
            const r = yield this.renderHeadless(SVGStringRenderer);
            return r.svg();
        });
    }
    toCanvas() {
        return __awaiter(this, void 0, void 0, function* () {
            const r = yield this.renderHeadless(CanvasRenderer);
            return r.canvas();
        });
    }
}
function renderToSVG(el, scene, width, height, renderer, origin, zoom) {
    // clear
    el.innerHTML = "";
    const Renderer = (renderer === "canvas") ? CanvasRenderer : SVGRenderer;
    return new Renderer()
        .initialize(el, width, height, origin, zoom)
        .render(scene);
}
export { GraphgraView };
//# sourceMappingURL=embedNetwork.js.map