import { repositionLineSegments } from "respacer";
import { select } from "../data/select";
import { codegenExpression, parseExpression } from "vega-expression";
import { forceCollide, forceSimulation, forceX, forceY } from "d3-force";
import { mean } from "d3-array";
const setField = (layout, spec) => {
    const transformedLayout = new Map();
    for (const e of layout.keys()) {
        // get datum by filtering? then pos to allow value?
        transformedLayout.set(e, Object.assign(Object.assign({}, layout.get(e)), { [spec.field]: spec.value }));
    }
    return transformedLayout;
};
const offset = (layout, data, spec) => {
    var _a, _b;
    const direction = spec.direction === "horizontal" ? "x" : "y";
    const width = (_a = spec.width) !== null && _a !== void 0 ? _a : 1000;
    const points = [];
    if ("condition" in spec) {
        const codegen = codegenExpression({
            globalvar: "globalVar"
        });
        const funcCode = codegen(parseExpression(spec.condition));
        // First argument means that the function argument will be interpreted as having name "globalVar"
        // so we can avoid using an actual global variable
        const fn = Function("globalVar", "\"use strict\"; return (" + funcCode.code + ")");
        for (const datum of data) {
            if (fn({ datum })) {
                points.push(Object.assign({ id: datum.id }, layout.get(datum.id)));
            }
        }
    }
    else {
        for (const id of layout.keys()) {
            points.push(Object.assign({ id }, layout.get(id)));
        }
    }
    // TODO: should pass in bounds as "width" option...
    const repositioned = repositionLineSegments(points, {
        oldPositionName: direction,
        widthName: (direction === "x") ? "width" : "height",
        newPositionName: "newPos",
        minSpacing: (_b = spec.minSpacing) !== null && _b !== void 0 ? _b : 0,
        width
    });
    const transformedLayout = new Map();
    for (const e of layout.keys()) {
        transformedLayout.set(e, layout.get(e));
    }
    for (const pos of repositioned) {
        const p = layout.get(pos.id);
        transformedLayout.set(pos.id, Object.assign(Object.assign({}, p), { [direction]: pos.newPos })); // , _shifted: p[direction] !== pos.newPos
    }
    return transformedLayout;
};
const orientLinks = (layout, spec, state) => {
    const network = state[spec.network];
    const field = spec.direction === "vertical" ? "x" : "y";
    const transformedLayout = new Map();
    for (const e of layout.keys()) {
        transformedLayout.set(e, layout.get(e));
    }
    for (const link of network.links) {
        transformedLayout.get(link.target.id)[field] = layout.get(link.source.id)[field];
    }
    return transformedLayout;
};
const jitter = (layout, spec) => {
    const transformedLayout = new Map();
    for (const e of layout.keys()) {
        const pos = layout.get(e);
        let x = pos.x;
        if ("x" in spec) {
            x += spec.x * 2 * (Math.random() - 0.5);
        }
        let y = pos.y;
        if ("y" in spec) {
            y += spec.y * 2 * (Math.random() - 0.5);
        }
        transformedLayout.set(e, Object.assign(Object.assign({}, pos), { x, y }));
    }
    return transformedLayout;
};
const beeswarm = (layout, spec) => {
    var _a, _b, _c;
    // const nodes = [...layout.values()];
    const nodes = [];
    for (const id of layout.keys()) {
        nodes.push(Object.assign({ id }, layout.get(id)));
    }
    const simulation = forceSimulation(nodes);
    simulation.force("x", forceX().strength((_a = spec.gravityX) !== null && _a !== void 0 ? _a : 0.1).x(d => d.x));
    simulation.force("y", forceY().strength((_b = spec.gravityY) !== null && _b !== void 0 ? _b : 0.1).y(d => d.y));
    simulation.force("collision", forceCollide().radius((d) => Math.max(d.width, d.height) / 2).strength((_c = spec.collisionStrength) !== null && _c !== void 0 ? _c : 1));
    // TODO: reconsider termination condition
    let meanV = 1e9;
    do {
        simulation.tick();
        meanV = mean(nodes.map((n) => Math.sqrt(Math.pow(n.vx, 2) + Math.pow(n.vy, 2)))) || 0;
        // console.log(mean_v);
    } while (meanV > 1e-3);
    simulation.stop();
    const transformedLayout = new Map();
    for (const node of nodes) {
        transformedLayout.set(node.id, Object.assign(Object.assign({}, layout.get(node.id)), { x: node.x, y: node.y }));
    }
    return transformedLayout;
};
const transformLayout = (transformSpec, layoutSpec, layout, state, funcs) => {
    // this is repetitive with computeAttributeLayout()
    let data;
    if ("network" in layoutSpec) {
        data = select(state, layoutSpec.network).nodes;
    }
    else {
        data = select(state, layoutSpec.data);
    }
    if (transformSpec.type === "set") {
        return setField(layout, transformSpec);
    }
    else if (transformSpec.type === "offset") {
        return offset(layout, data, transformSpec);
    }
    else if (transformSpec.type === "orientLinks") {
        return orientLinks(layout, transformSpec, state);
    }
    else if (transformSpec.type === "jitter") {
        return jitter(layout, transformSpec);
    }
    else if (transformSpec.type === "beeswarm") {
        return beeswarm(layout, transformSpec);
    }
    else if (transformSpec.type === "function") {
        return funcs[transformSpec.function](layout, transformSpec);
    }
};
export { transformLayout };
//# sourceMappingURL=layoutTransforms.js.map