import { select } from "../data/select";
import { constructExpressionFunction } from "../data/calculate";
import { max, min } from "d3-array";
const getId = (val) => {
    // We use this function as orders are usually defined over objects with ids, but we also need
    // orderings over numbers/strings (if we are constructing a table by grouping data rows, in which the entries do not have ids
    // [will this occur - nodes and links have ids?]
    if (typeof val === "object") {
        return val.id;
    }
    return val;
};
const applyOrderOrScale = (datum, spec, state) => {
    if ("value" in spec) {
        return spec.value;
    }
    else if ("order" in spec && typeof (spec.order) === "string") {
        return state[spec.order].get(getId(datum));
    }
    else if ("order" in spec) {
        // this is needed - e.g., for links where we may want to order based on id of source node (rather than id of link object)
        const objToOrder = ("field" in spec.order && spec.order.field) ? select(datum, spec.order.field) : datum;
        return state[spec.order.order].get(getId(objToOrder));
    }
    else {
        const scale = state[spec.scale];
        return scale(select(datum, spec.field));
    }
};
const getTicks = (spec, state) => {
    if ("order" in spec || "value" in spec) {
        return [];
    }
    const scale = state[spec.scale];
    if (typeof (scale.ticks) === "undefined") {
        return scale.domain().map(t => [scale(t), t]);
    }
    return scale.ticks().map(t => [scale(t), t]);
};
const getScale = (spec, state) => {
    if ("order" in spec) {
        return [];
    }
    return state[spec.scale];
};
// TODO: move
const getSize = (datum, spec, state, currentVal) => {
    // get the width or height of an element
    if (!spec) {
        return currentVal !== null && currentVal !== void 0 ? currentVal : 10; // TODO: is this a reasonable default?
    }
    else if (typeof spec === "number") {
        return spec;
    }
    else if ("scale" in spec) {
        const scale = state[spec.scale];
        return scale(datum[spec.field]);
    }
    else if ("field" in spec) {
        return select(datum, spec.field);
    }
    else if ("expression" in spec) {
        const f = constructExpressionFunction({ calculate: spec.expression });
        return f(datum, null, state);
    }
};
const scaleIndependently = (layout, varToScale, isOrder, positioning, domain) => {
    // This is a ugly hack to deal with the fact that we have to rescale both maps (layouts) and arrays (lists of axis ticks)
    let isTicks = false;
    const inLayout = layout;
    const l2 = new Map();
    if (Array.isArray(layout)) {
        isTicks = true;
        for (const i in layout) {
            l2.set(i, layout[i]);
        }
        layout = l2;
    }
    let maxVal = 0;
    let minVal = Infinity;
    let allIntegers = true;
    for (const i of layout.keys()) {
        const val = layout.get(i)[varToScale];
        maxVal = Math.max(maxVal, val);
        minVal = Math.min(minVal, val);
        allIntegers = allIntegers && Number.isInteger(val);
    }
    for (const i of layout.keys()) {
        // Rescale to fit in domain
        if (!allIntegers) {
            layout.get(i)[varToScale] = domain[0] + (domain[1] - domain[0]) * (layout.get(i)[varToScale] - minVal) / (maxVal - minVal);
        }
        else {
            // we don't want the first and last values to overlap
            // e.g., in we have three axes [minVal=0 maxVal=2, maxVal-minVal=2], they should be at 0 degrees, 120 degrees, 240 degrees
            layout.get(i)[varToScale] = domain[0] + (domain[1] - domain[0]) * (layout.get(i)[varToScale] - minVal) / (1 + maxVal - minVal);
        }
    }
    if (isTicks) {
        for (const [key, val] of layout) {
            inLayout[key] = val;
        }
    }
};
const scale = (layout, varToScale, otherVar, isOrder, positioning, domain) => {
    // group elements by value of the otherVar
    const assignedAxis = new Map();
    let maxVal = 0;
    for (const i of layout.keys()) {
        const element = layout.get(i);
        const otherVarVal = element[otherVar];
        if (!assignedAxis.has(otherVarVal)) {
            assignedAxis.set(otherVarVal, []);
        }
        assignedAxis.get(otherVarVal).push(element);
        if (element[varToScale] > maxVal) {
            maxVal = element[varToScale];
        }
    }
    // suppose we have split nodes across rings by node type, and assigned the angle based on an order by node id
    // within each ring, there are gaps in the ordering - we need to remove these before redistributing, unless mode is none
    // At the same time, we find the maximum value in each group
    const maxValInGroups = new Map();
    if (isOrder && positioning !== "none") {
        maxVal = 0;
        for (const v of assignedAxis.keys()) {
            const elements = assignedAxis.get(v);
            const vals = elements.map(e => e[varToScale]);
            vals.sort((a, b) => a - b);
            const squishedElements = [];
            for (const i of layout.keys()) {
                const l = layout.get(i);
                squishedElements[i] = Object.assign(Object.assign({}, l), { [varToScale]: vals.indexOf(l[varToScale]) });
            }
            //const squishedElements = layout.map(l => ({ ...l, [varToScale]: vals.indexOf(l[varToScale]) }));
            assignedAxis.set(v, squishedElements);
            for (const element of elements) {
                element[varToScale] = vals.indexOf(element[varToScale]);
            }
            maxValInGroups.set(v, vals.length);
            if (vals.length > maxVal) {
                maxVal = vals.length;
            }
        }
    }
    else {
        for (const v of assignedAxis.keys()) {
            const elements = assignedAxis.get(v);
            maxValInGroups.set(v, +max(elements.map(e => e[varToScale])));
        }
    }
    for (const i of layout.keys()) {
        const element = layout.get(i);
        const maxValInGroup = maxValInGroups.get(element[otherVar]);
        // adjust value of varToScale to meet specified position
        let shiftedN;
        if (positioning === "start" || positioning === "none" || !positioning) {
            shiftedN = element[varToScale];
        }
        else if (positioning === "end") {
            shiftedN = element[varToScale] + (maxVal - maxValInGroup);
        }
        else if (positioning === "middle") {
            shiftedN = element[varToScale] + (maxVal - maxValInGroup) / 2;
        }
        else if (positioning === "distribute") {
            shiftedN = element[varToScale] * (maxVal / maxValInGroup);
        }
        // Rescale to fit in domain
        // This is only needed when wrapping round a circle
        // otherwise we're scaling all axes by the same amount, which is pointless as they'll get rescaled together anyway
        if (domain) {
            element[varToScale] = domain[0] + domain[1] * (shiftedN / maxVal);
        }
        else {
            element[varToScale] = shiftedN;
        }
    }
};
const position = (n, maxElementsPerGroup, numElementsInGroup, domain, positioning, scaleOrOrder, state) => {
    if ("scale" in scaleOrOrder) {
        const scaleDomain = state[scaleOrOrder.scale].domain();
        return domain[0] + (domain[1] - domain[0]) * (n - scaleDomain[0]) / (scaleDomain[scaleDomain.length - 1] - scaleDomain[0]);
    }
    else {
        let shiftedN;
        if (positioning === "start" || !positioning) {
            shiftedN = n;
        }
        else if (positioning === "end") {
            shiftedN = n + (maxElementsPerGroup - numElementsInGroup);
        }
        else if (positioning === "middle") {
            shiftedN = n + (maxElementsPerGroup - numElementsInGroup) / 2;
        }
        else if (positioning === "distribute") {
            shiftedN = n * (maxElementsPerGroup / numElementsInGroup);
        }
        return domain[0] + domain[1] * (shiftedN / maxElementsPerGroup);
    }
};
const rescaleLayout = (spec, state, laidOutData1, layout, bounds, axes, useEqualScaling) => {
    // Rescales so that positions are within intervals [0, bounds.width] and [0, bounds.height]
    // don't translate by adding bounds.x or bounds.y - position is relative
    // TODO: this scales elements so their left edges fit, but probably better to scale so that their outer edges fit?
    const laidOutData = [...laidOutData1];
    if (layout.has("#TOP_LEFT_CORNER")) {
        laidOutData.push({ id: "#TOP_LEFT_CORNER" });
    }
    if (layout.has("#BOTTOM_RIGHT_CORNER")) {
        laidOutData.push({ id: "#BOTTOM_RIGHT_CORNER" });
    }
    // scale width/height first, so we can scale things so the bottom/right edge fits
    for (const n of laidOutData) {
        const pos = layout.get(n.id);
        const widthSpec = "width" in spec ? spec.width : undefined;
        pos.width = getSize(n, widthSpec, state, pos.width);
        const heightSpec = "height" in spec ? spec.height : undefined;
        pos.height = getSize(n, heightSpec, state, pos.height);
    }
    // work out scaling factors for positions
    const maxX = +max(laidOutData.map((n) => layout.get(n.id).x + layout.get(n.id).width)) || 0;
    const minX = +min(laidOutData.map((n) => layout.get(n.id).x)) || 0;
    const maxY = +max(laidOutData.map((n) => layout.get(n.id).y + layout.get(n.id).height));
    const minY = +min(laidOutData.map((n) => layout.get(n.id).y));
    let xScaling = (maxX === minX) ? 1 : bounds.width / (maxX - minX);
    let yScaling = (maxY === minY) ? 1 : bounds.height / (maxY - minY);
    //  true for network layouts to preserve aspect ratio
    // also for polar coordinates to avoid messing up scales
    if (useEqualScaling) {
        const scaling = Math.min(xScaling, yScaling);
        xScaling = scaling;
        yScaling = scaling;
    }
    console.log({ minY, maxY, xScaling, yScaling });
    // scale x/y positions of points
    for (const n of laidOutData) {
        const pos = layout.get(n.id);
        if (maxX === minX) {
            // avoid division-by-zero
            layout.get(n.id).x = Math.min(Math.max(maxX, 0), bounds.width);
        }
        else {
            layout.get(n.id).x = xScaling * (pos.x - minX);
        }
        if (maxY === minY) {
            layout.get(n.id).y = Math.min(Math.max(maxY, 0), bounds.width);
        }
        else {
            layout.get(n.id).y = yScaling * (pos.y - minY);
        }
        layout.get(n.id).width *= xScaling;
        layout.get(n.id).height *= yScaling;
    }
    // scale axes
    for (const axis of (axes || [])) {
        console.log({ minY, maxY, xScaling, yScaling });
        if (axis.type === "cartesian") {
            // scale ticks
            for (const tick of axis.ticks) {
                //  const {x, y} = scalePoint(tick.x, tick.y, minX, maxX, minY, maxY, scaling, bounds);
                //  tick.x = x;
                //  tick.y = y;
                if (maxX === minX) {
                    // avoid division-by-zero
                    tick.x = Math.min(Math.max(maxX, 0), bounds.width);
                }
                else {
                    tick.x = xScaling * (tick.x - minX);
                }
                if (maxY === minY) {
                    tick.y = Math.min(Math.max(maxY, 0), bounds.width);
                }
                else {
                    tick.y = yScaling * (tick.y - minY);
                }
            }
            // Transform ends of axis to appropriate positions (in coordinate system of layout pre-scaling)
            if (axis.extend === "x") {
                axis.x = minX;
                axis.x2 = maxX;
            }
            else if (axis.extend === "y") {
                axis.y = minY;
                axis.y = maxY;
            }
            else if (axis.extend === "radial") {
                const maxR = Math.max(maxX, maxY);
                const r = Math.sqrt(Math.pow(axis.x2, 2) + Math.pow(axis.y2, 2));
                axis.x = 0;
                axis.y = 0;
                axis.x2 = axis.x2 * (maxR / r);
                axis.y2 = axis.y2 * (maxR / r);
            }
            // Scale the ends of the axes in the same way as the data points
            const { x, y } = scalePoint(axis.x, axis.y, minX, maxX, minY, maxY, xScaling, yScaling, bounds);
            axis.x = x;
            axis.y = y;
            const { x: x2, y: y2 } = scalePoint(axis.x2, axis.y2, minX, maxX, minY, maxY, xScaling, yScaling, bounds);
            axis.x2 = x2;
            axis.y2 = y2;
        }
        else if (axis.type === "polar") {
            axis.r = axis.r * xScaling;
            axis.x = xScaling * (0 - minX);
            axis.y = xScaling * (0 - minY);
            // TODO: scale locations of ticks
            //  axis.x = scaling * (maxX - minX) / 2;
            //  axis.y = scaling * (maxY - minY) / 2;
            // the width is not necessarily (bounds.width, bounds.height), nor scaling * (maxX - minX) / 2;
            // the most extreme values are not necessarily symmetric about the centre
        }
    }
};
const scalePoint = (x, y, minX, maxX, minY, maxY, xScaling, yScaling, bounds) => {
    if (maxX === minX) {
        // avoid division-by-zero
        x = Math.min(Math.max(maxX, 0), bounds.width);
    }
    else {
        x = xScaling * (x - minX);
    }
    if (maxY === minY) {
        y = Math.min(Math.max(maxY, 0), bounds.width);
    }
    else {
        y = yScaling * (y - minY);
    }
    return { x, y };
};
export { rescaleLayout };
export { scale };
export { scaleIndependently };
export { getSize };
export { getTicks };
export { applyOrderOrScale };
//# sourceMappingURL=utils.js.map