import { select } from "../../data/select";
// Specify as {field: "fieldName"} or {value: "valueName"}
// allow just "fieldName" or "valueName" if unambiguous
const getValue = (dataObj, selector) => {
    if (selector === "*") {
        // TODO: we probably want to exclude already-used fields - e.g. the sorce and target for an edge
        return Object.assign({}, dataObj);
    }
    else if (typeof (selector) === "string") {
        return { [selector]: dataObj[selector] };
    }
    else if ("field" in selector) {
        const outputFieldName = selector.as ? selector.as : selector.field;
        return { [outputFieldName]: select(dataObj, selector.field) };
    }
    else if ("value" in selector) {
        const outputFieldName = selector.as ? selector.as : selector.value;
        return { [outputFieldName]: selector.value };
    }
};
const getNodeFromRow = (dataRow, nodeSpec, dataSource) => {
    const nodeData = {};
    for (const datumSelector of nodeSpec.data) {
        Object.assign(nodeData, getValue(dataRow, datumSelector));
    }
    return {
        id: select(dataRow, nodeSpec.id_field || "id"),
        _type: nodeSpec.type || "default",
        _dataSource: dataSource,
        data: nodeData
    };
};
const getLinkFromRow = (dataRow, linkSpec, dataSource) => {
    // TODO: get requested fields, and smoosh resulting object together
    const linkData = {};
    for (const datumSelector of (linkSpec.data || [])) {
        Object.assign(linkData, getValue(dataRow, datumSelector));
    }
    const obj2val = (obj) => Object.values(obj)[0];
    return {
        _dataSource: dataSource,
        _source_id: obj2val(getValue(dataRow, linkSpec.source_id)),
        _source_id_field: (linkSpec.source_id_field || "id"),
        _source_node_type: (linkSpec.source_node_type || "default"),
        _target_id: obj2val(getValue(dataRow, linkSpec.target_id)),
        _target_id_field: (linkSpec.target_id_field || "id"),
        _target_node_type: (linkSpec.target_node_type || "default"),
        data: linkData
    };
};
const addNodeToIndex = (index, type, field, val, node) => {
    if (!index[type]) {
        index[type] = {};
    }
    if (!index[type][field]) {
        index[type][field] = {};
    }
    index[type][field][val] = node;
    //  console.log("IN FUNCTION:", JSON.stringify(index['senator'], null, 4))
    //  console.log("IN FUNCTION:", JSON.stringify(index, null, 4))
    return index;
};
const constructNetworkComplex = (allData, spec) => {
    // THings to be aware of
    // connect nodes and links correctly, considering different ids for different node types
    // which field is the ID may depend on what is being connected: need to record  (source_id, source_node_type, source_id_field)
    // need to handle case of implicit node creation followed by explicit creation: handle this by creating nodes as explicitly defined, then joining up links later
    var _a, _b, _c, _d;
    const nodes = [];
    const links = [];
    let nodeIndex = [];
    // first, figure out how will need to index nodes to re-connect them
    const fieldsToIndex = {};
    for (const tableSpec of spec.parts) {
        for (const linkSpec of ([...(tableSpec.yieldsLinks || []), ...(tableSpec.mapYieldsLinks || [])])) {
            linkSpec.source_node_type = linkSpec.source_node_type || "default";
            linkSpec.target_node_type = linkSpec.target_node_type || "default";
            linkSpec.source_id_field = linkSpec.source_id_field || "id";
            linkSpec.target_id_field = linkSpec.target_id_field || "id";
            if (!fieldsToIndex[linkSpec.source_node_type]) {
                fieldsToIndex[linkSpec.source_node_type] = [];
            }
            fieldsToIndex[linkSpec.source_node_type].push(linkSpec.source_id_field);
            if (!fieldsToIndex[linkSpec.target_node_type]) {
                fieldsToIndex[linkSpec.target_node_type] = [];
            }
            fieldsToIndex[linkSpec.target_node_type].push(linkSpec.target_id_field);
        }
    }
    // record the explicitly specified nodes and links
    const nodeIds = [];
    const nodeIndexFromId = {};
    for (const tableSpec of spec.parts) {
        const data = select(allData, tableSpec.data);
        for (const dataRow of data) {
            for (const nodeSpec of (tableSpec.yieldsNodes || [])) {
                let node = getNodeFromRow(dataRow, nodeSpec, tableSpec.data);
                if (nodeIds.includes(node.id)) {
                    // ensure the link points to existing node object (rather than newly created one, which is identical but distinct and will not be saved)
                    node = nodes[nodeIndexFromId[node.id]];
                }
                else {
                    nodes.push(node);
                    nodeIds.push(node.id);
                    nodeIndexFromId[node.id] = nodes.length - 1;
                }
                // add index field for all fields that this node type is linked against
                // nodeIndex[type][field][value] = node
                for (const field of fieldsToIndex[node._type] || []) {
                    const val = (field === "id") ? node.id : dataRow[field]; // allow linking based on id of generated node, rather than field on input table
                    nodeIndex = addNodeToIndex(nodeIndex, node._type, field, val, node);
                }
            }
            for (const linkSpec of (tableSpec.yieldsLinks || [])) {
                const link = getLinkFromRow(dataRow, linkSpec, tableSpec.data);
                links.push(link);
                if (linkSpec.addReverseLinks) {
                    links.push(Object.assign(Object.assign({}, link), { _source_id: link._target_id, _source_id_field: link._target_id_field, _source_node_type: link._target_node_type, _target_id: link._source_id, _target_id_field: link._source_id_field, _target_node_type: link._source_node_type }));
                }
            }
            for (const mapLinkSpec of (tableSpec.mapYieldsLinks || [])) {
                const mappedData = select(dataRow, mapLinkSpec.field);
                const dataSource = tableSpec.data + "_" + mapLinkSpec.field;
                for (const d of mappedData) {
                    const link = getLinkFromRow({ parent: dataRow, child: d }, mapLinkSpec, dataSource);
                    links.push(link);
                    if (mapLinkSpec.addReverseLinks) {
                        links.push(Object.assign(Object.assign({}, link), { _source_id: link._target_id, _source_id_field: link._target_id_field, _source_node_type: link._target_node_type, _target_id: link._source_id, _target_id_field: link._source_id_field, _target_node_type: link._source_node_type }));
                    }
                }
            }
        }
    }
    // link everything up, and create any implicitly defined nodes
    for (const link of links) {
        let sourceNode = (_b = (_a = nodeIndex === null || nodeIndex === void 0 ? void 0 : nodeIndex[link._source_node_type]) === null || _a === void 0 ? void 0 : _a[link._source_id_field]) === null || _b === void 0 ? void 0 : _b[link._source_id];
        // if source node doesn't exist, it was defined implicitly so create it (and add to index to prevent duplication)
        if (!sourceNode) {
            //      console.log("FAILED TO FIND:", link._source_node_type, link._source_id_field, link._source_id);
            sourceNode = { [link._source_id_field]: link._source_id, _type: link._source_node_type };
            nodes.push(sourceNode);
            nodeIndex = addNodeToIndex(nodeIndex, link._source_node_type, link._source_id_field, link._source_id, sourceNode);
        }
        // if target node doesn't exist, it was defined implicitly so create it (and add to index to prevent duplication)
        let targetNode = (_d = (_c = nodeIndex === null || nodeIndex === void 0 ? void 0 : nodeIndex[link._target_node_type]) === null || _c === void 0 ? void 0 : _c[link._target_id_field]) === null || _d === void 0 ? void 0 : _d[link._target_id];
        if (!targetNode) {
            targetNode = { [link._target_id_field]: link._target_id, _type: link._target_node_type };
            nodes.push(targetNode);
            nodeIndex = addNodeToIndex(nodeIndex, link._target_node_type, link._target_id_field, link._target_id, targetNode);
        }
        link.source = sourceNode;
        link.target = targetNode;
        // strip out the _ fields that this function used internally
        delete link._source_id;
        delete link._source_node_type;
        delete link._source_id_field;
        delete link._target_id;
        delete link._target_node_type;
        delete link._target_id_field;
    }
    return { nodes, links };
};
export { constructNetworkComplex, getValue };
//# sourceMappingURL=constructNetworkComplex.js.map