import { constructAdjacencyList } from "../selectionTransforms/constructAdjacencyList";
import { select } from "../data/select";
const layoutBioFabric = (nodes, links, spec, state, addToState) => {
    // we assume that we're working on the original network ~~network is already split, so (node,link) pairs become (nodes)~~
    var _a, _b, _c;
    const adjacency = constructAdjacencyList(nodes, links, true);
    // find *distinct neighbours* node
    const neighbours = new Map();
    for (const link of links) {
        if (!neighbours.has(link.source.id)) {
            neighbours.set(link.source.id, new Set());
        }
        neighbours.get(link.source.id).add(link.target.id);
        if (!neighbours.has(link.target.id)) {
            neighbours.set(link.target.id, new Set());
        }
        neighbours.get(link.target.id).add(link.source.id);
    }
    // count these to get degree ignoring duplicate edges between same pair of nodes
    const degrees = new Map();
    for (const n of neighbours.keys()) {
        degrees.set(n, neighbours.get(n).size);
    }
    // sorted in ascending order
    const idsSortedByDegree = Array.from(degrees.keys())
        .sort((a, b) => degrees.get(a) - degrees.get(b));
    let nextAvailableRow = 1;
    const rowFromId = new Map();
    const idFromRow = new Map();
    let nodesUnassignedToRows = [...idsSortedByDegree];
    while (nodesUnassignedToRows.length > 0) {
        // assign node with highest degree to next available row
        const nodeId = nodesUnassignedToRows.pop();
        rowFromId.set(nodeId, nextAvailableRow);
        idFromRow.set(nextAvailableRow, nodeId);
        nextAvailableRow += 1;
        // then assign any of its neighbours that haven't already been assigned
        const links = adjacency.undirectedLinks.get(nodeId);
        // sort in descending degree order
        const neighbours = links.map(l => l.source.id === nodeId ? l.target : l.source)
            .sort((a, b) => degrees.get(b.id) - degrees.get(a.id));
        for (const neighbour of neighbours) {
            if (nodesUnassignedToRows.includes(neighbour.id)) {
                rowFromId.set(neighbour.id, nextAvailableRow);
                idFromRow.set(nextAvailableRow, neighbour.id);
                nextAvailableRow += 1;
                nodesUnassignedToRows = nodesUnassignedToRows.filter(n => n !== neighbour.id);
            }
        }
    }
    let nextAvailableColumn = 1;
    const colFromId = new Map();
    const idFromCol = new Map();
    for (let currentRow = 1; currentRow < idsSortedByDegree.length; currentRow += 1) {
        const nodeId = idFromRow.get(currentRow);
        const links = adjacency.undirectedLinks.get(nodeId);
        // group links incident to this node, based on node they connect to
        const groupedLinks = new Map();
        for (const l of links) {
            const neighbourId = l.source.id === nodeId ? l.target.id : l.source.id;
            const neighbourRow = rowFromId.get(neighbourId);
            if (neighbourRow < currentRow) {
                continue;
            }
            if (groupedLinks.has(neighbourId)) {
                groupedLinks.get(neighbourId).push(l);
            }
            else {
                groupedLinks.set(neighbourId, [l]);
            }
        }
        // "Order these sets by increasing row number r,
        // so that edges will be assigned in order of increasing
        // length."
        const orderedNeighbourIds = Array.from(groupedLinks.keys()).sort((a, b) => rowFromId.get(a) - rowFromId.get(b));
        for (const neighbourId of orderedNeighbourIds) {
            const sortField = (_a = spec.linkTypeColName) !== null && _a !== void 0 ? _a : "id";
            const linksToDraw = groupedLinks.get(neighbourId).sort((a, b) => compare(a, b, sortField));
            for (const link of linksToDraw) {
                colFromId.set(link.id, nextAvailableColumn);
                idFromCol.set(nextAvailableColumn, link.id);
                nextAvailableColumn += 1;
            }
        }
    }
    const newNodes = [];
    const newLinks = [];
    let layout = new Map();
    let nodeId = 0;
    const oldIdField = (_b = spec.saveIdAs) !== null && _b !== void 0 ? _b : "_originalId";
    for (const l of links) {
        const sourceNode = Object.assign(Object.assign({}, l.source), { id: nodeId++, [oldIdField]: l.source.id });
        const targetNode = Object.assign(Object.assign({}, l.target), { id: nodeId++, [oldIdField]: l.target.id });
        newNodes.push(sourceNode);
        newNodes.push(targetNode);
        newLinks.push(Object.assign(Object.assign({}, l), { source: sourceNode, target: targetNode }));
        layout.set(sourceNode.id, {
            x: colFromId.get(l.id),
            y: rowFromId.get(l.source.id),
            width: 1,
            height: 1
        });
        layout.set(targetNode.id, {
            x: colFromId.get(l.id),
            y: rowFromId.get(l.target.id),
            width: 1,
            height: 1
        });
    }
    const newNetworkName = (_c = spec.newNetworkName) !== null && _c !== void 0 ? _c : `${spec.network}#BioFabric`;
    //state[newNetworkName] = { nodes: newNodes, links: newLinks };
    const reshapedNetwork = { nodes: newNodes, links: newLinks };
    addToState(newNetworkName, reshapedNetwork);
    return { layout, reshapedNetwork };
};
const compare = (a, b, sortField) => {
    const aVal = select(a, sortField);
    const bVal = select(b, sortField);
    if (aVal < bVal) {
        return 1;
    }
    else if (aVal === bVal) {
        return 0;
    }
    return -1;
};
export { layoutBioFabric };
//# sourceMappingURL=layoutBioFabric.js.map