import { deviation, max, mean, median, min, mode, sum, variance, quantile, minIndex, maxIndex } from "d3-array";
import { group } from "./group";
import {select} from "./select";

import type {AggregationFunc, AggFuncType, AggregationSpec} from "../types/data-transforms";




const uncorrectedVariance = (xs: any[]) => variance(xs) * (xs.length - 1) / (xs.length);
const aggregationFunctions: Record<AggregationFunc, AggFuncType> = {
  count: (xs: any[]) => xs.length,
  valid: (xs: any[]) => xs.filter(x => x !== null && typeof (x) !== "undefined" && !(typeof (x) === "number" && Number.isNaN(x))).length,
  missing: (xs: any[]) => xs.filter(x => x === null || typeof (x) === "undefined" || x === "").length,
  distinct: (xs: any[]) => new Set(xs).size,

  sum,
  product: (xs: any[]) => xs.reduce((res, item) => res * item, 1),
  mean,
  average: mean,

  variance: uncorrectedVariance,
  variancep: variance,

  stdev: (xs: any[]) => Math.sqrt(uncorrectedVariance(xs)),
  stdevp: deviation, // bias-corrected

  median,
  q1: (xs: any[]) => quantile(xs, 0.25),
  q3: (xs: any[]) => quantile(xs, 0.75),

  min,
  max,
  argmin: (xs: any[]) => xs[minIndex(xs)],
  argmax: (xs: any[]) => xs[maxIndex(xs)],

  mode
};

// not implemented: stderr, ci0, ci1
// new addition not in Vega: mode
// unlike Vega, we don't provide a "values" aggregation; instead use a grouping



// performs a single aggregation operation
const computeAggregation = (data: any, spec: AggregationSpec) => {

// group data

  const grouped = spec.groupby ? group({ data }, "data", spec.groupby) : { all: data };
  const result = [];


  if (!spec.fields || spec.fields.length === 0) {
    // no fields specified, so default to just doing a COUNT aggregation

    for (const groupName of Object.keys(grouped)) {
      const groupData = grouped[groupName];

      const resultFieldName = spec.as?.[0] || `count`;
      const aggregatedGroup = {};

      for (const field of spec.groupby) {
        aggregatedGroup[field] =  select(groupData[0], field) // groupData[0][field];
      }

      aggregatedGroup[resultFieldName] = aggregationFunctions.count(groupData);

      result.push(aggregatedGroup);
    }

  } else {
    // do the specified aggregation

    for (const groupName of Object.keys(grouped)) {
      const groupData = grouped[groupName];
      const aggregatedGroup = {};
      for (const field of spec.groupby || []) {
        aggregatedGroup[field] = select(groupData[0], field)  // groupData[0][field];
      }

      for (const i in spec.fields) {
        const field = spec.fields[i];
        const agFuncName = spec.ops?.[i] || "count";
        const aggregationFunction = aggregationFunctions[agFuncName];
        const resultFieldName = spec.as?.[i] || agFuncName;
        aggregatedGroup[resultFieldName] = aggregationFunction(groupData.map(d => d[field]));
      }


      // const aggregatedGroup = spec.fields.map( (field, i) => aggregationFunctions[] )
      result.push(aggregatedGroup);
    }
  }


  return result;
};

export { computeAggregation, aggregationFunctions };