import { GeocodingSpec } from "../types/data-transforms";

// https://positionstack.com/documentation
// UGH: batch requests requires professional plan or higher :-(


// We use https://geocode.maps.co/
// which is powered by OSM and Nominatim


const sleep = (ms: number) => new Promise(resolve => setTimeout(resolve, ms));


const geocode = async (data: any, spec: GeocodingSpec) => {

  const locations = new Map();

  // read cached locations from LocalStorage
  let cachedLocations;
  if (window) {
    cachedLocations = JSON.parse(window.localStorage.getItem("locations") ?? "{}");
    for (const locationName of Object.keys(cachedLocations)) {
      locations.set(locationName, cachedLocations[locationName]);
    }
  }

  for (const datum of data) {

    let placeQueryString;

    if ("locationField" in spec) {
      placeQueryString = `q=${datum[spec.locationField]}`;
    } else {
      const terms = [];

      const f = n => encodeURIComponent(datum[spec[n]])

      if ("streetField" in spec) {
        terms.push(`street=${f("streetField")}`);
      }
      if ("cityField" in spec) {
        terms.push(`city=${f("cityField")}`);
      }
      if ("stateField" in spec) {
        terms.push(`state=${f("stateField")}`);
      }
      if ("countryField" in spec) {
        terms.push(`country=${f("countryField")}`);
      }
      if ("postalCodeField" in spec) {
        terms.push(`postalcode=${f("postalCodeField")}`);
      }

      placeQueryString = terms.join("&");
    }

    if (locations.has(placeQueryString)) {
      const location = locations.get(placeQueryString);
      datum[spec.lonAs ?? "lon"] = location.lon;
      datum[spec.latAs ?? "lat"] = location.lat

      if ("nameAs" in spec){
        datum[spec.nameAs] = location.display_name;
      }

    } else {

      // Respect the limit of 2 requests/s listed at (https://geocode.maps.co/)
      await sleep(800);
      const url = `https://geocode.maps.co/search?${placeQueryString}`;

      const response = await fetch(url, {
        mode: 'cors',
          headers: {
    'User-Agent': 'ANYTHING_WILL_WORK_HERE'
  }
      }).then(r => r.json());


      if (response.length > 0) {
        datum[spec.lonAs ?? "lon"] = response[0].lon;
        datum[spec.latAs ?? "lat"] = response[0].lat;

      if ("nameAs" in spec){
        datum[spec.nameAs] = response[0].display_name;
      }

        locations.set(placeQueryString, response[0]);
      } else {
        locations.set(placeQueryString, {lon: undefined, lat: undefined}); // otherwise a re-run will retry the query
      }
    }

  }

  // write fetched locations back to LocalStorage
  if (window) {
    for (const locationName of locations.keys()) {
      cachedLocations[locationName] = locations.get(locationName);
    }
    window.localStorage.setItem("locations", JSON.stringify(cachedLocations));
  }



  return data;
};


export {geocode};