import { repositionLineSegments } from "respacer";
import { select } from "../data/select";
import { codegenExpression, parseExpression } from "vega-expression";
import {
  Layout, LayoutJitterSpec,
  LayoutTransform, LayoutTransformBeeswarm,
  LayoutTransformOffset,
  LayoutTransformOrientLinks,
  LayoutTransformSet, Position
} from "../types/layout";
import { forceCollide, forceSimulation, forceX, forceY } from "d3-force";
import { mean } from "d3-array";

const setField = (layout: Layout, spec: LayoutTransformSet) => {
  const transformedLayout: Layout = new Map();
  for (const e of layout.keys()) {
    // get datum by filtering? then pos to allow value?
    transformedLayout.set(e, { ...layout.get(e), [spec.field]: spec.value });
  }
  return transformedLayout;
};


const offset = (layout: Layout, data: any[], spec: LayoutTransformOffset) => {
  const direction = spec.direction === "horizontal" ? "x" : "y";
  const width = spec.width ?? 1000;

  const points = [];

  if ("condition" in spec) {

    const codegen = codegenExpression({
      globalvar: "globalVar"
    });

    const funcCode = codegen(parseExpression(spec.condition));

    // First argument means that the function argument will be interpreted as having name "globalVar"
    // so we can avoid using an actual global variable
    const fn = Function("globalVar", "\"use strict\"; return (" + funcCode.code + ")");

    for (const datum of data) {
      if (fn({ datum })) {
        points.push({
          id: datum.id,
          ...layout.get(datum.id)
        });
      }
    }

  } else {
    for (const id of layout.keys()) {
      points.push({
        id,
        ...layout.get(id)
      });
    }
  }


  // TODO: should pass in bounds as "width" option...
  const repositioned = repositionLineSegments(points, {
    oldPositionName: direction,
    widthName: (direction === "x") ? "width" : "height",
    newPositionName: "newPos",
    minSpacing: spec.minSpacing ?? 0,
    width
  });


  const transformedLayout: Layout = new Map();
  for (const e of layout.keys()){
    transformedLayout.set(e, layout.get(e));
  }
  for (const pos of repositioned) {
    const p = layout.get(pos.id);
    transformedLayout.set(pos.id, { ...p, [direction]: pos.newPos}); // , _shifted: p[direction] !== pos.newPos
  }

  return transformedLayout;
};


const orientLinks = (layout: Layout, spec: LayoutTransformOrientLinks, state) => {
  const network = state[spec.network];
  const field = spec.direction === "vertical" ? "x" : "y";

  const transformedLayout: Layout = new Map();
  for (const e of layout.keys()){
    transformedLayout.set(e, layout.get(e));
  }

  for (const link of network.links) {
    transformedLayout.get(link.target.id)[field] = layout.get(link.source.id)[field];
  }
  return transformedLayout;
};

const jitter = (layout: Layout, spec: LayoutJitterSpec) => {

  const transformedLayout: Layout = new Map();
  for (const e of layout.keys()) {
    const pos = layout.get(e);

    let x = pos.x;
    if ("x" in spec) {
      x += spec.x * 2 * (Math.random() - 0.5);
    }

    let y = pos.y;
    if ("y" in spec) {
      y += spec.y * 2 * (Math.random() - 0.5);
    }

    transformedLayout.set(e, { ...pos, x, y });
  }

  return transformedLayout;
};

const beeswarm = (layout: Layout, spec: LayoutTransformBeeswarm) => {

  // const nodes = [...layout.values()];
  const nodes = [];
  for (const id of layout.keys()) {
    nodes.push({
      id,
      ...layout.get(id)
    });
  }

  const simulation = forceSimulation(nodes as any[]);

  simulation.force("x", forceX().strength(spec.gravityX ?? 0.1).x(d => d.x));
  simulation.force("y", forceY().strength(spec.gravityY ?? 0.1).y(d => d.y));

  simulation.force("collision", forceCollide().radius((d) => Math.max((d as Position).width, (d as Position).height) / 2).strength(spec.collisionStrength ?? 1));


  // TODO: reconsider termination condition
  let meanV = 1e9;
  do {
    simulation.tick();
    meanV = mean(nodes.map((n: any) => Math.sqrt(Math.pow(n.vx, 2) + Math.pow(n.vy, 2)))) || 0;
    // console.log(mean_v);
  } while (meanV > 1e-3);
  simulation.stop();

  const transformedLayout: Layout = new Map();
  for (const node of nodes) {
    transformedLayout.set(node.id, { ...layout.get(node.id), x: node.x, y: node.y });
  }
  return transformedLayout;

};


const transformLayout = (transformSpec: LayoutTransform, layoutSpec, layout, state, funcs) => {

  // this is repetitive with computeAttributeLayout()
  let data;
  if ("network" in layoutSpec) {
    data = select(state, layoutSpec.network).nodes;
  } else {
    data = select(state, layoutSpec.data);
  }

  if (transformSpec.type === "set") {
    return setField(layout, transformSpec);
  } else if (transformSpec.type === "offset") {
    return offset(layout, data, transformSpec);
  } else if (transformSpec.type === "orientLinks") {
    return orientLinks(layout, transformSpec, state);
  } else if (transformSpec.type === "jitter") {
    return jitter(layout, transformSpec);
  } else if (transformSpec.type === "beeswarm") {
    return beeswarm(layout, transformSpec);
  } else if (transformSpec.type === "function"){
    return funcs[transformSpec.function](layout, transformSpec)
  }

};

export { transformLayout };
