import { grid } from "./grid";

import type { AttributeLayoutSpec, GridLayout } from "../types/layout";
import { Axis, Layout } from "../types/layout";
import { applyOrderOrScale, getTicks } from "./utils";

export const getPos = (datum, spec: AttributeLayoutSpec, state, numPoints: number) => {

  const gridPatterns = ["row-by-row", "col-by-col", "row-snake", "col-snake", "diagonal", "diagonal-snake", "z-curve", "hilbert"];

  if (gridPatterns.includes(spec.pattern)) {
    return getPosOnGrid(datum, spec, state, numPoints);
  } else if (spec.pattern === "spiral") {
    return getPosOnSpiral(datum, spec, state, numPoints);
  } else if (spec.pattern === "circle") {
    return getPosOnCircle(datum, spec, state, numPoints);
  } else if (spec.pattern === "linear" && spec.orientation === "vertical") {
    return { x: 0, y: applyOrderOrScale(datum, spec, state) };
  } else if (spec.pattern === "linear") {
    return { x: applyOrderOrScale(datum, spec, state), y: 0 };
  }
};


const getPosOnGrid = (datum, spec, state, numPoints) => {
  const s = spec as GridLayout;

  // const i = state[s.order][datum.id];
  const i = applyOrderOrScale(datum, spec, state);

  const rows = s.numRows ??  Math.ceil(Math.sqrt(numPoints));
  const cols = s.numCols ??  Math.ceil(Math.sqrt(numPoints));

  const pos = grid(i, rows, cols, s.pattern);

  return {
    x: pos[0] * (1 + (spec.paddingX ?? 0)),
    y: pos[1] * (1 + (spec.paddingY ?? 0))
  };
};

const getPosOnSpiral = (datum, spec, state, numPoints) => {

  let angularSeparation = 30;
  if ("angularSeparation" in spec) {
    angularSeparation = spec.angularSeparation;
  } else if ("maxAngle" in spec) {
    angularSeparation = 360 / numPoints;
  }

  let startAngle = 2 * Math.PI;
  if ("startAngle" in spec) {
    startAngle = spec.startAngle;
  }


  const i = applyOrderOrScale(datum, spec, state);
  const angleRads = (i * angularSeparation) * (Math.PI / 180) + startAngle;

  let r;
  const a = 10;
  if (!("shape" in spec) || spec.shape === "Archimedes") {
    r = a * angleRads;
  } else if (spec.shape === "hyperbolic") {
    r = a / (angleRads);
    console.log({ i, r, angleRads });
  } else if (spec.shape === "logarithmic") {
    const k = 1;
    r = a * Math.exp(k * angleRads);
  } else if (spec.shape === "lituus") {
    r = a / Math.sqrt(angleRads);
  }

  return { x: r * Math.sin(angleRads), y: r * Math.cos(angleRads) };
};


const getPosOnCircle = (datum, spec, state, numPoints) => {
  const n = applyOrderOrScale(datum, spec, state);
  const r = 1;

  let theta;
  if ("order" in spec) {
    theta = 2 * Math.PI * n / numPoints;
  } else {
    const domain = state[spec.scale].domain();
    theta = 2 * Math.PI * (n - domain[0]) / (domain[domain.length - 1] - domain[0]);
  }

  return { x: r * Math.sin(theta), y: r * Math.cos(theta) };
};


export const layoutRings = (data, spec, state): [Layout, Axis[]] => {
  const layout: Layout = new Map();
  const axes: Axis[] = [];

  const ringsWithAxes = [];

  for (const datum of data) {

    const n = applyOrderOrScale(datum, spec, state);

    // ensure numSegments is an array
    let numSegmentsThisRing = spec.numSegments[0];
    let ring = 0;
    let remainder = n + 1; // add one as ordered position 0 is first element

    while (remainder > numSegmentsThisRing) {
      remainder -= numSegmentsThisRing;
      ring++;

      if (ring < spec.numSegments.length) {
        numSegmentsThisRing = spec.numSegments[ring];
      }
    }

    const theta = 2 * Math.PI * remainder / numSegmentsThisRing;

    layout.set(datum.id, { x: ring * Math.sin(theta), y: ring * Math.cos(theta) });

    if (!ringsWithAxes.includes(ring)) {
      ringsWithAxes.push(ring);
      axes.push({
        "type": "polar",
        "x": 0,
        "y": 0,
        "r": ring,
        ticks: [getTicks(spec, state).map(t => ({ x: ring * Math.sin(t[0]), y: ring * Math.cos(t[0]), label: t[1] }))]
      });
    }

  }

  return [layout, axes];
};