import { Axis, Layout, MapLayoutSpec, Position } from "../types/layout";
import { convertVal } from "../complete";

const computeMapLayout = (state, spec: MapLayoutSpec, bounds): [Layout, Axis[]] => {

  let elements;
  if ("data" in spec) {
    elements = state[spec.data];
  } else if ("network" in spec) {
    elements = state[spec.network].nodes;
  }

  const layout: Layout = new Map();

  const map = state[spec.map + "#leafletMap"];
  const mapSpec = state[spec.map + "#config"];

  for (const element of elements) {
    const lng = convertVal(spec.lng, element, state);
    const lat = convertVal(spec.lat, element, state);
    if (map) {
      return [state[spec.name], []]; // don't recompute layout: this will mess up the zoom
      // const xy = map.latLngToContainerPoint([lat, lng]);
      // layout.set(element.id, { x: xy.x, y: xy.y, width: 0, height: 0 });
    } else {
      // if map not created yet, return the map placeholder layout
      const pos = latLngToContainerPoint({ lat, lng }, { ...spec, center: mapSpec.center, zoomLevel: mapSpec.zoomLevel });
      layout.set(element.id, { x: pos.x, y: pos.y, width: 0, height: 0 });
    }
  }

  return [layout, []];
};


// see https://github.com/Leaflet/Leaflet/blob/fb0b2a169ebde77a8bed1715bb07d436b60328a2/src/geo/crs/CRS.EPSG3857.js
const earthRadius = 6378137;
const project = (latlng) => {
  const MAX_LATITUDE = 85.0511287798;

  const d = Math.PI / 180,
    max = MAX_LATITUDE,
    lat = Math.max(Math.min(max, latlng.lat), -max),
    sin = Math.sin(lat * d);

  return {
    x: earthRadius * latlng.lng * d,
    y: earthRadius * Math.log((1 + sin) / (1 - sin)) / 2
  };
};

// Entries of transformation matrix (a b; c d) from https://github.com/Leaflet/Leaflet/blob/fb0b2a169ebde77a8bed1715bb07d436b60328a2/src/geo/crs/CRS.EPSG3857.js#L21
const transform = (point, scale1, scale2) => {
  const a = scale1;
  const b = 0.5;
  const c = -scale1;
  const d = 0.5;

  const scale = scale2 || 1;
  return {
    x: scale * (a * point.x + b),
    y: scale * (c * point.y + d)
  };
};


const latLngToContainerPoint = (latlng, spec) => {
  // This does an operation equivalent to Leaflets map.latLngToContainerPoint()
  // It is only valid before any user interaction, when the mapPanePos is {x: 0, y: 0}.

  const width = spec.width ?? 1000;
  const height = spec.height ?? 1000;
  const zoomLevel = spec.zoomLevel;

  const scale1 = 0.5 / (Math.PI * earthRadius);
  const scale2 = 256 * Math.pow(2, zoomLevel);

  const centerPos = transform(project({ lat: spec.center[0], lng: spec.center[1] }), scale1, scale2);
  const pixelPos = transform(project(latlng), scale1, scale2);

  return {
    x: pixelPos.x - (centerPos.x - width/2),
    y: pixelPos.y - (centerPos.y - height/2)
  };
};

export { computeMapLayout };