import fs from "fs";

import { constructNetworkComplex, getValue } from "./constructNetworkComplex";

import { join } from "../../data/join";
import { unroll } from "../../data/unroll";
import { computeConnectivityBasedAttribute } from "../connectivityBasedAttribute";

test("can get all values", () => {
  const datum = {
    Name1: "Marie Boucher et Hubert Antheaume Cie",
    Place1: "St Nicolas, Nantes",
    Content: "Commerce",
    Name2: "François Lemasson",
    Place2: "Orléans",
    "Other places": "",
    "Etude Notaire": "Verger",
    "Cote ADLA": "4/E2/1954",
    "Nature acte": "Cambie",
    Date: "23/09/1672"
  };


  // wildcard selector
  expect(getValue(datum, "*")).toEqual(datum);

  // selector is name of a field, so selects value of that field
  expect(getValue(datum, "Name1")).toEqual({ Name1: "Marie Boucher et Hubert Antheaume Cie" });

  // selector is not name of a field, so selects literal value of the string
  // expect(getValue(datum, "Name111")).toEqual("Name111");

  // selector explicitly requests a field value
  expect(getValue(datum, { field: "Content" })).toEqual({ Content: "Commerce" });

  // but with a specified output field name
  expect(getValue(datum, { field: "Content", as: "FOO" })).toEqual({ FOO: "Commerce" });

  // selector explicitly requests a literal value
  expect(getValue(datum, { value: "Content" })).toEqual({ Content: "Content" });

  // but with a specified output field name
  expect(getValue(datum, { value: "Content", as: "FOO" })).toEqual({ FOO: "Content" });

});


/* ************************************************************************** */
test.skip("can import marie-boucher edge table", () => {

  // TODO: also handle node locations

  const marieBoucher = require("../../../example_datasets/vistorian/marieboucher.json");


  // N.B. the date parsing should have happened during the data block that loaded the OSN file

  const spec = [
    {
      data: "marieBoucher",
      yieldsLinks: [
        {
          source_id: "Name1",
          target_id: "Name2",
          data: ["*"]
        }
      ]
    }
  ];

  constructNetworkComplex({ marieBoucher }, spec);

  // If only one edge provided, can automatically pass through all other fields as edge data


});

test.skip("can import a node-table", () => {

  const nodeTable = require("../../../example_datasets/vistorian/nodeTable.json");

  const spec = [
    {
      data: "nodeTable",
      yieldsLinks: [
        {
          source_id: "Node",
          target_id: "Mother",
          data: [
            { value: "Mother", as: "type" }
          ]
        },
        {
          source_id: "Node",
          target_id: "Father",
          data: [
            { value: "Father", as: "type" }
          ]
        },
        {
          source_id: "Node",
          target_id: "God-Father",
          data: [
            { value: "God-father", as: "type" }
          ]
        },
        {
          source_id: "Node",
          target_id: "God-Mother",
          data: [
            { value: "God-Mother", as: "type" }
          ]
        },
        {
          source_id: "Node",
          target_id: "Place-of-birth", // or  {target: label: {"Place-of-birth"}, data: {type: "place"}  }
          data: [
            { value: "Place-of-Birth", as: "type" }
          ]
        }
      ]
    }

  ];

  constructNetworkComplex({ nodeTable }, spec);


  // alt spec yields one node per row, with edges to the others:
  // This is syntactic sugar that is not yet implemented
  /*
  const spec1 = [
    {
      data: "nodeTable",
      yieldsNodes: [
        { id: "Node" }
      ],
      yieldsLinks: [
        {
          target_id: "Mother",
          data: [
            {value: "Mother", as: "type"}
          ]
        },
        {
          target_id: "Father",
          data: [
            {value: "Father", as: "type"}
          ]
        },
        {
          target_id: "God-Father",
          data: [
            {value: "God-father", as: "type"}
          ]
        },
        {
          target_id: "God-Mother",
          data: [
            {value: "God-Mother", as: "type"}
          ]
        },
        {
          target_id: "Place-of-birth", // or  {target: label: {"Place-of-birth"}, data: {type: "place"}  }
          data: [
            {value: "Place-of-Birth", as: "type"}
          ]
        }
      ]
    }
  ];
  */
});

test("can import the origraph politics example", () => {
  const senators = require("../../../../editor/public/data/origraph/politics/senateMembers.json");
  const votes = require("../../../../editor/public/data/origraph/politics/senate_votes_yemen.json");
  const pressReleases = require("../../../../editor/public/data/origraph/politics/pressReleases.json");
  const contributions = require("../../../../editor/public/data/origraph/politics/allContributions.json");

  const spec =
    {
      "name": "politics-network",
      "parts": [
        {
          "data": "senators",
          "yieldsNodes": [
            {
              "type": "senator",
              "data": [
                "*"
              ]
            }
          ]
        },
        {
          "data": "votes",
          "yieldsLinks": [
            {
              "source_id": {
                "field": "member_id"
              },
              "source_node_type": "senator",
              "source_id_field": "id",
              "target_id": {
                "field": "vote_position"
              },
              "target_node_type": "vote"
            }
          ]
        },
        {
          "data": "pressReleases",
          "yieldsNodes": [
            {
              "type": "pressRelease",
              "id_field": "url",
              "data": [
                "*"
              ]
            }
          ],
          "yieldsLinks": [
            {
              "source_id": {
                "field": "member_id"
              },
              "source_node_type": "senator",
              "source_id_field": "id",
              "target_id": {
                "field": "vote_position"
              },
              "target_id_field": "member_id",
              "target_node_type": "pressRelease"
            }
          ]
        },
        {
          "data": "contributions",
          "yieldsLinks": [
            {
              "source_id": {
                "field": "candidate"
              },
              "source_node_type": "senator",
              "source_id_field": "fec_candidate_id",
              "target_id": {
                "field": "fec_committee_name"
              },
              "target_node_type": "Committee"
            }
          ]
        }
      ]
    }
  ;

  const network = constructNetworkComplex({ senators, votes, pressReleases, contributions }, spec);


  // This defines the network: as a post-processing step, also did projection committee-senator-vote


  fs.writeFile("senate-network-assembled.json", JSON.stringify(network, null, 4), () => null);


});

// TODO: setting link data from yieldLinks - e.g. role=cast

test("can import the origraph movies example", () => {

  let bechedel = require("../../../../editor/public/data/origraph/movies/data_files/bechdel.json");
  const credits = require("../../../../editor/public/data/origraph/movies/data_files/credits.json");
  const movies = require("../../../../editor/public/data/origraph/movies/data_files/movies.json");
  const people = require(".../../../../editor/public/data/origraph/movies/data_files/people.json");

  // join the bechedel and IMDB data
  bechedel = bechedel.map(d => ({ ...d, ttimdbid: "tt" + d.imdbid }));
  join(movies, bechedel, "imdb_id", "ttimdbid", ["rating"]);

  // reformat credits data
  const { cast, crew } = unroll(credits, {fields: ["cast", "crew"]});

  // connect up the network
  const spec = {
    "parts": [
      {
        data: "movies",
        yieldsNodes: [
          {
            type: "movie",
            data: ["*"]

          }
        ],

        mapYieldsLinks: [
          {
            field: "production_companies",
            // since target is set, the source could be inferred if there is only one entry in yieldNodes
            source_id: { field: "parent.id" },
            source_node_type: "movie",
            source_id_field: "id",

            target_id: { field: "child.name" },
            target_node_type: "production_company"
          }
        ]


        // TODO: yields lnk to production companies
      },
  {
    data: "people",
      yieldsNodes:
    [
      {
        type: "person",
        data: ["*"]
      }
    ]
  },

  {
    data: "cast",
      yieldsLinks:
    [
      {
        source_id: { field: "id" },
        source_node_type: "person",
        source_id_field: "id",

        target_id: { field: "movie_id" },
        target_node_type: "movie",
        target_id_field: "id"
      }
    ]
  },

  {
    data: "crew",
      yieldsLinks:
    [
      {
        source_id: { field: "id" },
        source_node_type: "person",
        source_id_field: "id",

        target_id: { field: "movie_id" },
        target_node_type: "movie",
        target_id_field: "id"
      }
    ]
  }
]
};


  const network = constructNetworkComplex({ movies, people, cast, crew }, spec);

  //const people_nodes = network.nodes.filter(n => n._type === "person");
  //computeConnectivityasedAttribute(people_nodes, network.links, "node.data.rating", "mean_bechedel_score", "mean", );

  const companyNodes = network.nodes.filter(n => n._type === "production_company"); // actual example uses film studios
  const ll = network.links.filter(l => l.target._type === "production_company" && l.source._type === "movie");
  computeConnectivityBasedAttribute(companyNodes, ll, {
    attributeSelector: "node.data.rating",
    as: "mean_bechedel_score",
    operation: "mean",
    direction: "in"
  });

  fs.writeFile("movies-network-assembled.json", JSON.stringify(network, null, 4), () => null);

});