import { parseGraphml } from "./importGraphml";

test("can import simple network", () => {
  // http://graphml.graphdrawing.org/primer/simple.graphml

  const networkString = `
<?xml version="1.0" encoding="UTF-8"?>
<graphml xmlns="http://graphml.graphdrawing.org/xmlns"  
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="http://graphml.graphdrawing.org/xmlns
     http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd">
  <graph id="G" edgedefault="undirected">
    <node id="n0"/>
    <node id="n1"/>
    <node id="n2"/>
    <node id="n3"/>
    <node id="n4"/>
    <node id="n5"/>
    <node id="n6"/>
    <node id="n7"/>
    <node id="n8"/>
    <node id="n9"/>
    <node id="n10"/>
    <edge source="n0" target="n2"/>
    <edge source="n1" target="n2"/>
    <edge source="n2" target="n3"/>
    <edge source="n3" target="n5"/>
    <edge source="n3" target="n4"/>
    <edge source="n4" target="n6"/>
    <edge source="n6" target="n5"/>
    <edge source="n5" target="n7"/>
    <edge source="n6" target="n8"/>
    <edge source="n8" target="n7"/>
    <edge source="n8" target="n9"/>
    <edge source="n8" target="n10"/>
  </graph>
</graphml>
`;

  const network = parseGraphml(networkString, { format: "graphml" });

  expect(network).toEqual({
      "nodes": [
        { "id": "n0" },
        { "id": "n1" },
        { "id": "n2" },
        { "id": "n3" },
        { "id": "n4" },
        { "id": "n5" },
        { "id": "n6" },
        { "id": "n7" },
        { "id": "n8" },
        { "id": "n9" },
        { "id": "n10" }
      ],
      "links": [
        { "source": "n0", "target": "n2", "directed": false },
        { "source": "n1", "target": "n2", "directed": false },
        { "source": "n2", "target": "n3", "directed": false },
        { "source": "n3", "target": "n5", "directed": false },
        { "source": "n3", "target": "n4", "directed": false },
        { "source": "n4", "target": "n6", "directed": false },
        { "source": "n6", "target": "n5", "directed": false },
        { "source": "n5", "target": "n7", "directed": false },
        { "source": "n6", "target": "n8", "directed": false },
        { "source": "n8", "target": "n7", "directed": false },
        { "source": "n8", "target": "n9", "directed": false },
        { "source": "n8", "target": "n10", "directed": false }
      ]
    }
  );
});


test("adds correct reverse edges", () => {
  // http://graphml.graphdrawing.org/primer/simple.graphml

  const networkString = `
<?xml version="1.0" encoding="UTF-8"?>
<graphml xmlns="http://graphml.graphdrawing.org/xmlns"  
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="http://graphml.graphdrawing.org/xmlns
     http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd">
  <graph id="G" edgedefault="undirected">
    <node id="n0"/>
    <node id="n1"/>
    <node id="n2"/>
    <node id="n3"/>
    <edge source="n0" target="n2" directed="true"/>
    <edge source="n1" target="n2" directed="false"/>
    <edge source="n2" target="n3"/>
  </graph>
</graphml>
`;

  const network = parseGraphml(networkString, { format: "graphml", addReverseLinks: true });

  expect(network).toEqual({
      "nodes": [
        { "id": "n0" },
        { "id": "n1" },
        { "id": "n2" },
        { "id": "n3" }
      ],
      "links": [
        { "source": "n0", "target": "n2", "directed": true },

        { "source": "n1", "target": "n2", "directed": true },
        { "source": "n2", "target": "n1", "directed": true },

        { "source": "n2", "target": "n3", "directed": true },
        { "source": "n3", "target": "n2", "directed": true }
      ]
    }
  );

  const network2 = parseGraphml(networkString, { format: "graphml", addReverseLinks: false });

  expect(network2).toEqual({
      "nodes": [
        { "id": "n0" },
        { "id": "n1" },
        { "id": "n2" },
        { "id": "n3" }
      ],
      "links": [
        { "source": "n0", "target": "n2", "directed": true },

        { "source": "n1", "target": "n2", "directed": false },

        { "source": "n2", "target": "n3", "directed": false }
      ]
    }
  );
});


test("can import a network with attributes", () => {

  // this test includes missing values with and without defaults

  // http://graphml.graphdrawing.org/primer/attributes.graphml

  const networkString = `
<?xml version="1.0" encoding="UTF-8"?>
<graphml xmlns="http://graphml.graphdrawing.org/xmlns"  
      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
      xsi:schemaLocation="http://graphml.graphdrawing.org/xmlns 
        http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd">
  <key id="d0" for="node" attr.name="color" attr.type="string">
    <default>yellow</default>
  </key>
  <key id="d1" for="edge" attr.name="weight" attr.type="double"/>
  <graph id="G" edgedefault="undirected">
    <node id="n0">
      <data key="d0">green</data>
    </node>
    <node id="n1"/>
    <node id="n2">
      <data key="d0">blue</data>
    </node>
    <node id="n3">
      <data key="d0">red</data>
    </node>
    <node id="n4"/>
    <node id="n5">
      <data key="d0">turquoise</data>
    </node>
    <edge id="e0" source="n0" target="n2">
      <data key="d1">1.0</data>
    </edge>
    <edge id="e1" source="n0" target="n1">
      <data key="d1">1.0</data>
    </edge>
    <edge id="e2" source="n1" target="n3">
      <data key="d1">2.0</data>
    </edge>
    <edge id="e3" source="n3" target="n2"/>
    <edge id="e4" source="n2" target="n4"/>
    <edge id="e5" source="n3" target="n5"/>
    <edge id="e6" source="n5" target="n4">
      <data key="d1">1.1</data>
    </edge>
  </graph>
</graphml>
`;

  const network = parseGraphml(networkString, { format: "graphml" });

  expect(network).toEqual({
      "nodes": [
        { "id": "n0", "color": "green" },
        { "id": "n1", "color": "yellow" },
        { "id": "n2", "color": "blue" },
        { "id": "n3", "color": "red" },
        { "id": "n4", "color": "yellow" },
        { "id": "n5", "color": "turquoise" }
      ],
      "links": [
        {
          "id": "e0",
          "source": "n0",
          "target": "n2",
          "weight": 1,
          "directed": false
        },
        {
          "id": "e1",
          "source": "n0",
          "target": "n1",
          "weight": 1,
          "directed": false
        },
        {
          "id": "e2",
          "source": "n1",
          "target": "n3",
          "weight": 2,
          "directed": false
        },
        {
          "id": "e3",
          "source": "n3",
          "target": "n2",
          "weight": undefined,
          "directed": false
        },
        {
          "id": "e4",
          "source": "n2",
          "target": "n4",
          "weight": undefined,
          "directed": false
        },
        {
          "id": "e5",
          "source": "n3",
          "target": "n5",
          "weight": undefined,
          "directed": false
        },
        {
          "id": "e6",
          "source": "n5",
          "target": "n4",
          "weight": 1.1,
          "directed": false
        }
      ]
    }
  );
});


test("adds specify which graph to import", () => {
  // http://graphml.graphdrawing.org/primer/simple.graphml

  const networkString = `
<?xml version="1.0" encoding="UTF-8"?>
<graphml xmlns="http://graphml.graphdrawing.org/xmlns"  
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="http://graphml.graphdrawing.org/xmlns
     http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd">
  <graph id="G" edgedefault="undirected">
    <node id="n0"/>
    <node id="n1"/>
    <node id="n2"/>
    <node id="n3"/>
    <edge source="n0" target="n2" directed="true"/>
    <edge source="n1" target="n2" directed="false"/>
    <edge source="n2" target="n3"/>
  </graph>
  
   <graph id="G2" edgedefault="undirected">
    <node id="n0"/>
    <node id="n1"/>
    <node id="n2"/>
    <node id="n3"/>
    <edge source="n1" target="n3" directed="true"/>
    <edge source="n2" target="n3" directed="false"/>
    <edge source="n3" target="n0"/>
  </graph>


</graphml>
`;

  const firstNetwork = {
    "nodes": [
      { "id": "n0" },
      { "id": "n1" },
      { "id": "n2" },
      { "id": "n3" }
    ],
    "links": [
      { "source": "n0", "target": "n2", "directed": true },
      { "source": "n1", "target": "n2", "directed": false },
      { "source": "n2", "target": "n3", "directed": false }
    ]
  };

  const secondNetwork = {
    "nodes": [
      { "id": "n0" },
      { "id": "n1" },
      { "id": "n2" },
      { "id": "n3" }
    ],
    "links": [
      { "source": "n1", "target": "n3", "directed": true },
      { "source": "n2", "target": "n3", "directed": false },
      { "source": "n3", "target": "n0", "directed": false }
    ]
  };

  const network = parseGraphml(networkString, { format: "graphml" });
  expect(network).toEqual(firstNetwork);

  const network2 = parseGraphml(networkString, { format: "graphml", graphIndex: 0 });
  expect(network2).toEqual(firstNetwork);

  const network3 = parseGraphml(networkString, { format: "graphml", graphIndex: 1 });
  expect(network3).toEqual(secondNetwork);

  const network4 = parseGraphml(networkString, { format: "graphml", graphId: "G" });
  expect(network4).toEqual(firstNetwork);

  const network5 = parseGraphml(networkString, { format: "graphml", graphId: "G2" });
  expect(network5).toEqual(secondNetwork);
});
