import { Network, NetworkImportGraphml } from "../../types/network";

import { XMLParser } from "fast-xml-parser";

// For debugging, see https://naturalintelligence.github.io/fast-xml-parser/

export async function importGraphml(spec: NetworkImportGraphml) {
  const data = await fetch(spec.url)
    .then((response) => response.text());

  return parseGraphml(data, spec);

}

export function parseGraphml(graphmlString: string, spec: NetworkImportGraphml): Network {

  const nodes = [];
  const links = [];

  const fieldName = {};
  const fieldDefault = {};
  const nodeFields = [];
  const edgeFields = [];

  const parser = new XMLParser({
    ignoreAttributes: false,
    attributeNamePrefix: ""
  });
  const graphml = parser.parse(graphmlString).graphml;

  //const graph = graphml.graph;
  let graph;
  if (Object.prototype.toString.call(graphml.graph) === "[object Array]") {
    if (typeof spec.graphIndex !== "undefined"){
      graph = graphml.graph[spec.graphIndex];
    } else if (typeof spec.graphId !== "undefined"){
      graph = graphml.graph.filter(g => g.id == spec.graphId)[0];
    } else {
      graph = graphml.graph[0];
    }
  } else {
    graph = graphml.graph;
  }

  let defaultDirected;
  if (graph.edgedefault === "directed"){
    defaultDirected = true;
  } else if (graph.edgedefault === "undirected"){
    defaultDirected = false;
  }

  // process list of attributes for nodes/edges
  if (Object.prototype.toString.call(graphml.key) === "[object Array]") {
    for (const f of graphml.key) {
      fieldName[f.id] = f["attr.name"];

      if (f.default){
        fieldDefault[f.id] = f.default;
      }

      if (f.for === "node") {
        nodeFields.push(f.id);
      } else if (f.for === "edge") {
        edgeFields.push(f.id);
      }
    }
  }

  for (const node of graph.node) {
    let data = {};
    if (Object.prototype.toString.call(node.data) === "[object Array]") {
      data = Object.fromEntries(node.data.map(n => [fieldName[n.key], n["#text"]]));
    } else if (typeof node.data === "object"){
      data = {[fieldName[node.data.key]]: node.data["#text"]};
    }

    // set attributes to default values where needed
    for (const fieldId of nodeFields){
      const name = fieldName[fieldId];
      data[name] = data[name] ?? fieldDefault[fieldId];
    }

    nodes.push({ id: node.id, ...data });
  }

  for (const edge of graph.edge) {
    let data = {};
    if (Object.prototype.toString.call(edge.data) === "[object Array]") {
      data = Object.fromEntries(edge.data.map(n => [fieldName[n.key], n["#text"]]));
    } else if (typeof edge.data === "object"){
      data = {[fieldName[edge.data.key]]: edge.data["#text"]};
    }

    // set attributes to default values where needed
    for (const fieldId of edgeFields){
      const name = fieldName[fieldId];
      data[name] = data[name] ?? fieldDefault[fieldId];
    }

    let isDirected;
    if (typeof edge.directed == "undefined"){
      isDirected = defaultDirected;
    } else {
      isDirected = (edge.directed === "true");
    }

    // the edge identifier is optional
    if (typeof edge.id !== "undefined"){
      data["id"] = edge.id
    }

    if (!isDirected && spec.addReverseLinks) {
      links.push({source: edge.source, target: edge.target, ...data, directed: true});
      links.push({ source: edge.target, target: edge.source, ...data, directed: true });
    } else {
      links.push({ source: edge.source, target: edge.target, ...data, directed: isDirected});
    }
  }

  return { nodes, links };
}
