import { attributeOrdering } from "./attributeOrdering";

const data = [
  { name: "bar", val: 2, id: 0 },
  { name: "baz", val: 3, id: 1 },
  { name: "foo", val: 1, id: 2 }
];


test("can order ascending by numerical field", () => {

  const spec = {
    orderBy: [{ field: "val", direction: "ascending" }]
  };

  const order = attributeOrdering(data, spec);

  // Map {2 => 0, 0 => 1, 1 => 2}
  expect(order).toEqual(new Map([[2,0], [0,1], [1,2]]));
});

test("can give single sorting criteria as object rather than list", () => {

  const spec = {
    orderBy: { field: "val", direction: "ascending" }
  };

  const order = attributeOrdering(data, spec);

  //  Map {2 => 0, 0 => 1, 1 => 2}
  expect(order).toEqual(new Map([[2,0], [0,1], [1,2]]));
});

test("can give single sorting criteria as field name", () => {
  const spec = {
    orderBy: "val"
  };

  const order = attributeOrdering(data, spec);

  //  Map {2 => 0, 0 => 1, 1 => 2}
  expect(order).toEqual(new Map([[2,0], [0,1], [1,2]]));
});


test("can order descending by numerical field", () => {
  const spec = {
    orderBy: [{ field: "val", direction: "descending" }]
  };

  const order = attributeOrdering(data, spec);

  // Map {1 => 0, 0 => 1, 2 => 2}
  expect(order).toEqual(new Map([[1,0], [0,1], [2,2]]));
});


test("respects the allowTies option", () => {
  const repeatedData = [
    { name: "bar", val: 2, id: 0 },
    { name: "baz", val: 3, id: 1 },
    { name: "foo", val: 1, id: 2 },

    { name: "bar", val: 2, id: 3 },
    { name: "baz", val: 3, id: 4 },
    { name: "foo", val: 1, id: 5 }
  ];

  const spec = {
    orderBy: [{ field: "val", direction: "ascending" }],
    allowTies: true
  };

  const order = attributeOrdering(repeatedData, spec);
  //  Map {2 => 0, 5 => 0, 0 => 1, 3 => 1, 1 => 2, 4 => 2}
  expect(order).toEqual(new Map([
    [2, 0],
    [5, 0],
    [0, 1],
    [3, 1],
    [1, 2],
    [4, 2]
  ]));


  const specNoTies = {
    orderBy: [{ field: "val", direction: "ascending" }],
    allowTies: false
  };
  const orderWithoutTies = attributeOrdering(repeatedData, specNoTies);

  //  Map {2 => 0, 5 => 1, 0 => 2, 3 => 3, 1 => 4, 4 => 5}
  expect(orderWithoutTies).toEqual(new Map([
    [2, 0],
    [5, 1],
    [0, 2],
    [3, 3],
    [1, 4],
    [4, 5]
  ]));

});


test("can order ascending by a numerical expression", () => {
  const spec = {
    orderBy: [{ expression: "datum.val * datum.id", direction: "ascending" }]
  };

  const order = attributeOrdering(data, spec);

  // Map {0 => 0, 2 => 1, 1 => 2}
  expect(order).toEqual(new Map([
    [0, 0],
    [2, 1],
    [1, 2]
  ]));
});

test("can order ascending by a string field", () => {

  const spec = {
    orderBy: [{ field: "name", direction: "ascending" }]
  };


  const order = attributeOrdering(data, spec);
  //  Map {0 => 0, 1 => 1, 2 => 2}
  expect(order).toEqual(new Map([
    [0, 0],
    [1, 1],
    [2, 2]
  ]));
});


test("can order descending by a string field", () => {
  const spec = {
    orderBy: [{ field: "name", direction: "descending" }]
  };

  const order = attributeOrdering(data, spec);
  // Map {2 => 0, 1 => 1, 0 => 2}
  expect(order).toEqual(new Map([
    [2, 0],
    [1, 1],
    [0, 2]
  ]));
});


test("can sort on multiple variables", () => {

  let d = [];
  let id = 0;
  for (let a = 0; a < 3; a++) {
    for (let b = 0; b < 3; b++) {
      for (let c = 0; c < 3; c++) {
        d.push({ a, b, c, id });
        id++;
      }
    }
  }

  let spec1 = {
    orderBy: [
      { field: "a" },
      { field: "b" },
      { field: "c" }
    ]
  };


  const order1 = attributeOrdering(d, spec1);
  const sorted1 = getSorted(d, order1);

  expect(sorted1).toEqual(
    [
      { a: 0, b: 0, c: 0, id: 0 },
      { a: 0, b: 0, c: 1, id: 1 },
      { a: 0, b: 0, c: 2, id: 2 },
      { a: 0, b: 1, c: 0, id: 3 },
      { a: 0, b: 1, c: 1, id: 4 },
      { a: 0, b: 1, c: 2, id: 5 },
      { a: 0, b: 2, c: 0, id: 6 },
      { a: 0, b: 2, c: 1, id: 7 },
      { a: 0, b: 2, c: 2, id: 8 },
      { a: 1, b: 0, c: 0, id: 9 },
      { a: 1, b: 0, c: 1, id: 10 },
      { a: 1, b: 0, c: 2, id: 11 },
      { a: 1, b: 1, c: 0, id: 12 },
      { a: 1, b: 1, c: 1, id: 13 },
      { a: 1, b: 1, c: 2, id: 14 },
      { a: 1, b: 2, c: 0, id: 15 },
      { a: 1, b: 2, c: 1, id: 16 },
      { a: 1, b: 2, c: 2, id: 17 },
      { a: 2, b: 0, c: 0, id: 18 },
      { a: 2, b: 0, c: 1, id: 19 },
      { a: 2, b: 0, c: 2, id: 20 },
      { a: 2, b: 1, c: 0, id: 21 },
      { a: 2, b: 1, c: 1, id: 22 },
      { a: 2, b: 1, c: 2, id: 23 },
      { a: 2, b: 2, c: 0, id: 24 },
      { a: 2, b: 2, c: 1, id: 25 },
      { a: 2, b: 2, c: 2, id: 26 }
    ]
  );


  let spec2 = {
    orderBy: [
      { field: "c" },
      { field: "b" },
      { field: "a" }
    ]
  };


  const order2 = attributeOrdering(d, spec2);
  const sorted2 = getSorted(d, order2);

  expect(sorted2).toEqual([
      { a: 0, b: 0, c: 0, id: 0 },
      { a: 1, b: 0, c: 0, id: 9 },
      { a: 2, b: 0, c: 0, id: 18 },
      { a: 0, b: 1, c: 0, id: 3 },
      { a: 1, b: 1, c: 0, id: 12 },
      { a: 2, b: 1, c: 0, id: 21 },
      { a: 0, b: 2, c: 0, id: 6 },
      { a: 1, b: 2, c: 0, id: 15 },
      { a: 2, b: 2, c: 0, id: 24 },
      { a: 0, b: 0, c: 1, id: 1 },
      { a: 1, b: 0, c: 1, id: 10 },
      { a: 2, b: 0, c: 1, id: 19 },
      { a: 0, b: 1, c: 1, id: 4 },
      { a: 1, b: 1, c: 1, id: 13 },
      { a: 2, b: 1, c: 1, id: 22 },
      { a: 0, b: 2, c: 1, id: 7 },
      { a: 1, b: 2, c: 1, id: 16 },
      { a: 2, b: 2, c: 1, id: 25 },
      { a: 0, b: 0, c: 2, id: 2 },
      { a: 1, b: 0, c: 2, id: 11 },
      { a: 2, b: 0, c: 2, id: 20 },
      { a: 0, b: 1, c: 2, id: 5 },
      { a: 1, b: 1, c: 2, id: 14 },
      { a: 2, b: 1, c: 2, id: 23 },
      { a: 0, b: 2, c: 2, id: 8 },
      { a: 1, b: 2, c: 2, id: 17 },
      { a: 2, b: 2, c: 2, id: 26 }
    ]
  );


  let spec3 = {
    orderBy: [
      { field: "a", direction: "descending" },
      { field: "b", direction: "ascending" },
      { field: "c" }
    ]
  };

  const order3 = attributeOrdering(d, spec3);
  const sorted3 = getSorted(d, order3);

  expect(sorted3).toEqual([
      { a: 2, b: 0, c: 0, id: 18 },
      { a: 2, b: 0, c: 1, id: 19 },
      { a: 2, b: 0, c: 2, id: 20 },
      { a: 2, b: 1, c: 0, id: 21 },
      { a: 2, b: 1, c: 1, id: 22 },
      { a: 2, b: 1, c: 2, id: 23 },
      { a: 2, b: 2, c: 0, id: 24 },
      { a: 2, b: 2, c: 1, id: 25 },
      { a: 2, b: 2, c: 2, id: 26 },
      { a: 1, b: 0, c: 0, id: 9 },
      { a: 1, b: 0, c: 1, id: 10 },
      { a: 1, b: 0, c: 2, id: 11 },
      { a: 1, b: 1, c: 0, id: 12 },
      { a: 1, b: 1, c: 1, id: 13 },
      { a: 1, b: 1, c: 2, id: 14 },
      { a: 1, b: 2, c: 0, id: 15 },
      { a: 1, b: 2, c: 1, id: 16 },
      { a: 1, b: 2, c: 2, id: 17 },
      { a: 0, b: 0, c: 0, id: 0 },
      { a: 0, b: 0, c: 1, id: 1 },
      { a: 0, b: 0, c: 2, id: 2 },
      { a: 0, b: 1, c: 0, id: 3 },
      { a: 0, b: 1, c: 1, id: 4 },
      { a: 0, b: 1, c: 2, id: 5 },
      { a: 0, b: 2, c: 0, id: 6 },
      { a: 0, b: 2, c: 1, id: 7 },
      { a: 0, b: 2, c: 2, id: 8 }
    ]
  );

});


const getSorted = (data, order) => {

  let res = [];
  for (const id of order.keys()) {
    res[order.get(id)] = data.filter(d => d.id === +id)[0];
  }
  return res;
};