import { NetworkLink, NetworkNode } from "../types/network";
import { LayoutSpecBipartite } from "../types/topologicalLayout";
import { constructExpressionFunction } from "../data/calculate";
import { applyOrderOrScale } from "../layouts/utils";
import { Layout } from "../types/layout";


const squishOrder = (order, ids) => {

  // make a map from position => item
  const index = [];
  for (const e of order.entries()) {
    const [id, ind] = e;
    //const ind = order.get(id);
    index[ind] = id;
  }

  let ind = 0;
  const newOrder = new Map();
  for (const i in index) {
    const id = index[i];
    if (ids.includes(id)) {
      newOrder.set(id, ind);
      ind++;
    }
  }

  return newOrder;
};


const layoutBipartite = (nodes: NetworkNode[], links: NetworkLink[], spec: LayoutSpecBipartite, state: any, addToState) => {


  const layer1: NetworkNode[] = [];
  const layer2: NetworkNode[] = [];

  const f = constructExpressionFunction({ calculate: spec.inFirstLayer });


  for (const node of nodes) {
    if (f(node, null, state)) {
      layer1.push(node);
    } else {
      layer2.push(node);
    }
  }

  const fixedLayerNumber = spec.fixedLayer ?? 1;

  const fixedNodes = (fixedLayerNumber === 1) ? layer1 : layer2;
  const movableNodes = (fixedLayerNumber === 1) ? layer2 : layer1;

  const movableLayerPos = (fixedLayerNumber === 1) ? 100 : 1;
  const fixedLayerPos = (fixedLayerNumber === 1) ? 1 : 100;


  const fixedPositions = ("x" in spec) ? spec.x : spec.y;
  const xOrYSpec = ("order" in fixedPositions) ? { order: fixedPositions.order } : fixedPositions;

  const layout: Layout = new Map();


  // This is a bit of a hack.
  // If we'e applied an order to a network, it will include the nodes for both layers.
  // This means there will be gaps in the fixed layer where the nodes from the movable layer would be,
  // resulting in an inefficient use of space.
  // We therefore construct a new ordering, by removing these gaps.
  if ("order" in xOrYSpec) {
    state[xOrYSpec.order + "_SQUISHED"] = squishOrder(state[xOrYSpec.order], fixedNodes.map(n => n.id));
    xOrYSpec.order = xOrYSpec.order + "_SQUISHED";
  }


  // Position the fixed nodes
  for (const node of fixedNodes) {
    const xOrY = applyOrderOrScale(node, xOrYSpec, state);

    if ("x" in spec) {
      layout.set(node.id, { x: xOrY, y: fixedLayerPos, width: 1, height: 1 });
    } else {
      layout.set(node.id, { x: fixedLayerPos, y: xOrY, width: 1, height: 1 });
    }
  }

  const fixedNodeIds = fixedNodes.map(n => n.id);
  for (const node of movableNodes) {
    const neighbours = [];
    for (const link of links) {
      if (link.source.id === node.id && fixedNodeIds.includes(link.target.id)) {
        neighbours.push(link.target.id);
      }
      if (link.target.id === node.id && fixedNodeIds.includes(link.source.id)) {
        neighbours.push(link.source.id);
      }
    }

    const average = arr => arr.reduce((p, c) => p + c, 0) / arr.length;

    const uniqueNeighbours = [...new Set(neighbours)];

    if ("x" in spec) {
      const barycentreX = average(uniqueNeighbours.map(id => layout.get(id).x));
      layout.set(node.id, { x: barycentreX, y: movableLayerPos, width: 1, height: 1 });
    } else {
      const barycentreY = average(uniqueNeighbours.map(id => layout.get(id).y));
      layout.set(node.id, { x: movableLayerPos, y: barycentreY, width: 1, height: 1 });
    }
  }

  return layout;
};


export { layoutBipartite };