// https://observablehq.com/collection/@d3/d3-force
//

import type { SimulationLinkDatum, SimulationNodeDatum } from "d3-force";
import {
  forceCenter,
  forceCollide,
  forceLink,
  forceManyBody,
  forceRadial,
  forceSimulation,
  forceX,
  forceY
} from "d3-force";

import { mean } from "d3-array";

import type { LayoutSpecD3Force } from "../types/topologicalLayoutD3Force";
import type { NetworkLink, NetworkNode } from "../types/network";
import { getVal } from "./getVal";
import { convertVal } from "../complete";
import { Bounds } from "../types/spec";

const layoutD3Force = (nodes: NetworkNode[], links: NetworkLink[], spec: LayoutSpecD3Force, state: any, bounds: Bounds) => {
  const simulation = forceSimulation(nodes as any[]);

  const get = (force, field, defaultVal) => {
    return force[field] ? convertVal(force[field], {}, state, defaultVal, bounds ) :  defaultVal;
  }

  if ("forces" in spec) {
    let i = 0;
    for (const force of spec.forces) {

      let d3force = null;

      if (force.force == "center") {
        const x = get(force, "x", 0);
        const y = get(force, "y", 0);
        const strength = get(force, "strength", 1);
        d3force = forceCenter(x, y).strength(strength);
      } else if (force.force == "collide") {
        const strength = get(force, "strength", 1);
        const iterations = get(force, "iterations", 1);
        const radius = get(force, "radius", 1);
        d3force = forceCollide(radius).strength(strength).iterations(iterations);
      } else if (force.force === "nbody") {
        const theta = get(force, "theta", 0.9);
        const distanceMin = get(force, "distanceMin", 1);
        const distanceMax = get(force, "distanceMax", Infinity);
        const strength = get(force, "strength", -20);
        d3force = forceManyBody().strength(strength).theta(theta).distanceMin(distanceMin).distanceMax(distanceMax);
      } else if (force.force === "x") {
        const strength = get(force, "strength", 0.1);
        const x = get(force, "x", 0);
        d3force = forceX().strength(strength).x(x);
      } else if (force.force === "y") {
        const strength = get(force, "strength", 0.1);
        const y = get(force, "y", 0);
        d3force = forceY().strength(strength).y(y);
      } else if (force.force === "link") {
        const distance = get(force, "distance", 30);
        const iterations = get(force, "iterations", 1);

        d3force = forceLink().links(links as unknown as SimulationLinkDatum<SimulationNodeDatum>[]).distance(distance).iterations(iterations);

        // the default strength is not a constant : it is 1 / Math.min(count(link.source), count(link.target))
        if ("strength" in force) {
          const strength = get(force, "strength", 0);
          d3force.strength(strength);
        }

      } else if (force.force === "radial") {
        const radius = get(force, "radius", 1);
        const strength = get(force, "strength", 0.1);
        const x = get(force, "x", 0);
        const y = get(force, "y", 0);
        d3force = forceRadial(radius).x(x).y(y).strength(strength);
      }

      simulation.force(`force_${i}`, d3force);
      i++;
    }
  } else {
    // No forces were specified, so fall back on a sensible default
    simulation.force("link", forceLink(links as any[]).id((d: any) => d.id))
      .force("charge", forceManyBody().strength(-400));
  }

  // TODO: reconsider termination condition
  let meanV = 1e9;
  do {
    simulation.tick();
    meanV = mean(nodes.map((n: any) => Math.sqrt(Math.pow(n.vx, 2) + Math.pow(n.vy, 2)))) || 0;
    // console.log(mean_v);
  } while (meanV > 1e-3);
  simulation.stop();
};

export { layoutD3Force };
