import { Value } from "./network";
import { LiteralOrFieldOrExpression } from "./topologicalLayoutD3Force";

type MarkerShape =
  "triangle"
  | "circle-triangle"
  | "circle"
  | "tee"
  | "vee"
  | "square"
  | "diamond"
  | "triangle-tee"
  | "triangle-cross";


/**
 * A marker for the start or end of an edge.
 * Can be used to represent directionality, or to distinguish edges of different types.
 */
export type MarkerSpec = {
  /**
   * Offset (in pixels) between the end of the edge and the marker.
   */
  stepBack?: number;

  /**
   * The marker shape.
   */
  shape: MarkerShape | LiteralOrFieldOrExpression;

  /**
   * If the direction is set to "reverse" then the marker will point towards the source node (rather than the target node).
   */
  direction?: "forwards" | "reverse";

  /**
   * Fill color for the marker.
   */
  fill?: string;

  /**
   * Opacity of the marker (0 is fully transparent, and 1 is fully opaque).
   */
  opacity?: number;


  /**
   * Stroke color for the marker.
   */
  stroke?: string;

  /**
   * Area of the bounding-box of the marker.
   */
  size?: number;
}


type LinkPathSpecBase = {
  type: "linkpath";

  /**
   * The name of the field in the link object pointing at the source node.
   */
  start: string;

  /**
   * The name of the field in the link object pointing at the target node.
   */
  end: string;

  /**
   * Definition of marker to be drawn at the start of an edge.
   */
  startMarker?: MarkerSpec;

  /**
   * Definition of marker to be drawn at the end of an edge.
   */
  endMarker?: MarkerSpec;

  /**
   * Tooltip to be displayed on mouseover.
   */
  tooltip?: Value;

  /**
   * For some link types, the shape of the link depends on which node is treated as the source or target.
   * If set, this determines how the two nodes will be interpreted by applying an ordering; this has no effect on the drawing of endMarkers.
   */
  directionForShape?: {
    /**
     * Name of an order to apply.
     */
    ordering?: string;

    /**
     * If true, will reverse the direction of edges from final element in order to first element in order.
     * This is useful, for examples, when drawing hive plots.
     */
    cyclic?: boolean;

    /**
     * If true, reverse the ordering when it is applied.
     */
    reverse?: boolean;
  }

}


type ColorGradient = string | {
  /**
   * Name of a color scheme provided by d3-scale-chromatic (e.g., blues, reds, viridis).
   */
  colorScheme?: string;

  /**
   * If direction is "reverse", the color scale will be applied with the initial color applied to the target node (rather than the source node).
   */
  direction?: "forwards" | "reverse";

  /**
   * If applied to a line, the line is represented by a series of line segments, each with a different color. This option sets the number of line segments used for each edge.
   */
  numSegments?: number;
}


type LinkPathSpecLine = LinkPathSpecBase & {

  /**
   * Represent edges by straight lines.
   */
  shape?: "line";

  /**
   * Specifies a color gradient along the line.
   */
  colorGradient?: ColorGradient;

  /**
   * Stroke color for the line.
   */
  stroke?: string;

  /**
   Dash pattern for the line ([dash length, gap length])
   */
  strokeDash?: [number, number];

  /**
   * Stroke width for the line (in pixels).
   */
  strokeWidth?: number;

  /**
   * Opacity for the curve (0 is fully transparent, and 1 is fully opaque).
   */
  strokeOpacity?: number;
}

type LinkPathSpecWedge = LinkPathSpecBase & {
  /**
   * Represent edges by wedges/triangles.
   */
  shape: "wedge";

  /**
   * Stroke color for the wedges.
   */
  stroke?: string;

  /**
   * Opacity of the stroke around the edge of the wedge (0 is fully transparent, and 1 is fully opaque).
   */
  strokeOpacity?: number;

  /**
   * Fill color for the wedges.
   */
  fill?: string;

  /**
   * Opacity of the wedge (0 is fully transparent, and 1 is fully opaque).
   */
  opacity?: number;


  /**
   * If the direction is set to "reverse" then the wedge will point towards the source node (rather than the target node).
   */
  direction?: "forwards" | "reverse";

  /**
   * Width of the widest past of the wedge.
   */
  width?: number;
}

type LinkPathSpecCurveSymmetric = LinkPathSpecBase & {
  /**
   * Represent edges by symmetric curves.
   */
  shape: "curveSymmetric";

  /**
   * Stroke color for the curve.
   */
  stroke?: string;

  /**
   * Dash pattern for the line ([dash length, gap length])
   */
  strokeDash?: [number, number];

  /**
   * Stroke width for the curve (in pixels).
   */
  strokeWidth?: number;

  /**
   * Opacity for the curve (0 is fully transparent, and 1 is fully opaque).
   */
  strokeOpacity?: number;

    /**
   * If set to true, apply positiveAngle by using positions of two ends only, ignoring which is the source and which is the target.
   * This can be useful if you are drawing an arc diagram and want to draw all arcs on the same side of the line of nodes, regardless of their direction.
   */
  swap?: boolean | "reverse";

}

type LinkPathSpecArc = LinkPathSpecBase & {
  /**
   * Represent edges by semi-circular arcs curves.
   */
  shape: "arc";

  /**
   * Stroke color for the curve.
   */
  stroke?: string;

  /**
   * Dash pattern for the line ([dash length, gap length])
   */
  strokeDash?: [number, number];

  /**
   * Stroke width for the curve (in pixels).
   */
  strokeWidth?: number;

  /**
   * Opacity for the curve (0 is fully transparent, and 1 is fully opaque).
   */
  strokeOpacity?: number;

  /**
   * If positiveAngle is true (the default), then `sweep-flag=1` for the generated path, and the path will be drawn in a "positive-angle" direction.
   * As defined by the SVG Specification:
   * "(i.e., the ellipse formula x=cx+rx*cos(theta) and y=cy+ry*sin(theta) is evaluated such that theta starts at an angle corresponding to the current point and increases positively until the arc reaches (x,y))"
   */
  positiveAngle?: boolean;

  /**
   * If set to true, apply positiveAngle by using positions of two ends only, ignoring which is the source and which is the target.
   * This can be useful if you are drawing an arc diagram and want to draw all arcs on the same side of the line of nodes, regardless of their direction.
   */
  swap?: boolean | "reverse";
}

type LinkPathSpecDiagonal = LinkPathSpecBase & {
  /**
   * Represent edges by sigmoidal curves (the Vega LinkPath transform calls this a "diagonal").
   */
  shape: "sigmoid";

  /**
   * This parameter determines the orientation of the asymptotes of the sigmoid.
   */
  orientation?: "vertical" | "horizontal";

  /**
   * Stroke color for the curve.
   */
  stroke?: string;

  /**
   * Dash pattern for the line ([dash length, gap length])
   */
  strokeDash?: [number, number];

  /**
   * Stroke width for the curve (in pixels).
   */
  strokeWidth?: number;

  /**
   * Opacity for the curve (0 is fully transparent, and 1 is fully opaque).
   */
  strokeOpacity?: number;
}

type LinkPathSpecOrthogonal = LinkPathSpecBase & {
  /**
   * Represent edges by a pair of vertical and horizontal lines.
   */
  shape: "orthogonal";

  /**
   * Which direction to move first.
   */
  orientation?: "vertical" | "horizontal";

  /**
   * Stroke color for the curve.
   */
  stroke?: string;

  /**
   * Dash pattern for the line ([dash length, gap length])
   */
  strokeDash?: [number, number];

  /**
   * Stroke width for the curve (in pixels).
   */
  strokeWidth?: number;

  /**
   * Opacity for the curve (0 is fully transparent, and 1 is fully opaque).
   */
  strokeOpacity?: number;
}

type LinkPathSpecCurve = LinkPathSpecBase & {
  /**
   * Represent edges by asymmetric curves, with curvature distinguishing the source and target nodes.
   */
  shape: "curve";

  /**
   * The angle controls the amount of curvature.
   * The curve is a cubic Bezier curve, with control points positioned at the start and end nodes.
   * The intermediate control point is obtained by taking the line segment from the source node to the midpoint of the source and edge, and rotating from the line joining the source and edge by angle.
   * The angle is in degrees, so a value or 0 or 180 gives a straight line.
   */
  angle?: number;

  /**
   * Stroke color for the curve.
   */
  stroke?: string;

  /**
   * Dash pattern for the line ([dash length, gap length])
   */
  strokeDash?: [number, number];

  /**
   * Stroke width for the curve (in pixels).
   */
  strokeWidth?: number;

  /**
   * Opacity for the curve (0 is fully transparent, and 1 is fully opaque).
   */
  strokeOpacity?: number;
}

export type LinkPathSpecGlyphs = LinkPathSpecBase & {
  /**
   * Represent edges by a series of glyphs positioned along their lengths.
   */
  shape: "glyphs";

  /**
   * The shape of glyph to be used.
   */
  glyphShape: MarkerShape;

  /**
   * Specifies a color gradient along the line.
   */
  colorGradient?: ColorGradient;

  /**
   * Distance between glyphs.
   */
  separation?: number;


  /**
   * Stroke color for outer edge of the glyphs.
   */
  stroke?: string;

  /**
   * Stroke width for the outer edge of the glyphs (in pixels).
   */
  strokeWidth?: number;

  /**
   * Opacity for the outer edge of the glyphs (0 is fully transparent, and 1 is fully opaque).
   */
  strokeOpacity?: number;

  /**
   * Fill color for the line.
   */
  fill?: string;

  /**
   * Opacity of the line (0 is fully transparent, and 1 is fully opaque).
   */
  opacity?: number;

}


type LinkPathSpecSquiggle = LinkPathSpecBase & {
  /**
   * Represent edges by AbySS-Explorer style squiggles.
   * The number of oscillations cna be set directly (as n), or indirectly by setting the spacing between them.
   */
  shape: "squiggle";

  /**
   * The number of squiggles.
   */
  n?: number;

  /**
   * The spacing between squiggles.
   */
  spacing?: number;

  /**
   * The amplitude of the squiggles
   */
  amplitude?: number;


  /**
   * Stroke color for the curve.
   */
  stroke?: string;

  /**
   * Dash pattern for the line ([dash length, gap length])
   */
  strokeDash?: [number, number];

  /**
   * Stroke width for the curve (in pixels).
   */
  strokeWidth?: number;

  /**
   * Opacity for the curve (0 is fully transparent, and 1 is fully opaque).
   */
  strokeOpacity?: number;
}


export type LinkPathMarkSpec =
  LinkPathSpecWedge
  | LinkPathSpecGlyphs
  | LinkPathSpecLine
  | LinkPathSpecCurve
  | LinkPathSpecCurveSymmetric
  | LinkPathSpecArc
  | LinkPathSpecDiagonal
  | LinkPathSpecOrthogonal
  | LinkPathSpecSquiggle;

export type LinkPathSpec = {
  entries: string;
  layout: string;
  mark: LinkPathMarkSpec;

  inducedLayout?: {
    name: string;
    position?: number;

    /**
     * Value of width to assign to entries in the induced layout.
     */
    width?: number;
    /**
     * Value of height to assign to entries in the induced layout.
     */
    height?: number;
    /**
     * Offset to apply to position of element to obtain x- position in induced layout.
     * For example, you may wish to offset by the width of a mark, so that the offset layout coincides with the centre
     * (or a specific edge) of mark rather than the top-left corner.
     */
    x?: LiteralOrFieldOrExpression;

    /**
     * Offset to apply to position of element to obtain x- position in induced layout.
     * For example, you may wish to offset by the width of a mark, so that the offset layout coincides with the centre
     * (or a specific edge) of mark rather than the top-left corner.
     */
    y?: LiteralOrFieldOrExpression;
  }
}