

export type AggregationFunc = "count" | "valid" | "missing" | "distinct" | "sum" | "product" | "mean" | "average" |
  "variance" | "variancep" | "stdev" | "stdevp" | "median" | "q1" | "q3" | "min" | "max" | "argmin" | "argmax" | "mode";


type AggFuncType = (xs: any[]) => (number);

type AggregationSpec = {
  /**
   * Aggregate rows of the data table that have the same values in one (or multiple) columns.
   */
  type: "aggregation";

  /**
   * The data fields to group by.
   */
  groupby?: string[];
  /**
   * The data fields for which to compute aggregate functions.
   */
  fields?: string[];
  /**
   * The aggregation functions to apply.
   */
  ops?: AggregationFunc[],
  /**
   * The field names to use for the computed aggregate fields.
   */
  as?: string,
  //cross: boolean,
  // drop: boolean,
  // key not implemented
}

type BinSpec = {
  /**
   * Group continuous numerical data into discrete bins.
   */
  type: "bin";
  /**
   * The field to be binned.
   */
  field: string;
}
export type CalculateExpressionSpec = {
  calculate: string;
}
export type CalculateSpec = {
  /**
   * Evaluate an expression, and record the resulting value as a field.
   */
  type: "calculate";
  /**
   * The expression to evaluate.
   */
  calculate: string;
  /**
   * The name of the field in which to record the calculated value.
   */
  as: string;
}

type FilterSpec = {
  /**
   * Filter a list of data items, keeping only those for which the specified expression evaluates to true.
   */
  type: "filter";
  /**
   * The expression to evaluate for each data item.
   */
  expr: string;
}

export type FlattenSpec = {
  /**
   * Maps array-valued fields to a set of individual data objects, one per array entry.
   */
  type: "flatten";
  /**
   * Array containing the names of fields to flattten.
   */
  flatten: string[];
}

export type FoldSpec = {
  /**
   * Collapses one or more data fields into a key and value.
   */
  type: "fold";
  /**
   * Array containing the name of fields to fold.
   */
  fields: string[];
}

export type LookupSpec = {
  /**
   * Extend the elements in a (primary) data array by looking up additional values in another (secondary) data array.
   */
  type: "lookup";
  /**
   * The name of the data array to join with.
   */
  from: string;

  /**
   * The field in the secondary array containing the key that will be used for matching.
   */
  key: string;
  /**
   * The fields in the primary data array that will be matched against the key in the secondary array.
   */
  fields: string[];

  /**
   * The fields on the secondary data array hose values should be copied.
   */
  values: string[];
  /**
   * The name of the fields on the primary data array in which to store the looked-up values.
   */
  as?: string[];

  /**
   * The default value to use if lookup fails.
   */
  default: any;
}

export type PivotSpec = {
  /**
   * Maps unique values from a field to new aggregated fields.
   */
  type: "pivot";
  /**
   * The name of the field to pivot on; the unique values of this field become new field names in the output.
   */
  pivot: string;
  /**
   * The name of the field to populate pivoted fields; the aggregate values of this field become the values of the new pivoted fields.
   */
  value: string;
  /**
   * Names of (optional) fields to group by.
   */
  groupby: string[];
  /**
   * The aggregation operation to apply.
   */
  op: string;

  // limit not implemented
}

export type SampleSpec = {
  /**
   * Select a random subset of a data array.
   */
  type: "sample";
  /**
   * The number of data elements to keep.
   */
  sample: number;
}


export type UnrollSpec = {
  /**
   * Unroll, by extracting the values of the specified fields from each element of the data array, and concatenating these into a new array.
   */
  type: "unroll";
  /**
   * The name of the fields whose values should be extracted
   */
  fields: string[];
}

type GeocodingSpec = {
  /**
   * Geocode locations, by finding a latitude and longitude from a place name.
   */
  type: "geocode";

  /**
   * The name of a field containing a location name.
   */
  locationField?: string;
  /**
   * The name of a field containing a street name.
   */
  streetField?: string;
  /**
   * The name of a field containing a city name.
   */
  cityField?: string;
  /**
   * The name of a field containing a state name.
   */
  stateField?: string;
  /**
   * The name of a field containing a country name.
   */
  countryField?: string;
  /**
   * The name of a field containing a postal code.
   */
  postalCodeField?: string;

  /**
   * The name of a field in which the longitude should be saved.
   */
  lonAs?: string;
  /**
   * The name of a field in which the latitude should be saved.
   */
  latAs?: string;

  /**
   * If specified, the displayName of the geocoded location will be recorded as a field of this name.
   * This can be helpful in identifying where the geocoder has picked the wrong location for an ambiguous place name.
   */
  nameAs?: string;
};

type AddIndexSpec = {
  /**
   * Add a sequential id field to each element in a data array.
   */
  type: "addIndex";

  /**
   * The name of the field in which to record the id.
   */
  as?: string;
}

type FunctionDataTransformSpec = {
  /**
   * Transforms a function by calling a function (passed in on object of functions when rendering the spec).
   * This is an escape-hatch to allow users to do things that are not possible within NetworkPanorama itself.
   * The specification fragment for this transform is passed as an argument to the function.
   */
  type: "function";
  /**
   * The name of the function to call.
   */
  function: string;
}

export type TransformSpec = AggregationSpec | AddIndexSpec | BinSpec | CalculateSpec | FilterSpec | FlattenSpec | FoldSpec | PivotSpec | SampleSpec | UnrollSpec | LookupSpec | GeocodingSpec | FunctionDataTransformSpec;

