import { TopolgicalLayoutSpec } from "./topologicalLayout";
import { Bounds } from "./spec";

// TODO: check if this duplicates a type used in mark encoding definition

interface Position {
  x: number;
  y: number;
  width?: number;
  height?: number;
}

type Layout = Map<string | number, Position | Bounds>;


type ScaleAndField = {
  scale: string;
  field: string;
}

export type OrderWithOptionalField = string | {
  order: string;
  field?: string;
}

export type PositionType = "start" | "end" | "middle" | "distribute" | "none";

type LayoutJitterSpec = {
  /**
   * Randomly perturb positions.
   */
  type: "jitter";
  /**
   * x-coordinate may be altered by  -x/2 <= dx <= x/2
   */
  x?: number;
  /**
   * y-coordinate may be altered by  -y/2 <= dy <= y/2
   */
  y?: number;
}

type LayoutTransformSet = {
  /**
   * Set x/y/width/height of elements in a layout.
   */
  type: "set";
  /**
   * Which field to set.
   */
  field: "width" | "height" | "x" | "y";
  /**
   * New value to assign to field.
   */
  value: number;
}

type LayoutTransformOffset = {
  /**
   * Offset positions of elements (either horizontally or vertically) to avoid overlap.
   * Minimises the sum-of-squares of displacements.
   */
  type: "offset";
  /**
   * Which direction to offset positions in.
   */
  direction: "vertical" | "horizontal";
  /**
   * Maximise width (or height) within which offset elements should fit.
   */
  width?: number;
  /**
   * Minimum spacing between adjacent elements.
   */
  minSpacing?: number;
  /**
   * If set, elements for which this expression evaluates to false will be ignored.
   */
  condition?: string;
}

type LayoutTransformOrientLinks = {
  /**
   * Move target of links to align horizontally with source.
   */
  type: "orientLinks";
  /**
   * Direction in which to orient links.
   */
  direction: "vertical" | "horizontal";
  /**
   * The network containing the links.
   */
  network: string;
};

type LayoutTransformBeeswarm = {
  /**
   * Apply a Beeswarm transformation, by doing a force-based simulation in which
   * elements are attracted to their original positions, but also pushed apart by a collision force.
   */
  type: "beeswarm";
  /**
   *  Strength of horizontal force pulling nodes to their original position (default: 0.1).
   */
  gravityX?: number;
  /**
   *  Strength of vertical force pulling nodes to their original position (default: 0.1).
   */
  gravityY?: number;
  /**
   * Strength of force avoiding overlaps (default: 1)
   */
  collisionStrength?: number;
}

type LayoutTransformFunction = {
  /**
   * Transform a layout by calling a function (passed in on object of functions when rendering the spec).
   * This is an escape-hatch to allow users to do things that are not possible within NetworkPanorama itself.
   */
  type: "function";
  /**
   * The name of the function to call.
   */
  function: string;
}


type LayoutTransform = LayoutTransformSet | LayoutTransformOffset | LayoutTransformOrientLinks | LayoutJitterSpec | LayoutTransformBeeswarm | LayoutTransformFunction;


type BaseLayoutData = {
  /**
   * The name of the data array containing the items to be laid out.
   */
  data: any;
  transform?: LayoutTransform[];
}

type BaseLayoutNetwork = {
  /**
   * The name of the network whose node should be laid out.
   */
  network: any;
  transform?: LayoutTransform[];
}


type BaseLayout = (BaseLayoutData | BaseLayoutNetwork) & {
  /**
   * The name that will be given to the layout when it is created.
   */
  name: string;

  // TODO: explain
  positioning?: PositionType;

  // TODO: width and height can't always be passed - e.g., they would mess up a grid layout

  /**
   * width of individual elements in the layout: field, expression, or number
   */
  width?: Val;
  height?: Val;

  /**
   * List of transforms to be apply to this layout.
   */
  transform?: LayoutTransform[];

}

type LayoutLinearOrder = BaseLayout & {
  /**
   * A Linear layout positions items along a straight line. It is a 1D layout that depends on a single order or scale.
   */
  pattern: "linear",
  order: OrderWithOptionalField;
  orientation?: "vertical" | "horizontal";
}

type LayoutLinearScale = BaseLayout & {
  /**
   * A Linear layout positions items along a straight line. It is a 1D layout that depends on a single order or scale.
   */
  pattern: "linear",
  scale: ScaleAndField;
  orientation?: "vertical" | "horizontal";

}

type LayoutLinear = LayoutLinearOrder | LayoutLinearScale;

// Grid layouts
// TODO: rename to pattern
type Val = number | { expression: string } | { field: string } | {scale: string, field: string};


type GridLayout = BaseLayout & {
  /**
   * There are several grid layouts that positions items in a 2D grid. They depend on a single order or scale.
   */

  pattern: GridType;
  numRows: number;
  numCols: number;
  order: OrderWithOptionalField;

  /**
   * Padding between columns in the grid (relative to the size of one element in the grid).
   */
  paddingX?: number;
  /**
 * Padding between rows in the grid (relative to the size of one element in the grid).
 */
  paddingY?: number;
}


// type ScaleOrOrderSpec = {scale: ScaleAndField} | {order: OrderWithOptionalField}
type DimensionSpec = { scale: string, field: string } | { order: string } | { value: number};
type DistributionPattern = "start" | "end" | "center" | "distribute";

type AttributeLayoutCartesian = BaseLayout & {
  pattern: "cartesian";

  /**
   * (default: x-coordinate of 0)
   */
  x?: DimensionSpec[];
  xDistributed?: DistributionPattern;
  /**
   * (default: y-coordinate of 0)
   */
  y?: DimensionSpec[];
  yDistributed?: DistributionPattern;
}

type AttributeLayoutPolar = BaseLayout & {
  pattern: "polar";
  radius: DimensionSpec[];
  radiusDistributed?: DistributionPattern;
  angle: DimensionSpec[];
  angleDistributed?: DistributionPattern;
}

type AttributeLayoutExplicit = BaseLayout & {
  pattern: "explicit";
  x: {field: string, scale?: string} | {expression: string} | {parameter: string};
  y: {field: string, scale?: string} | {expression: string} | {parameter: string};
  scaleToBounds?: boolean;
};

type AttributeLayoutSpec =
  AttributeLayoutCartesian
  | AttributeLayoutPolar
  | LayoutLinear
  | GridLayout
  | RoundPatternSpec
  | AttributeLayoutExplicit;

type DerivedLayout = {
  /**
   * The name of the new layout that will be created.
   */
  name: string;

  /**
   * The name of the layout that this layout should be created from.
   */
  source: string;

  /**
   * The transforms to be applied to create this layout from the source layout.
   */
  transform?: LayoutTransform[];
};

type FieldOrExpression = {field: string} | {expression: string};

type MapLayoutSpec = (BaseLayoutData | BaseLayoutNetwork) & {
  /**
   * Name of this layout.
   */
  name: string;
  /**
   * Name of the map on which the layout is based.
   */
  map: string;
  /**
   * Name of field (or expression) containing the longitude.
   */
  lng: FieldOrExpression;
  /**
   * Name of field (or expression) containing the latitude.
   */
  lat: FieldOrExpression;
}


type LayoutFunctionSpec = {
  /**
   * Computes a layout by calling a function (passed in on object of functions when rendering the spec).
   * This is an escape-hatch to allow users to do things that are not possible within NetworkPanorama itself.
   */
  type: "function";

  /**
   * Name of this layout.
   */
  name: string;

  /**
   * The name of the function to call.
   */
  function: string;
}


type LayoutSpec = TopolgicalLayoutSpec | AttributeLayoutSpec | MapLayoutSpec | DerivedLayout | LayoutFunctionSpec;

export type GridType = "row-by-row" | "col-by-col" | "row-snake" | "col-snake" | "diagonal" | "diagonal-snake" |
  "z-curve" | "hilbert";


export type RoundPatternSpec = SpiralPatternSpec | SectorsPatternSpec | CirclePatternSpec;

// TODO: implement
export type SpiralPatternSpec = BaseLayout & {
  /**
   * A spiral layout positions items in a spiral. It is a 1D layout that depends on a single order or scale.
   */
  pattern: "spiral";

  /**
   * The specific shape of spiral to be used (default: "Archimedes")
   */
  shape?: "Archimedes" | "hyperbolic" | "logarithmic" | "lituus";

  /** The angular separation (in degrees) between successive items on the spiral
   */
  angularSeparation?: number;

//  maxAngle?: number;

  /**
   * The hyperbolic spiral and lituus spiral inwards; they are initially quite flat, and become increasingly curved, so it is useful
   * to start at an angle larger than 0; increasing the angle moves the outer edge of the spiral inwards.
   *
   * For the Archimedes and logarithmic spiral, increasing the angle moves the start of the spiral outwards.
   * (the default is 2 * pi).
   */
  startAngle?: number;
}

// TODO: only works with an order, not a scale
export type SectorsPatternSpec = BaseLayout & OrderWithOptionalField & {
  /**
   *
   * Positions items in concentric rings, with a user-specified number of items in each ring.
   * This layout is similar to a spiral, but rather than radial distance increasingly continuously, it increases in discrete steps.
   * It depends on a single layout (if you want angular position to encode a continuous variable, use the polar layout instead).
   */
  pattern: "rings";
  // direction?: "clockwise" | "counterclockwise";

  /**
   * The number of elements to position within each ring.
   * The first element should typically  be 0 (to have no element in the center) or 1 (to have the first element in the center).
   * If the array is shorter than the number of rings required, the last element of the array will determine the number of sectors in these additional rings.
   */
  numSegments: number[]; // if array, allows different number of sectors per annulus
 // increaseSize?: boolean; // either increase size or separation; default to increasing separation
//  order: OrderWithOptionalField;
}

export type CirclePatternSpec = BaseLayout & DimensionSpec & {
  /**
   * A circle layout positions items in a circle. It is a 1D layout that depends on a single order or scale.
   */

  pattern: "circle";
}

type CartesianAxis = {
  type: "cartesian";
  x: number;
  x2: number;
  y: number;
  y2: number;
  extend: "x" | "y" | "radial";// the direction along which the scale varies
  ticks: {x: number, y: number, label: string}[],
}

type PolarAxis = {
  type: "polar";
  x: number;
  y: number;
  r: number;
  ticks: {x: number, y: number, label: string}[]
}

type Axis = CartesianAxis | PolarAxis;

