import { NetworkTransformSpec, SeparateMetricSpec } from "./networkTransforms";

interface NetworkNode {
  id: number; // or string?

  [id: string]: any;
}


// TODO: record whether directed here
interface NetworkLink {
  id?: string;
  source: NetworkNode;
  target: NetworkNode;
  data?: any;
  directed?: boolean; // a network may contain a mixture of directed and undirected edges: ensure handled correctly during addingReverse edges and when calculating metrics
}


interface Network {
  nodes: NetworkNode[];
  links: NetworkLink[];
}

interface Range {
  min: number;
  max: number;
}


export type SimpleNetworkSpec = {
  /**
   * The name to be given to the network when it is created.
   */
  name?: string;
  /**
   * The name of the data array containing the list of nodes (or a selector)
   */
  nodes: string;
  /**
   * The name of the data array containing the list of links (or a selector)
   */
  links: string;
  /**
   * Boolean variable recording whether the edge sin this network are directed.
   */
  directed: boolean;

  /**
   * The first element is the field on the nodes containing the node id;
   * the second element is the field on the links containing the id of the source node.
   */
  source_node: [string, string];

  /**
   * The first element is the field on the nodes containing the node id;
   * the second element is the field on the links containing the id of the target node.
   */
  target_node: [string, string];

  /**
   * List of metrics to calculate for each node.
   */
  metrics?: SeparateMetricSpec[];
  /**
   * If true, replace any undirected links by a pair of directed links.
   */
  addReverseLinks?: boolean;

  /**
   * List of network transformations to apply.
   */
  transform?: NetworkTransformSpec[];

  /**
   * If true, links for which the source or target node cannot be found are ignored.
   * If false, these links are created, but their source or target attribute is undefined.
   * This is likely to cause errors computing layouts, but may be useful when looking for misisng node definitions in a dataset.
   * Default: true
   */
  ignoreDanglingLinks: boolean;
}


// Types for complex network import:

export type ComplexNetworkSpec = {
  /**
   * The name to be given to the network when it is created.
   */
  name?: string;
  /**
   * Each entry in this list describes the data that should be extracted from a data array.
   */
  parts: ComplexNetworkSpecComponent[];
  /**
   * List of network transformations to apply.
   */
  transform?: NetworkTransformSpec[];
}

export type ComplexNetworkSpecComponent = {
  /**
   * The name of the array being processed
   */
  data: string;

  /**
   * Definitions for how nodes should be extracted from this data array.
   */
  yieldsNodes?: ExtractedNodeSpec[];
  /**
   * Definitions for how links should be extracted from this data array.
   */
  yieldsLinks?: ExtractedLinkSpec[];
  /**
   * Definitions for how nested links should be extracted from this data array.
   */
  mapYieldsLinks?: MapLinkSpec[];
}

export type ExtractedNodeSpec = {
  /**
   * The node type of nodes created by this rule.
   */
  type?: string;
  /**
   * The name of the field which contains the node id.
   */
  id_field: string;
  /**
   * List of fields from the data array to be copied to the nodes, and their new names.
   */
  data: Value[];
}

type MapLinkSpec = ExtractedLinkSpec & {
  field: string;
}

// TODO: should source_id/target_id be changed to ValFromField or expression?

export type ExtractedLinkSpec = {
  /**
   * The field in the data array that contains the id of the source node.
   */
  source_id: string | ValFromField | ValLiteral; // the id of the source

  /**
   * The field on the source node which contains the value that will be used as the id (default is `id`).
   */
  source_id_field?: string; // which field on the source node

  /**
   * The type of the source node.
   */
  source_node_type: string; //

  /**
   * The field in the data array that contains the id of the target node.
   */
  target_id: string | ValFromField | ValLiteral;

  /**
   * The field on the target node which contains the value that will be used as the id (default is `id`).
   */
  target_id_field?: string;
  /**
   * The type of the source node.
   */
  target_node_type?: string;

  /**
   * List of fields from the data array to be copied to the links, and their new names.
   */
  data?: Value[];

  /**
   * If true, replace any undirected links by a pair of directed links.
   */
  addReverseLinks?: boolean;
}

type Value = string | ValFromField | ValLiteral;
export type ValFromField = {
  field: string;
  as?: string;
}
export type ValLiteral = {
  value: any;
  as?: string;
}

// Types for network import

type NetworkImportBase = {
  /**
   * The name to be given to the network when it is created.
   */
  name?: string;

  /**
   * URL from which the network data should be downloaded.
   */
  url: string;

  transform?: NetworkTransformSpec[];
}

type NetworkImportPajek = NetworkImportBase & {
  /**
   * Import a file in Pajek format.
   */
  format: "pajek";

  /**
   * If true, replace any undirected links by a pair of directed links.
   */
  addReverseLinks?: boolean;
}

type NetworkImportGraphml = NetworkImportBase & {
  /**
   * Import a file in Pajek format.
   */
  format: "graphml";

  /**
   * If true, replace any undirected links by a pair of directed links.
   */
  addReverseLinks?: boolean;

  /***
   * A GraphML file may contain multiple graphs: this option specifies the index
   * of the graph to import (default: 0)
   */
  graphIndex?: number;

  /***
   * A GraphML file may contain multiple graphs: this option specifies the id
   * of the graph to import
   */
  graphId?: string;
}

type NetworkSpecDerived = {
  /**
   * The name to be given to the network when it is created.
   */
  name?: string;

  /**
   * The name of an existing network, from which this new network should be derived by applying transformations.
   */
  source: string;

  /**
   * List of metrics to calculate for each node.
   */
  metrics?: SeparateMetricSpec[];

  /**
   * List of network transformations to apply.
   */
  transform?: NetworkTransformSpec[];
}

type NetworkImport = NetworkImportPajek | NetworkImportGraphml;


export type NetworkSpec = SimpleNetworkSpec | ComplexNetworkSpec | NetworkImport | NetworkSpecDerived;