import type { AggregationFunc } from "./data-transforms";


type MetricName = "degree" | "in-degree" | "out-degree" | "undirected-betweeness" | "directed-betweeness" |
  "closeness" | "eccentricity";

// metrics can also be listed directly in network definition (outside a transform block)
interface SeparateMetricSpec {
  /**
   * Which metric to calculate.
   */
  metric: MetricName;

  /**
   * The name of the field in which the computed values should be saved (default is the same as the name of the metric).
   */
  as?: string;
}

export interface ClusterSpec {
  /**
   * Perform clustering/community detection.
   */
  type: "cluster";

  /**
   * The clustering method to use.
   */
  method?: "louvain";

  /**
   * The name of the field in which the cluster id of each node should be stored (defaults to the same as the method).
   */
  as?: string;
}

type AggregateEdgesSpec = {
  /**
   * Aggregate multiple edges with the same source and target node into a single edge.
   */
  type: "aggregateEdges";


  /**
   * Aggregation function to be applied to the edges that are being merged, to obtain the data for the new aggregate edge.
   */
  operation?: AggregationFunc;

  /**
   * If directed is false, two edges between the same pair of nodes will be merged even if they have opposite directions.
   */
  directed?: boolean;
}

type ConnectivityBasedAttributesSpec = {
  /**
   * Computes a new data attribute for each node, based on the data associated with its neighbours.
   */
  type: "connectivityBasedAttribute";

  /**
   * Which data attribute to select for each neighbouring node.
   */
  attributeSelector: string;

  /**
   * The name of the field in which the result should be stored.
   */
  as: string;

  /**
   * The aggregation operation to be applied.
   */
  operation?: AggregationFunc;

  /**
   * When calculating value for a node, consider edges directed in towards that node, out away from that node, or both.
   * Default is both.
   */
  direction?: "in" | "out" | "both";
}

type ProjectNetworkSpec = {
  /**
   * Projection replaces a path between two nodes via a third node with a single link directly joining them (i.e., A->B->C become A->C)
   */
  type: "project";

  /**
   * Field that indicates the node type (default: "_type").
   */
  fieldName?: string;

  /**
   * The value of the node type field for the nodes that will be joined by the newly-created link (i.e., nodes A and C in the example A->B->C).
   */
  remainingNodeType: string | number;

  /**
   * The value of the node type filed for the node that will be replace (i.e., node B in the example A->B->C)
   */
  removedNodeType: string | number;

  /**
   If directed is false, ignore edge directions when considering paths to project (i.e, A->B->C and A<-B->C would become A->C).
   */
  directed: boolean;


  /**
   * How to aggregate the data from the links that are being merged.
   */
  linkDataReducer: any; //
  // really this is (a: any, b: any) => [any, any], but JSON Schema generation fails for function signatures... // TODO: need some way to actually specify in JSON
}

type MetricSpec = {
  /**
   * Compute a metric for each node. These may depend on the structure of the network as a whole.
   */
  type: "metric";

  /**
   * Which metric to calculate.
   */
  metric: MetricName;

  /**
   * The name of the field in which the value of the metric calculated for each node should be stored (defaults to the same as the metric).
   */
  as?: string;
}

// todo: rename this as can apply to links
type NodeCalculateSpec = {
  /**
   * Calculate new attribute for each node or link, based on the current attributes of only the same node or link.
   */
  type: "calculate";

  /**
   * Whether to do calculation for each node or edge (default is nodes).
   */
  for?: "nodes" | "links";

  /**
   * The expression to calculate.
   */
  calculate: string;

  /**
   * The name of the field in which result should be stored (defaults to the same as the method).
   */
  as: string;
}

type RemoveIsolatedSpec = {
  /**
   * Remove isolate nodes (i.e., nodes with no attached edges)
   */
  type: "removeIsolated";
}

type FilterEdgesSpec = {
  /**
   * Remove some edges, retaining only those for which the expression is true.
   */
  type: "filterEdges";
  /**
   * The expression to evaluate for each edge.
   */
  expression: string;
}

type FilterNodeSpec = {
  /**
   * Remove some nodes, retaining only those for which the expression is true. When a node is removed, so are any edges that were attached to it.
   */
  type: "filterNodes";

  /**
   * The expression to evaluate for each node.
   */
  expression: string;
}

//

type AggregateNodeSpecBase = {
  /**
   * Aggregate nodes into a single node.
   */
  type: "aggregateNodes";
}

type AggregateNodesByField = AggregateNodeSpecBase & {
  /**
   * Nodes with the same value for this field will be merged.
   */
  field: string;
}

type AggregateNodesByFields = AggregateNodeSpecBase & {
  /**
   * Nodes with the same value for this field will be merged.
   */
  fields: string[];
  /**
   * The values of this field wil be recorded in this field.
   */
  as?: string;
}

type AggregateNodesByFieldOrFields = AggregateNodesByField | AggregateNodesByFields;

type AggregateNodesByExpression = AggregateNodeSpecBase & {
  /**
   * Nodes for which evaluating this expression gives the same result will be merged.
   */
  expression: string;

  /**
   * The merged nodes will contain the result of evaluating the expression as a field with this name (default: key);
   */
  as?: string;
}

type ConnectAttributeSpec = {
  /***
   * Add a connection between each pair of nodes for which the value of a specified field in the same.
   */
  type: "connect";
  /**
   * Value to be stored in the "_type" field for edges created by this operation.
   */
  as?: string;
  /**
   * The name of the field to be examined.
   */
  field: string;
  /**
   * Each pair of nodes will be joined by a single edge (if `false`), or a pair of edges with opposite directions (if `true`). Default is `true`.
   */
  bothDirections?: boolean;
}

type ConnectExpressionSpec = {
  /***
   * Add a connection between each pair of nodes for which the value of a specified expression in the same.
   */
  type: "connect";
  /**
   * Value to be stored in the "_type" field for edges created by this operation.
   */
  as?: string;
  /**
   * The expression to be evaluated.
   */
  expression: string;
  /**
   * Each pair of nodes will be joined by a single edge (if `false`), or a pair of edges with opposite directions (if `true`). Default is `true`.
   */
  bothDirections?: boolean;
}

type ConnectSpec = ConnectAttributeSpec | ConnectExpressionSpec;


type SwapNodesAndLinksSpec = {
  /**
   * Swap nodes and links by computing the Line Graph.
   * This consists of a node for each edge in the original graph;
   * these are joined if they share a common edge in the original graph (in the undirected case)
   * or if they correspond to edges that form a directed path of length two in the original graph.
   */
  type: "swapNodesAndEdges";
  /**
   * Whether to treat edges as directed
   */
  directed?: boolean;
}

type ReverseEdgesSpec = {
  /**
   * Swap the direction of edges.
   */
  type: "reverseEdges";
  /**
   * If an expression is provided as a condition, then only edges for which it evaluates to `true` will be reversed.
   */
  condition?: string;
}

type SetDirectednessValueSpec = {
  /**
   * Change whether or not edges are directed.
   */
  type: "setDirectness";
  /**
   * The value of directedness to apply to all edges.
   */
  value: "directed" | "undirected";
}

type SetDirectednessExpressionSpec = {
    /**
   * Change whether or not edges are directed.
   */

  type: "setDirectness";
  /**
   * This expression will be evaluated for all edges; if it evaluates to `true`, then the edge will be treated as directed.
   */
  expression: string;
}
type SetDirectednessSpec = SetDirectednessValueSpec | SetDirectednessExpressionSpec;


type PromoteFieldSpec = {
  /**
   * Promote a field to a node type.
   * This creates a node for each distinct value of a field on existing nodes, and adds an edge from each original node to the corresponding new value node.
   */
  type: "promote";
  /**
   * The name of the field to be promoted.
   */
  field: string;
  /**
   * Prefix to be applied before the value when assigning the ID for the new nodes.
   */
  prefix: string;
  /**
   * Prefix to be applied to the ID of newly created edges.
   */
  linkPrefix?: string;

  /**
   * Whether the new edges should be directed from the existing nodes to the new value nodes or vice-versa.
   */
  direction?: "toValue" | "fromValue";
};

type PromoteExpressionSpec = {
  /**
   * Promote an expression to a node type.
   * This creates a node for each distinct value of a expression evaluated for existing nodes, and adds an edge from each original node to the corresponding new value node.
   */
  type: "promote";
  /**
   * The expression to be evaluated for each node.
   */
  expression: string;

  /**
   * Prefix to be applied before the value when assigning the ID for the new nodes.
   */
  prefix: string;
  /**
   * Prefix to be applied to the ID of newly created edges.
   */
  linkPrefix?: string;

  /**
   * Whether the new edges should be directed from the existing nodes to the new value nodes or vice-versa.
   */
  direction?: "toValue" | "fromValue";
};

type PromoteSpec = PromoteFieldSpec | PromoteExpressionSpec;



type AggregateNodeSpec = AggregateNodesByFieldOrFields | AggregateNodesByExpression;

type NetworkTransformSpec = ClusterSpec | AggregateEdgesSpec | ConnectivityBasedAttributesSpec | ProjectNetworkSpec |
  MetricSpec | NodeCalculateSpec | RemoveIsolatedSpec | FilterEdgesSpec | FilterNodeSpec | AggregateNodeSpec | ConnectSpec | SwapNodesAndLinksSpec | ReverseEdgesSpec | SetDirectednessSpec | PromoteSpec;
