// TODO: replace with the transforms implemented in networks/paths
// TODO: what if all selections are mixed, and have both a nodes and edges array (i.e., they are like networks, but not neccessarily closed)? Operations amy act on both or nether

// A selection contains a list of node ids, and a list of link ids; one or both may be empty.
import { NetworkLink, NetworkNode } from "./network";
import { ModifierKeys } from "./interaction";

// TODO: actually generate selection objects in this form, rather than as lists of ids
type Selection = {
  nodes: NetworkNode[]; //(string | number)[];
  links: NetworkLink[]; // (string | number)[];
  network: string;
};


// Transforms applied to selections
type BaseSelectionTransform = {
  type: string; // or "operation?" - the name of the transform to apply

  /**
   * If action is set to "add", the result of the transformation will include the contents of the original selection.
   * (N.B. the cytoscape API refers to this as "open" or "closed")
   */
  action: "add" | "replace";
};

type TransformShortestPath = BaseSelectionTransform & {
  /**
   * Find the shortest path between a selection containing a pair of nodes (if the selection contains more than 2 nodes, the path is found between the first 2 nodes, and the remained are ignored).
   * If selection contains more than 2 nodes, it will find the shortest path between all pairs of nodes.
   */
  type: "shortestPath";

  /**
   * Field name containing the edge weight (this could be pre-computed with a calculate transform).
   * As we use Dijiska's algorithm, all weights must be positive.
   */
  weight?: string;
}

type TransformPrune = BaseSelectionTransform & {
  /**
   * Remove nodes that have no incoming connections, no outgoing connections, or no connections at all.
   */
  type: "prune";

  /**
   * Specifies whether to remove nodes with no incoming connections ("roots"), nodes with no outgoing connections ("leaves")
   * or no connections at all ("isolated").
   */
  remove: "roots" | "leaves" | "isolated";
};

type TransformFilter = BaseSelectionTransform & {
  /**
   * Filter (either nodes or edges) retaining only those for which the expression is true.
   */
  type: "filter";

  /**
   * Whether the filter expression applies to nodes or edges.
   */
  affects: "nodes" | "edges";

  /**
   * The expression to evaluate for each edge.
   */
  expression: string;
};

type TransformProject = BaseSelectionTransform & {

  /**
   * Given a selection, make a new selection containing elements for which the specified fields hae the same values.
   */
  type: "project";

    /**
   * Whether the operation applies to nodes or edges.
   */
  affects: "nodes" | "edges";

  /**
   * List of fields to match on.
   */
  fields: string[];

  /**
   * If selection contains "red cars" and "blue bikes", should we also select "red bikes" and "blue cars"?
   */
  semantics?: "separate" | "combined";

}

type TransformParallelEdges = BaseSelectionTransform & {
  /**
   *  Select edges with the same source and target nodes as edges in a selection.
   */
  type: "parallelEdges";
  ignoreDirection?: boolean;
}

type TransformCommonNeighbours = BaseSelectionTransform & {
  /**
   * Select the nodes connected to all the nodes in a selection.
   */
  type: "commonNeighbours";
  direction: "source" | "target" | "both";
}

type TransformConnectedNodes = BaseSelectionTransform & {
  /**
   * Select the nodes connected to a set of edges.
   */
  type: "connectedNodes";
  direction: "source" | "target" | "both";
}

type TransformConnectedEdges = BaseSelectionTransform & {
  /**
   * Select the edges connected to a set of nodes.
   */
  type: "connectedEdges";
  direction: "source" | "target" | "both";
}
// ?? just Edges?

type TransformEdgesBetween = BaseSelectionTransform & {
  /**
   * Find the edges joining nodes in one selection with nodes in a second selection.
   */
  type: "edgesBetween";

  secondSelection: string;

  ignoreDirection?: boolean;
}

type TransformNeighbours = BaseSelectionTransform & {
  /**
   * Find the neighbours of a node (or set of nodes)
   */
  type: "neighbours";
  direction: "in" | "out" | "both";

  /**
   * default 1
   */
  depth?: number | "infinity";
  // way to select nodes or edges?
};

type SelectionTransform =
  TransformShortestPath
  | TransformPrune
  | TransformFilter
  | TransformProject
  | TransformParallelEdges
  | TransformConnectedEdges
  | TransformConnectedNodes
  | TransformCommonNeighbours
  | TransformEdgesBetween
  | TransformNeighbours;


// Definitions of params

type SelectionParameter = {
  /**
   * Name of the selection parameter.
   */
  name: string;

  /**
   * Initial value of selection.
   */
  value?: string[];

  type: "selection";

  /**
   * Name of the corresponding network.
   */
  network: string;

  /**
   * If set, then dragging whilst holding down these keys will move the selected nodes.
   */
  dragging?: {
    /**
     * Key that must be held down whilst the interaction occurs (default is "any").
     */
    modifierKeyRequired?: ModifierKeys[];

    /**
     * Key that must be *not* held down whilst the interaction occurs (default is none).
     */
    modifierKeyAbsent?: ModifierKeys[];
  }

    /**
     * If true, then clicking on the background will clear the current selection.
     */
    clearOnBackgroundClick?: boolean;
}

type DerivedSelectionParameter = {
  /**
   * Name of the derived selection.
   */
  name: string;

  /**
   * Name of the selection from this selection is derived.
   */
  source: string;

  /**
   * List of selections that will be applied to the source selection to obtain this selection.
   */
  transform: SelectionTransform[];
}

type VariableParameter = {

  /**
   * The name of the parameter.
   */
  name: string;

  /**
   * The default vale of the parameter.
   */
  value?: any;

  /**
   * Binds the parameter to an input element.
   */
  bind?: Binding;

  /**
   * Specification for how the parameter should be animated.
   */
  animate?: Animation;
}

type Animation = {

  /**
   * Label displayed by animation controls (by default, the name of the parameter will be used)
  **/
  label?: string;

  /**
   * Minimum values for the animation
   */
  min: number;

  /**
   * Maximum value for the animation.
   **/
  max: number;

  /**
   * Array of discrete values to transition between.
   */
  values: (string | number)[];

  /**
   * Time, in seconds, between updates to the parameter's value. Incompatible with totalTime option.
   */
  stepTime?: number;
  /**
   * Time taken for a complete cycle of the animation. Incompatible with stepTime option.
   */
  totalTime?: number;

  /**
   * Value by which the parameter should be increased on each update. Incompatible with the numSteps option.
   */
  stepSize?: number;
  /**
   * Total number of updates in one cycle of the animation. Incompatible with the stepSize option.
   */
  numSteps?: number;

  /**
   * If true, the animation will start automatically; if false, animation will not start until the pay button is pressed.
   */
  autoStart?: boolean;

  /**
   * If true, the animation will loop: after reaching the max value, the parameter wil reset to the min value and begin increasing again.
   */
  loop?: boolean;

  /**
   * If true, display a play/pause button.
   */
  showPauseControl?: boolean;

  /**
   * If true, display a button that allows users to step the next/previous state of the animation.
   */
  showStepControl?:boolean;

  /**
   * If true, display a button allowing users to jump to the first/last state of the animation.
   */
  showJumpControl?: boolean;

  /**
   * If true, display the current value of the parameter.
   */
  showValue?: boolean;
}


export type Parameter = VariableParameter | SelectionParameter | DerivedSelectionParameter;


// N.B. Currently we only support binding to input elements, not binding point selections to legends or interval selections to scales.
// N.B. Vega calls the `label' the `name'

// Bindings

type BindingBase = {
  /**
   * The type of input element to use.
   */
  input: "checkbox" | "radio" | "range" | "select" | "brush" | undefined;

  /**
   * An optional CSS selector string indicating the parent element to which the input element should be added.
   */
  element?: string;

  /**
   * Label to use for the input element (by default, the name of the bound signal will be used).
   */
  label?: string;

  /**
   * Values for attributes to set on the HTML element.
   */
  [attribute: string]: string;

}


type BrushBinding = BindingBase & {
  /**
   * Filter by brushing on an axis. In order to determine the scale for this axis, it is necessary to specify
   * which network is to be filtered, whether nodes or links are to be filtered, and which field wll be filtered on.
   */
  input: "brush";

  axisType: "linear" | "time";

  /**
   * Name of the network to be filtered.
   */
  network: string;

  /**
   * Whether this control will be used to filter nodes or edges.
   */
  for: "nodes"| "links";

  /**
   * The field on the node or link to be filtered.
   */
  field: string;

  /**
   * If set to true, parameter will be updated continuously as the brush is moved, without waiting for interaction to end.
   */
  updateContinuously?: boolean
}


type CheckboxBinding = BindingBase & {

  input: "checkbox";

}

type RadioOrSelectBinding = BindingBase & {

  input: "radio" | "select";

  /**
   * Array of the available options.
   */
  options: any[];

  /**
   * Array containing labels for the options. By default, the label is the obtained by converting the option value to a string.
   */
  labels?: string[];
}

type RangeBinding = BindingBase & {

  input: "range";

  /**
   * Maximum slider value. Defaults to 100 (or the default signal value, if that is larger).
   */
  max?: number;

  /**
   * The minimum slider value. Defaults to 0 (or the default signal value, if that is smaller).
   */
  min?: number;

  /**
   * The minimum slider increment.
   */
  step?: number;

}

type ExternalBinding = BindingBase & {

  input: undefined;

  /**
   * CSS selector for an input element.
   */
  element: string;

  /**
   * The name of the event to listen for (default is oninput).
   */
  event?: string;
}

export type Binding = BrushBinding | CheckboxBinding | RadioOrSelectBinding | RangeBinding | ExternalBinding;
