import type { UrlData } from "vega-typings";
import type { Scale } from "vega-typings/types/spec/scale";

import type { LayoutSpec } from "./layout";
import type { OrderingSpec } from "./ordering";
import type { TableSpec } from "./table";
import type { TransformSpec } from "./data-transforms";
import type { BaseData, SourceData, ValuesData } from "vega-typings/types/spec/data";
import type { NetworkSpec } from "./network";
import type { ModifierDefinitions } from "./interaction";

// TODO: replace these approximate typings with correct types
//import { Mark } from "vega-typings";
import type { MarkDef as Mark } from "vega-lite/src/mark";
import type { LinkPathMarkSpec } from "./LinkMarks";
import type { Interaction } from "./interaction";
import type { Parameter } from "./params";
import { Format } from "vega-typings/types/spec/data";
import { LiteralOrFieldOrExpression } from "./topologicalLayoutD3Force";


type Bounds = {
  x: number;
  y: number;
  width: number;
  height: number;
}

// Data is a union type, and applying omit to it rather than the individual
// types seems to cause problems
type DataWithoutTransform =
  Omit<SourceData, "transform"> |
  Omit<ValuesData, "transform"> |
  Omit<UrlData, "transform"> |
  Omit<BaseData, "transform"> |
  { localStorage: string, name: string, format?: Format } |
  { name: string, function: string };


// TODO: update this type, as well also allow circle/square/diamond marks (rather than symbol marks), and expressions for fields
type GraphgraMark = LinkPathMarkSpec | Mark;

// TODO: think about this
type VisEntryBase = {
  /**
   * Name for this vis entry.
   */
  name?: string;

  /**
   * Mark to be created for each data element.
   */
  mark: GraphgraMark;

  /**
   * Save the location of each positioned element as a
   */
  inducedLayout?: {
    name: string

    /**
     * Value of width to assign to entries in the induced layout.
     */
    width?: number;
    /**
     * Value of height to assign to entries in the induced layout.
     */
    height?: number;

    /**
     * Offset to apply to position of element to obtain x- position in induced layout.
     * For example, you may wish to offset by the width of a mark, so that the offset layout coincides with the centre
     * (or a specific edge) of mark rather than the top-left corner.
     */
    x?: LiteralOrFieldOrExpression;

    /**
     * Offset to apply to position of element to obtain x- position in induced layout.
     * For example, you may wish to offset by the width of a mark, so that the offset layout coincides with the centre
     * (or a specific edge) of mark rather than the top-left corner.
     */
    y?: LiteralOrFieldOrExpression;
  };

  /**
   * List of enabled interactions.
   */
  actions?: Interaction[];

  /**
   * Layout definitions scoped to this vis element.
   */
  layouts?: LayoutSpec[];

  /**
   * Grouping definitions scoped to this vis element.
   */
  groupings?: Grouping[];

  /**
   * Scale definitions scoped to this vis element.
   */
  scales?: Scale[];

  /**
   * Ordering definitions scoped to this vis element.
   */
  orderings?: OrderingSpec[];

  /**
   * Vis elements to be nested inside this vis element.
   */
  vis?: VisEntryBase[];

  /**
   * The name of a selection: if set, the mark will only be drawn for data elements that are included in the selection.
   */
  ifInSelection?: string;

  /**
   * An optional expression: if specified, the mark will only be drawn for data elements for which the expression evaluates to true.
   */
  if?: string;
}

type VisEntryNonTable = VisEntryBase & {
  /**
   *  Name of the layout ot be used to assign a position to each element.
   */
    layout?: string;
  /**
   * If `true`, will draw axes corresponding to the layout.
   */
  drawAxes?: boolean;

  /**
   * Set to either true or an object to draw border around this vis entry.
   * `stroke`/`strokeWidth`/`strokeOpacity`/`strokeDash` properties of the object affect appearance of the border.
   */
  border?: BorderSpec | boolean;
}

type BorderSpec = {
  stroke?: string;
  strokeWidth?: number;
  strokeOpacity?: number;
  strokeDash?: [number, number];
}

type VisEntryNormal = VisEntryNonTable & {
  entries: string;
};


type VisEntryTable = VisEntryBase & {
  /**
   * Name of the table to render.
   */
  table: string;

  /**
   * Specification for row labels.
   */
  rowLabels?: any;
  /**
   * Specification for column labels.
   */
  colLabels?: any;

  /**
   * Properties for ruled horizontal lines between each row in the table.
   */
  rowLines?: any;

  /**
   * Properties for ruled horizontal lines through the middle of each row in the table.
   */
  rowMidLines?: any;

  /**
   * Properties for ruled vertical lines between each column in the table.
   */
  colLines?: any;

  /**
   * Properties for ruled vertical lines through the middle of each column in the table.
   */
  colMidLines?: any;
};

type VisEntry =  VisEntryTable | VisEntryNormal | VisEntryNonTable;


type Grouping = {

  /**
   * The name that will be given to the grouping when it is created.
   */
  name: string;
  data?: string;

  network?: string;

  /**
   * The fields that wll be used for grouping.
   */
  common: string[];
};

type GaGraphDataSpec = DataWithoutTransform & { transform?: TransformSpec[]; }

//Omit<UrlData, "transform"> //& {transform?: TransformSpec[];}

export type GraphgraSpec = {
  pan?: ModifierDefinitions; // modifier keys to enable panning
  zoom?: boolean;

  width?: number;
  height?: number;

  x?: number;
  y?: number;

  /**
   * A list of parameters, who values can be changed dynamically by interactions.
   */
  parameters?: Parameter[];

  /**
   * An array of objects describing the data sources.
   */
  data?: GaGraphDataSpec[];
  /**
   * Grouping
   */
  groupings?: Grouping[];
  /**
   * An array of objects defining how networks are created from the imported data.
   * Generation of a network is necessary to compute network metrics (e.g., node degree) or apply topology-based network layouts.
   */
  networks?: NetworkSpec[];
  /**
   * A layout is a mapping from a datum to a 2D location.
   * This includes layouts based on network topology, as well as attribute-based layouts.
   */
  layouts?: LayoutSpec[];
  /**
   * A scale is a mapping from data to an aspect of a visual encoding.
   */
  scales?: Scale[];

  /**
   * An ordering is a mapping from data to a natural number.
   */
  orderings?: OrderingSpec[];
  tables?: TableSpec[];
  /**
   * Definition of the mapping from data to visual elements.
   */
  vis?: VisEntry[];

  maps?: MapSpec[];

}


type MapSpecBase = {
  /**
   * Name for this map.
   */
  name: string;

  x?: number;
  y?: number;

  width?: number;
  height?: number;

  /**
   * Latitude and longitude of the center of this map.
   */
  center?: [number, number];
  /**
   * Zoom level of the map.
   */
  zoomLevel?: number;
}

// See https://leaflet-extras.github.io/leaflet-providers/preview/
type MapSpecNamed = MapSpecBase & {
  tiles: "Mapnik" | "Stadia.AlidadeSmooth" |  "Stamen_Terrain" | "Stamen_TerrainBackground" | "Stamen.TonerBackground";
}

type MapSpecCustom = MapSpecBase & {
  /**
   * Url for the map tiles, in the format accepted by Leaflet.js
   */
  url: string;

  /**
   * Attribution message to be displayed below the map.
   */
  attribution?: string;
  /**
   * API access token.
   */
  accessToken?: string;
  /**
   * The minimum enabled zoom level.
   */
  minZoom?: number;
  /**
   * The maximum enabled zoom level.
   */
  maxZoom?: number;
  /**
   * Subdomains of the tile service (a work-around for browser limits on parallel requests per domain).
   */
  subdomains?: string;
}

type MapSpec = MapSpecNamed | MapSpecCustom;


// TODO: tighten this up - eg. exactly one of test/isSelection preent
// TODO: use FullCondition in spec
type BaseCondition = {
    value: any;
    [field: string]: any;
};

type Condition = BaseCondition & ({test: string} | {inSelection: string} | {selectionIsEmpty: string} | {selectionIsNOtEmpty: string})
