import { LayoutSpecBase } from "./topologicalLayout";

export type LayoutSpecSetCola = LayoutSpecBase & {
  /**
   * Use SetCoLa tp generate a set of layout constraints, which are then supplied to the CoLA layout algorithm.
   * For further details, see https://github.com/jamesscottbrown/setcola
   */
  type: "setcola";

  /**
   * Reference elements used for positioning: horizontal or vertical lines, or points.
   */
  guides?: SetColaGuide[];

  /**
   * Constraints in SetCola format.
   */
  constraints?: SetColaConstraint[]

  size?: [number, number];
  avoidOverlaps?: boolean;
  convergenceThreshold?: number;
  nodeSize?: number;
  linkDist?: number;
  numIterations?: 100;

}


type SetColaAlignmentConstraint = {
  /**
   * Constrain nodes to be vertically or horizontally alogned.
   */
  constraint: "align";
  /**
   * Axis alogn which the nodes will be aligned.
   */
  axis: "x" | "y";

  "orientation": "center" | "left" | "right" | "top" | "bottom";
}

type PositionConstraint = {
  /**
   * Position nodes on one side of a guide.
   */
  constraint: "position";

  position: "left" | "right" | "above" | "below";

  /**
   * Name of the guide to position relative to.
   */
  of: string;


  /**
   * Minimum amount of space between the node and guide.
   */
  gap?: number;
}


type OrderConstraint = {
  /**
   * Position nodes in a particular order (either horizontally or vertically)
   */
  constraint: "order";

  /**
   * The axis along which the nodes are ordered.
   */
  axis: "x" | "y";

  /**
   * the name of te property used to determine the order.
   * TODO: should add wrapper to accept a NetPaorama order.
   */
  by: string;

  /**
   * Explicitly set the order of nodes; this is interpreted as a list of node ids.
   */
  order?: number[];

  /**
   * If true, the sort order is reversed.
   */
  reverse: boolean;

  /**
   * Defines the amount of space that each section of the order should take up.
   */
  band?: number;

};

type CircleConstraint = {
  /**
   * Arrange nodes so that they approximate a circle layout.
   */
  constraint: "circle";

  /**
   * Can be either "center" or the name of a guide.
   */
  around: "center" | string;

  /**
   * Radius of the circle.
   */
  gap: number;
}

type HullConstraint = {
  /**
   * Adds an enclosing boundary around the nodes.
   */
  constraint: "hull";
}

type PaddingConstraint = {
  /**
   * Adds specified padding around nodes.
   */
  constraint: "padding";

  /**
   * Amount of padding.
   */
  amount: number;

}


/////////////

/**
 * This strategy partitions all the nodes into disjoint sets based on the value of the partitionProperty of the node.
 * You may also specify a list of values to include or exclude, which checks for those values explicitly when completing the partition.
 */
type SetDefinitionByPartition = {
  partition: string;
  include?: any[];
  exclude?: any[];
}

/**
 * This strategy defines explict sets based on a predicate on the graph nodes. You may refer to properties of the node using dot syntax
 */
type SetDefinitionByPredicate = {
  expr: string;

  /**
   * The (optional) name for the set allows you to refer to this set later in the specificaiton.
   */
  name?: string;
}

/**
 * This strategy generates sets by identify node keys.
 */
type SetDefinitionByKeys = {
  collect: string[];
}

type SetDefinition = SetDefinitionByPartition | SetDefinitionByPredicate | SetDefinitionByKeys | string;

//////////////////

type SetColaGuideHorizontal = {
  name: string;
  y: number;
}

type SetColaGuideVertical = {
  name: string;
  x: number;
}

type SetColaGuidePoint = {
  name: string;
  x: number;
  y: number;
}

type SetColaGuide = SetColaGuideVertical | SetColaGuideHorizontal | SetColaGuidePoint;

//////////////////


type SetColaConstraint = {
  name?: string;
  sets: SetDefinition | SetDefinition[];
  forEach: SetColaConstraintSingle[];
}

type SetColaConstraintSingle = SetColaAlignmentConstraint | PositionConstraint | OrderConstraint | CircleConstraint | HullConstraint | PaddingConstraint;

