import type { LayoutSpecBase } from "./topologicalLayout";

type ColaInequalityConstraint = {
  /**
   * The axis along which the constraint applies.
   * */
  axis: "x" | "y";

  /**
   * Index of node 1
   * */
  left: number;
  /**
   * Index of node 2
   * */
  right: number;
  /**
   * Minimum separation along specified axis (or exact separation, if `equality` is true).
   * */
  gap: number;
  /**
   * If equality is true, then this is an equality constraint
   * */
  equality?: boolean;
}
type NodeAndOffset = {
  /**
   * Index of node.
   * */
  node: string;

  /**
   * Offset from alignment line.
   * */
  offset: string;
}

type ColaAlignmentConstraint = {
  type: "alignment";
  /**
   * The axis along which the constraint applies.
   * */
  axis: "x" | "y";
  /**
   * Offsets of each node from the alignment line.
   */
  offsets: NodeAndOffset[];
}

type ColaConstraint = ColaInequalityConstraint | ColaAlignmentConstraint;

export type LayoutSpecCola = LayoutSpecBase & {
  /**
   * Use WebCoLa/cola.js tp apply the CoLA layout algorithm.
   * For further details, see https://ialab.it.monash.edu/webcola/doc/classes/_layout_.layout.html
   */
  type: "webcola";

  /**
   * "The automatic layout will compute layout that tries to keep links (AKA edges) as close as possible to this length"
   */
  linkDistance?: number;
  /**
   * "list of constraints of various types"
   */
  constraints?: ColaConstraint[];
  /**
   * "if true, the layout will not permit overlaps of the node bounding boxes (defined by the width and height properties on nodes)"
   */
  avoidOverlap?: boolean;

  /**
   * "Default size (assume nodes are square so both width and height) to use in packing if node width/height are not specified."
   */
  defaultNodeSize?: number;

  /**
   * "
   *
   * if true, the final step of the start method will be to nicely pack connected components of the graph.
   * works best if start() is called with a reasonable number of iterations specified and each node has a bounding box (defined by the width and height properties on nodes)."
   */
  handleDisconnected?: boolean;

  convergenceThreshold?: number;

  /**
   * "causes constraints to be generated such that directed graphs are laid out either from left-to-right or top-to-bottom.
   * a separation constraint is generated in the selected axis for each edge that is not involved in a cycle (part of a strongly connected component)"
   */
  flowLayout?: { axis: "x" | "y", minSeparation: number };

  /**
   * " compute an ideal length for each link based on the graph structure around that link.
   * you can use this (for example) to create extra space around hub-nodes in dense graphs.
   * In particular this calculation is based on the "symmetric difference" in the neighbour sets of the source and target:
   * i.e. if neighbours of source is a and neighbours of target are b then calculation is: |a intersection b|/|a union b|n."
   */
  jaccardLinkLength?: number;

  /**
   * " compute an ideal length for each link based on the graph structure around that link.
   * you can use this (for example) to create extra space around hub-nodes in dense graphs.
   * In particular this calculation is based on the "symmetric difference" in the neighbour sets of the source and target:
   * i.e. if neighbours of source is a and neighbours of target are b then calculation is: |a intersection b|/|a union b|n."
   */
  symmetricDiffLinkLength?: number;
  /**
   *
   */
  numStepsUnconstrained?: number;
}