// For cytoscape
// TODO: add type with options for each alyout algorithm

// TODO: flag options as optional
// TODO: defaults

// N.B - ignore cola as it can be used directly


import type { LayoutSpecBase } from "./topologicalLayout";
import type { LiteralOrFieldOrExpression } from "./topologicalLayoutD3Force";

export type LayoutSpecCytoscape =
  CytoscapeLayoutGrid |
  CytoscapeLayoutCircle |
  CytoscapeLayoutConcentric |
  CytoscapeLayoutBreadthfirst |
  CytoscapeLayoutCose |
  CytoscapeLayoutAvsdf |
  CytoscapeLayoutDagre |
  CytoscapeLayoutElk |
  CytoscapeLayoutKlay |
  CytoscapeLayoutFcose |
  CytoscapeLayoutCoseBilkent |
  CytoscapeLayoutCise |
  CytoscapeLayoutEuler |
  CytoscapeLayoutSpread;


export type LayoutSpecCytoscapeBase = LayoutSpecBase & {
  type: "cytoscape";
//  algorithm?: CytoscapeLayoutAlgorithm;

  animate?: false;

  /**
   * If true, prevents node overlap.
   */
  avoidOverlap?: boolean;
  /**
   * "extra spacing around nodes when avoidOverlap is true"
   */
  avoidOverlapPadding?: number;
//  [otherOptions: string]: string | number;
}

type CytoscapeLayoutGrid = LayoutSpecCytoscapeBase & {

  /**
   * "The grid layout puts nodes in a well-spaced grid."
   * https://js.cytoscape.org/#layouts/grid
   */
  algorithm: "grid";

  /**
   * "Force number of rows in the grid"
   * Default: undefined
   */
  rows?: number;
  /**
   * "force num of columns in the grid"
   * Default: undefined
   */
  cols?: number;
}

type CytoscapeLayoutCircle = LayoutSpecCytoscapeBase & {

  /**
   * "The circle layout puts nodes in a circle."
   * https://js.cytoscape.org/#layouts/circle
   */
  algorithm: "circle";

  /**
   * "where nodes start in radians"
   * Default: 3 / 2 * Math.PI
   */
  startAngle?: number;
  /**
   * "how many radians should be between the first and last node (defaults to full circle)"
   * Default: undefined
   */
  sweep?: number;

  /**
   * "whether the layout should go clockwise (true) or counterclockwise/anticlockwise (false)"
   * Default: true
   */
  clockwise?: boolean,
}

type CytoscapeLayoutConcentric = LayoutSpecCytoscapeBase & {

  /**
   * "The concentric layout positions nodes in concentric circles, based on a metric that you specify to segregate the nodes into levels."
   * https://js.cytoscape.org/#layouts/concentric
   */
  algorithm: "concentric";

  /**
   * "where nodes start in radians"
   * Default: 3 / 2 * Math.PI
   */
  startAngle?: number;
  /**
   * "how many radians should be between the first and last node (defaults to full circle)"
   * Default: undefined
   */
  sweep?: number;

  /**
   * "whether the layout should go clockwise (true) or counterclockwise/anticlockwise (false)"
   * Default: true
   */
  clockwise?: boolean;

  /**
   * "whether levels have an equal radial distance between them"
   * Default: false
   */
  equidistant?: boolean;
  /**
   * "min spacing between outside of nodes (used for radius adjustment)"
   * Default: 10
   */
  minNodeSpacing?: number;

  /**
   * "Returns numeric value for each node, placing higher nodes in levels towards the centre"
   * Default is node degree
   */
  concentric?: LiteralOrFieldOrExpression;

  /**
   * "the variation of concentric values in each level"
   * Default: 4
   */
  levelWidth?: LiteralOrFieldOrExpression;
}


type CytoscapeLayoutBreadthfirst = LayoutSpecCytoscapeBase & {
  /**
   * "The breadth-first layout puts nodes in a hierarchy, based on a breadth-first traversal of the graph.
   * It is best suited to trees and forests in its default top-down mode, and it is best suited to DAGs in its circle mode."
   * https://js.cytoscape.org/#layouts/breadthfirst
   */
  algorithm: "breadthfirst";

  /**
   * "put depths in concentric circles if true, put depths top down if false"
   * Default: false
   */
  circle?: boolean;
  /**
   * "whether to create an even grid into which the DAG is placed (circle:false only)"
   * Default: false
   */
  grid?: boolean;
  /**
   * "whether to shift nodes down their natural BFS depths in order to avoid upwards edges (DAGS only)"
   * Default: false
   */
  maximal?: boolean;

  // TODO: depthsort function is a function of 2 values
}

type CytoscapeLayoutCose = LayoutSpecCytoscapeBase & {
  /**
   * "The cose (Compound Spring Embedder) layout uses a physics simulation to lay out graphs. It works well with noncompound graphs and it has additional logic to support compound graphs well.
   *
   * It was implemented by Gerardo Huck as part of Google Summer of Code 2013 (Mentors: Max Franz, Christian Lopes, Anders Riutta, Ugur Dogrusoz).
   *
   * Based on the article “A layout algorithm for undirected compound graphs” by Ugur Dogrusoz, Erhan Giral, Ahmet Cetintas, Ali Civril and Emek Demir.
   *
   * The cose layout is very fast and produces good results. The cose-bilkent extension is an evolution of the algorithm that is more computationally expensive but produces near-perfect results.."
   * https://js.cytoscape.org/#layouts/cose
   */
  algorithm: "cose";


  /**
   * "Extra spacing between components in non-compound graphs"
   * Default: 40
   */
  componentSpacing?: number;

  /**
   * "Node repulsion (non overlapping) multiplier"
   * Default: 2048
   */
  nodeRepulsion?: LiteralOrFieldOrExpression;

  /**
   * "Node repulsion (overlapping) multiplier"
   * Default: 4
   */
  nodeOverlap?: number;


  /**
   * "Ideal edge (non nested) length"
   * Default: 32
   */
  idealEdgeLength?: LiteralOrFieldOrExpression;

  /**
   * "Divisor to compute edge forces"
   * Default: 32
   */
  edgeElasticity?: LiteralOrFieldOrExpression;

  /**
   * "Nesting factor (multiplier) to compute ideal edge length for nested edges"
   * Default: 1.2
   */
  nestingFactor?: number;

  /**
   * "Gravity force (constant)"
   * Default: 1
   */
  gravity?: number;

  /**
   * "Maximum number of iterations to perform"
   * Default: 1000
   */
  numIter?: number;

  /**
   * "Initial temperature (maximum node displacement)"
   * Default: 1000
   */
  initialTemp?: number;

  /**
   * "Cooling factor (how the temperature is reduced between consecutive iterations)"
   * Default: 0.99
   */
  coolingFactor?: number;

  /**
   * "Lower temperature threshold (below this point the layout will end)"
   * Default 1.0
   */
  minTemp?: number;
}

type CytoscapeLayoutAvsdf = LayoutSpecCytoscapeBase & {
  /**
   * "An implementation of the Circular Drawing Algorithm by Hongmei He & Ondrej Sýkora."
   * https://github.com/iVis-at-Bilkent/cytoscape.js-avsdf
   */
  algorithm?: "avsdf";

  /**
   * "How apart the nodes are"
   * Default: 60
   */
  nodeSeparation?: number;
}


type CytoscapeLayoutDagre = LayoutSpecCytoscapeBase & {
  /**
   * "The Dagre layout for DAGs and trees for Cytoscape.js
   *
   * The dagre layout organises the graph using a DAG (directed acyclic graph) system, written by Chris Pettitt. It is especially suitable for DAGs and trees. For more information, please refer to Dagre's documentation."
   * https://github.com/cytoscape/cytoscape.js-dagre
   */
  algorithm: "dagre";

  /**
   * "the separation between adjacent nodes in the same rank"
   */
  nodeSep?: number;
  /**
   * "the separation between adjacent edges in the same rank"
   */
  edgeSep?: number;
  /**
   * "the separation between each rank in the layout"
   */
  rankSep?: number;

  /**
   * "'TB' for top to bottom flow, 'LR' for left to right,"
   */
  rankDir?: "TB" | "LR";

  /**
   * "alignment for rank nodes. Can be 'UL', 'UR', 'DL', or 'DR', where U = up, D = down, L = left, and R = right"
   */
  align?: "UL" | "UR" | "DL" | "DR",

  /**
   * "If set to 'greedy', uses a greedy heuristic for finding a feedback arc set for a graph.
   *  A feedback arc set is a set of edges that can be removed to make a graph acyclic."
   */
  acyclicer?: "greedy" | undefined;

  /**
   * "Type of algorithm to assign a rank to each node in the input graph"
   */
  ranker?: "network-simplex" | "tight-tree" | "longest-path";

  /**
   * "number of ranks to keep between the source and target of the edge"
   * Default: 1
   */
  minLen?: LiteralOrFieldOrExpression;

  /**
   * "higher weight edges are generally made shorter and straighter than lower weight edges"
   * Default: 1
   */
  edgeWeight?: LiteralOrFieldOrExpression;
}


// TODO: add options from https://www.eclipse.org/elk/reference/algorithms.html
type CytoscapeLayoutElk = LayoutSpecCytoscapeBase & {
  /**
   * "The elk layout algorithm adapter for Cytoscape.js
   *
   * ELK is a set of layout algorithms implemented by the Eclipse Foundation in Java. The source code is compiled to JS by the ELK.js project using GWT."
   * https://github.com/cytoscape/cytoscape.js-elk
   */
  algorithm: "elk";

  elk?: {

    /**
     * "The main field to set is `algorithm`, which controls which particular layout algorithm is used.
     * - box : Pack the nodes like boxes.
     * disco : Pack the (disconnected) components. A secondary layout may be applied to each component via options.elk.componentLayoutAlgorithm.
     * force : Apply a basic force-directed layout.
     * layered : Apply a hierarchical layout, appropriate for DAGs and trees.
     * mrtree : Apply a traditional, hierarchical tree layout.
     * random : Apply random positions to the nodes.
     * stress : Apply a force-directed layout"
     * Default: "layered"
     */
    algorithm?: "box" | "disco" | "force" | "layered" | "mrtree" | "random" | "stress";
    [option: string]: string | number;
  }

}


//////////////////////////////////////

// TODO: finish conversion
type CytoscapeLayoutKlay = LayoutSpecCytoscapeBase & {
  /**
   * "The Klay layout algorithm for Cytoscape.js (demo)
   *
   * This discrete layout creates good results for most graphs and it supports compound nodes."
   * https://github.com/cytoscape/cytoscape.js-klay
   */
  algorithm: "klay";

  /**
   * "Edges with a non-nil value are skipped when greedy edge cycle breaking is enabled"
   * Default: () => null
   */
  priority?: LiteralOrFieldOrExpression;


  klay: {
    /**
     * "Adds bend points even if an edge does not change direction"
     * Default: false
     */
    addUnnecessaryBendpoints?: boolean;
    /**
     * "The aimed aspect ratio of the drawing, that is the quotient of width by height"
     * Default: 1.6
     */
    aspectRatio?: number;

    /**
     * "Tries to further compact components (disconnected sub-graphs)."
     * Default: false
     */
    compactComponents?: boolean;

    /**
     * "Strategy for crossing minimization.
     *   LAYER_SWEEP The layer sweep algorithm iterates multiple times over the layers, trying to find node orderings that minimize the number of crossings. The algorithm uses randomization to increase the odds of finding a good result. To improve its results, consider increasing the Thoroughness option, which influences the number of iterations done. The Randomization seed also influences results.
     *   INTERACTIVE Orders the nodes of each layer by comparing their positions before the layout algorithm was started. The idea is that the relative order of nodes as it was before layout was applied is not changed. This of course requires valid positions for all nodes to have been set on the input graph before calling the layout algorithm. The interactive layer sweep algorithm uses the Interactive Reference Point option to determine which reference point of nodes are used to compare positions."
     * Default: "LAYER_SWEEP"
     */
    crossingMinimization?: "LAYER_SWEEP" | "INTERACTOVE";

    /**
     *  "Strategy for cycle breaking. Cycle breaking looks for cycles in the graph and determines which edges to reverse to break the cycles. Reversed edges will end up pointing to the opposite direction of regular edges (that is, reversed edges will point left if edges usually point right).
     *   GREEDY This algorithm reverses edges greedily. The algorithm tries to avoid edges that have the Priority property set.
     *   INTERACTIVE The interactive algorithm tries to reverse edges that already pointed leftwards in the input graph. This requires node and port coordinates to have been set to sensible values."
     * Default: "GREEDY"
     */
    cycleBreaking?: "GREEDY" | "INTERACTIVE",

    /**
     * "Overall direction of edges: horizontal (right / left) or vertical (down / up)"
     * Default: "UNDEFINED
     */
    direction?: "UNDEFINED" | "RIGHT" | "LEFT" | "DOWN" | "UP",

    /**
     * "Defines how edges are routed (POLYLINE, ORTHOGONAL, SPLINES)"
     * Defult: ORTHOGONAL
     */
    edgeRouting?: "ORTHOGONAL" | "POLYLINE" | "SPLINES",
    /**
     * "Factor by which the object spacing is multiplied to arrive at the minimal spacing between edges."
     * Default: 0.5
     */
    edgeSpacingFactor?: number;
    /**
     * "Whether feedback edges should be highlighted by routing around the nodes"
     * Default: false
     */
    feedbackEdges?: boolean;

    /**
     * "Tells the BK node placer to use a certain alignment instead of taking the optimal result.  This option should usually be left alone.
     *   /* NONE Chooses the smallest layout from the four possible candidates.
     *   LEFTUP Chooses the left-up candidate from the four possible candidates.
     *   RIGHTUP Chooses the right-up candidate from the four possible candidates.
     *   LEFTDOWN Chooses the left-down candidate from the four possible candidates.
     *   RIGHTDOWN Chooses the right-down candidate from the four possible candidates.
     *   BALANCED Creates a balanced layout from the four possible candidates."
     *   Default: "NONE"
     */
    fixedAlignment?: "NONE" | "LEFTUP" | "RIGHTUP" | "LEFTDOWN" | "RIGHTDOWN" | "BALANCED";

    /**
     * "Factor by which the usual spacing is multiplied to determine the in-layer spacing between objects."
     * Default: 1.0
     */
    inLayerSpacingFactor?: number;

    /**
     * "Whether the selected layouter should consider the full hierarchy"
     * Default: false
     */
    layoutHierarchy?: boolean;

    /**
     * "Dampens the movement of nodes to keep the diagram from getting too large"
     * Default: 0.3
     */
    linearSegmentsDeflectionDampening?: number;

    /**
     * "Edges that have no ports are merged so they touch the connected nodes at the same points."
     * Default: false
     */
    mergeEdges?: boolean;

    /**
     * "If hierarchical layout is active, hierarchy-crossing edges use as few hierarchical ports as possible"
     * Default: true
     */
    mergeHierarchyCrossingEdges?: boolean;
    /**
     *   "Strategy for node layering
     *   NETWORK_SIMPLEX This algorithm tries to minimize the length of edges. This is the most computationally intensive algorithm. The number of iterations after which it aborts if it hasn't found a result yet can be set with the Maximal Iterations option.
     *   LONGEST_PATH A very simple algorithm that distributes nodes along their longest path to a sink node.
     *   INTERACTIVE Distributes the nodes into layers by comparing their positions before the layout algorithm was started.
     *   The idea is that the relative horizontal order of nodes as it was before layout was applied is not changed.
     *   This of course requires valid positions for all nodes to have been set on the input graph before calling the layout algorithm.
     *   The interactive node layering algorithm uses the Interactive Reference Point option to determine which reference point of nodes are used to compare positions."
     *   Default: "NETWORK_SIMPLEX"
     */
    nodeLayering?: "NETWORK_SIMPLEX" | "LONGEST_PATH" | "INTERACTIVE";

    /**
     * "Strategy for Node Placement
     *   BRANDES_KOEPF Minimizes the number of edge bends at the expense of diagram size: diagrams drawn with this algorithm are usually higher than diagrams drawn with other algorithms.
     *   LINEAR_SEGMENTS Computes a balanced placement.
     *   INTERACTIVE Tries to keep the preset y coordinates of nodes from the original layout. For dummy nodes, a guess is made to infer their coordinates. Requires the other interactive phase implementations to have run as well.
     *   SIMPLE Minimizes the area at the expense of... well, pretty much everything else."
     *   Default: "BRANDES_KOEPF"
     */
    nodePlacement?: "BRANDES_KOEPF" | "LINEAR_SEGMENTS" | "INTERACTIVE" | "SIMPLE";

    /**
     * "Seed used for pseudo-random number generators to control the layout algorithm; 0 means a new seed is generated"
     * Default: 1
     */
    randomizationSeed?: number;
    /**
     * "Whether a self-loop is routed around or inside its node"
     * Default: false
     */
    routeSelfLoopInside?: boolean;
    /**
     * "Whether each connected component should be processed separately"
     * Default: true
     */
    separateConnectedComponents?: boolean;
    /**
     * "Overall setting for the minimal amount of space to be left between objects"
     * Default: 20
     */
    spacing?: number;

    /**
     * "How much effort should be spent to produce a nice layout"
     * Default: 7
     */
    thoroughness?: number;
  }
}

type CytoscapeLayoutFcose = LayoutSpecCytoscapeBase & {
  /**
   * "fCoSE (pron. "f-cosay", fast Compound Spring Embedder), is a faster version of our earlier compound spring embedder algorithm named CoSE, implemented as a Cytoscape.js extension by i-Vis Lab in Bilkent University."
   * https://github.com/iVis-at-Bilkent/cytoscape.js-fcose
   */
  algorithm: "fcose";

  /**
   *" - "draft" only applies spectral layout
   * - "default" improves the quality with incremental layout (fast cooling rate)
   * - "proof" improves the quality with incremental layout (slow cooling rate)"
   * Default: "default"
   */
  quality?: "draft" | "default" | "proof",

  //
  //

  /**
   * "Use random node positions at beginning of layout. If this is set to false, then quality option must be "proof""
   * Default: true
   */
  randomize?: boolean;

  /* spectral layout options */

  /**
   * "False for random, true for greedy sampling"
   * Default: true
   */
  //
  samplingType?: boolean;
  /**
   * "Sample size to construct distance matrix"
   * Default: 25
   */
  sampleSize?: number;
  /**
   * "Separation amount between nodes"
   * Default: 75
   */
  nodeSeparation?: number;
  /**
   * "Power iteration tolerance"
   * Default: 0.0000001
   */
  piTol?: number;

  /* incremental layout options */

  /**
   * "Node repulsion (non overlapping) multiplier"
   * Default: 4500
   */
  nodeRepulsion?: LiteralOrFieldOrExpression;

  /**
   * "Ideal edge (non nested) length"
   * Default: 50
   */
  idealEdgeLength?: 50;

  /**
   * "Divisor to compute edge forces"
   * Default: 0.45
   */
  edgeElasticity?: 0.45;

  /**
   * "Nesting factor (multiplier) to compute ideal edge length for nested edges"
   * Default: 0.1
   */
  nestingFactor?: number;

  /**
   * "Maximum number of iterations to perform - this is a suggested value and might be adjusted by the algorithm as required"
   * Default: 2500
   */
  numIter?: number;

  /**
   * "For enabling tiling"
   * Default: true
   */
  tile?: boolean;

  /**
   * "Represents the amount of the vertical space to put between the zero degree members during the tiling operation"
   * Default: 10
   */
  tilingPaddingVertical?: LiteralOrFieldOrExpression;

  /**
   * "Represents the amount of the horizontal space to put between the zero degree members during the tiling operation"
   * Default: 10
   */
  tilingPaddingHorizontal?: LiteralOrFieldOrExpression;

  /**
   * "Gravity force (constant)"
   * Default: 0.25
   */
  gravity?: number;

  /**
   * "Gravity range (constant) for compounds"
   * Default: 1.5
   */
  gravityRangeCompound?: number;

  /**
   * "Gravity force (constant) for compounds"
   * Default: 1.0
   */
  gravityCompound?: number;

  /**
   * "Gravity range (constant)"
   * Default: 3.8
   */
  gravityRange?: number;

  /**
   * "Initial cooling factor for incremental layout"
   * Default: 0.3
   */
  initialEnergyOnIncremental?: number;

  /* constraint options */

  /**
   * "Fix desired nodes to predefined positions
   * [{nodeId: 'n1', position: {x: 100, y: 200}}, {...}]"
   * Default: undefined
   */
  fixedNodeConstraint?: FixedConstraint[],

  /**
   * "Align desired nodes in vertical/horizontal direction
   * {vertical: [['n1', 'n2'], [...]], horizontal: [['n2', 'n4'], [...]]} "
   * Default: undefined
   */
  alignmentConstraint?: AlignmentConstraint[],
  /**
   * "Place two nodes relatively in vertical/horizontal direction
   * [{top: 'n1', bottom: 'n2', gap: 100}, {left: 'n3', right: 'n4', gap: 75}, {...}]"
   * Default: undefined
   */
  relativePlacementConstraint?: RelativeConstraint[];
}

type FixedConstraint = {
  nodeId: string;
  position: { x: number, y: number }
}

type AlignmentConstraint = {
  vertical?: string[][];
  horizontal?: string[][];
}

type RelativeConstraintH = {
  top: string;
  bottom: string;
  gap: number;
}

type RelativeConstraintV = {
  left: string;
  right: string;
  gap: number;
}

type RelativeConstraint = RelativeConstraintV | RelativeConstraintH;


type CytoscapeLayoutCoseBilkent = LayoutSpecCytoscapeBase & {

  /**
   * "The CoSE (pron. "cosay", Compound Spring Embedder) layout for Cytoscape.js developed by i-Vis Lab in Bilkent University is a spring embedder layout with support for compound graphs (nested structures) and varying (non-uniform) node dimensions. A faster version of this layout style called fCoSE, also supporting user-defined placement constraints can be found here. (demo, compound demo)
   *
   * Please cite the following when using this layout:
   *
   * U. Dogrusoz, E. Giral, A. Cetintas, A. Civril, and E. Demir, "A Layout Algorithm For Undirected Compound Graphs", Information Sciences, 179, pp. 980-994, 2009."
   * https://github.com/cytoscape/cytoscape.js-cose-bilkent
   */
  algorithm: "cose-bilkent";

  /**
   * "- 'draft' fast cooling rate
   *  - 'default' moderate cooling rate
   *  - "proof" slow cooling rate"
   * Default: "default"
   */
  quality?: "draft" | "default" | "proof";

  /**
   * "Whether to include labels in node dimensions. Useful for avoiding label overlap
   * Default: false
   * */
  nodeDimensionsIncludeLabels?: boolean;

  /**
   * "number of ticks per frame; higher is faster but more jerky"
   * Default: 30
   */
  refresh?: number;

  /**
   * "Whether to fit the network view after when done"
   * Default: true
   */
  fit?: boolean;

  /**
   * "Padding on fit"
   * Default: 10
   */
  padding?: number;

  /**
   * "Whether to enable incremental mode"
   * Default: true
   */
  randomize?: boolean;

  /**
   * "Node repulsion (non overlapping) multiplier"
   * Default: 4500
   */
  nodeRepulsion?: number;

  /**
   * "Ideal (intra-graph) edge length"
   * Default: 50
   */
  idealEdgeLength?: number;

  /**
   * "Divisor to compute edge forces"
   * Default: 0.45
   */
  edgeElasticity?: number;

  /**
   * "Nesting factor (multiplier) to compute ideal edge length for inter-graph edges"
   * Default: 0.1
   */
  nestingFactor?: number;

  /**
   * "Gravity force (constant)"
   * Default: 0.25
   */
  gravity?: number;

  /**
   * "Maximum number of iterations to perform"
   * Default: 2500
   */
  numIter?: number;

  /**
   * "Whether to tile disconnected nodes"
   * Default: true
   */
  tile?: boolean;

  /**
   * "Amount of vertical space to put between degree zero nodes during tiling"
   * Default: 10
   */
  tilingPaddingVertical?: LiteralOrFieldOrExpression;

  /**
   * "Amount of horizontal space to put between degree zero nodes during tiling"
   * Default: 10
   */
  tilingPaddingHorizontal?: LiteralOrFieldOrExpression;

  /**
   * "Gravity range (constant) for compounds"
   * Default: 1.5
   */
  gravityRangeCompound?: number;

  /**
   * "Gravity force (constant) for compounds"
   * Default: 1.0
   */
  gravityCompound?: number;

  /**
   * "Gravity range (constant)"
   * Default: 3.8
   */
  gravityRange?: number;

  /**
   * "Initial cooling factor for incremental layout"
   * Default: 0.5
   */
  initialEnergyOnIncremental?: number;

}


//////////////////////////////////////////

// TODO: clusterInfo
type CytoscapeLayoutCise = LayoutSpecCytoscapeBase & {

  /**
   * "CiSE(Circular Spring Embedder) is an algorithm based on the traditional force-directed layout scheme with extensions to move and rotate nodes in the same cluster as a group. Further local improvements may be obtained by flipping clusters and by swapping neighboring node pairs in the same cluster, reducing the edge crossing number.
   *
   * The algorithm is implemented as a Cytoscape.js extension by i-Vis Lab in Bilkent University (demo)
   *
   * Please cite the following when using this layout:
   *
   * M. Belviranli, A. Dilek and U. Dogrusoz, "CiSE: A Circular Spring Embedder Layout Algorithm" in IEEE Transactions on Visualization & Computer Graphics, vol. 19, no. 06, pp. 953-966, 2013."
   * https://github.com/iVis-at-Bilkent/cytoscape.js-cise
   */
  algorithm: "cise";


  /**
   * "separation amount between nodes in a cluster. note: increasing this amount will also increase the simulation time"
   * Default: 12.5
   */
  nodeSeparation?: number;

  /**
   * "Inter-cluster edge length factor  (2.0 means inter-cluster edges should be twice as long as intra-cluster edges)"
   * Default: 1.4
   */
  idealInterClusterEdgeLengthCoefficient?: number;

  /**
   * "Whether to pull on-circle nodes inside of the circle"
   * Default: false
   */
  allowNodesInsideCircle?: boolean;

  /**
   * "Max percentage of the nodes in a circle that can move inside the circle"
   * Default: 0.1
   */
  maxRatioOfNodesInsideCircle?: number;

  /**
   * "Lower values give looser springs; Higher values give tighter springs"
   * Default: 0.45
   */
  springCoeff?: number;

  /**
   * "Node repulsion (non overlapping) multiplier"
   * Default: 4500
   */
  nodeRepulsion?: number;

  /**
   * "Gravity force (constant)"
   * Default: 0.25
   */
  gravity?: number;

  /**
   * "Gravity range (constant)"
   * Default: 3.8
   */
  gravityRange?: number;

}


type CytoscapeLayoutEuler = LayoutSpecCytoscapeBase & {
  /**
   * "Euler is a fast, high-quality force-directed (physics simulation) layout for Cytoscape.js
   * It is based on cytoscape-ngraph.forcelayout, with several parts reworked and several general improvements."
   * https://github.com/cytoscape/cytoscape.js-euler
   */
  algorithm: "euler";

  /**
   * "The ideal length of a spring
   *   - This acts as a hint for the edge length
   *   - The edge length can be longer or shorter if the forces are set to extreme values"
   *   Default: 80
   */
  springLength?: LiteralOrFieldOrExpression;

  /**
   * "  Hooke's law coefficient
   *   - The value ranges on [0, 1]
   *   - Lower values give looser springs
   *   - Higher values give tighter springs"
   *   Default: 0.0008
   */
  springCoeff?: LiteralOrFieldOrExpression;

  /**
   * "  Coulomb's law coefficient
   *   - Makes the nodes repel each other for negative values
   *   - Makes the nodes attract each other for positive values"
   *   Default: 4
   */
  mass?: LiteralOrFieldOrExpression;

  /**
   * "Coulomb's law coefficient
   *  - Makes the nodes repel each other for negative values
   *  - Makes the nodes attract each other for positive values"
   *  Default: -1.2
   */
  gravity?: number;

  /**
   * "A force that pulls nodes towards the origin (0, 0). Higher values keep the components less spread out"
   * Default: 0.001
   */
  pull?: number;

  /**
   * "Theta coefficient from Barnes-Hut simulation
   * - Value ranges on [0, 1]
   * - Performance is better with smaller values
   *  - Very small values may not create enough force to give a good result"
   *  Default: 0.666
   */
  theta?: number;

  /**
   * "Friction / drag coefficient to make the system stabilise over time"
   * Default: 0.02
   */
  dragCoeff?: number;

  /**
   * "When the total of the squared position deltas is less than this value, the simulation ends"
   * Default: 1
   */
  movementThreshold?: 1;
}

// TODO: think about specifying the prelayout
type CytoscapeLayoutSpread = LayoutSpecCytoscapeBase & {
  /**
   * "The spread layout uses a force-directed physics simulation with several external libraries. The layout tries to keep elements spread out evenly, making good use of constrained space.
   *
   * The layout makes use of CoSE (MIT) and rhill-voronoi-core.js (MIT). CoSE is already bundled in Cytoscape.js, and rhill-voronoi-core.js is bundled in cytoscape-spread.
   *
   * There are two phases to this layout:
   *
   * (1) A force-directed layout provides initial positions for the nodes. By default, the embedded version of CoSE is used, because it is fast and because it does not increase your app's bundle size any more than using Cytoscape.js itself. You can use an alternative layout by specifying options.prelayout with the layout options you want to use for the first phase (e.g. { name: 'grid' }). Alternatively, you can specify options.prelayout: false (falsey) to just use the node's existing positions for the first phase.
   *
   * (2) Voronoi is used to spread out the nodes in the remaining space.
   *
   * Note that since you are composing layouts with phase (1), where options.prelayout is non-falsey, you will have more layout events. For example, you will have more than one layoutstop event -- one for the Spread layout overall and one for the prelayout within phase (1) of Spread.
   *
   * If you skip phase (1) with options.prelayout falsey, you will not get extra events within Spread. You can use promise chaining with two layouts to get the same effect as running a layout in phase (1), i.e.:"
   * https://github.com/cytoscape/cytoscape.js-spread
   */
  algorithm: "spread";

  /**
   * "Layout options for the first phase"
   * Default "{ name: 'cose' }"
   */
  prelayout?: any;
  /**
   * "Maximum number of expanding iterations"
   * Default: 4
   */
  maxExpandIterations?: number;
  /**
   * "Uses random initial node positions on true"
   * Default: false
   */
  randomize?: boolean;
}