import type { LayoutSpecBase } from "./topologicalLayout";

// TODO: check if this is redundant with another type elsewhere

type LiteralOrFieldOrExpression = number | { expression: string } | { field: string };

type ForceCenter = {
  /**
   * "The centering force translates nodes uniformly so that the mean position of all nodes (the center of mass if all nodes have equal weight) is at the given position ⟨x,y⟩.
   * This force modifies the positions of nodes on each application; it does not modify velocities, as doing so would typically cause the nodes to overshoot and oscillate around the desired center.
   * This force helps keeps nodes in the center of the viewport, and unlike the positioning force, it does not distort their relative positions."
   */
  force: "center";

  /**
   * The x-coordinate of the centering position.
   */
  x?: number;

  /**
   * The y-coordinate of the centering position.
   */
  y?: number;

  /**
   * the centering force’s strength. A reduced strength of e.g. 0.05 softens the movements on interactive graphs in which new nodes enter or exit the graph
   * Default: 1
   */
  strength: number;
}

type ForceCollide = {
  /**
   * "The collision force treats nodes as circles with a given radius, rather than points, and prevents nodes from overlapping.
   * More formally, two nodes a and b are separated so that the distance between a and b is at least radius(a) + radius(b).
   * To reduce jitter, this is by default a “soft” constraint with a configurable strength and iteration count."
   */
  force: "collide";

  /**
   * "The radius of each node"
   * Default: 1
   */
  radius?: LiteralOrFieldOrExpression;

  /**
   * The force strength: a number in the range [0,1].
   * Default 1
   */
  strength?: number;

  /**
   * the number of iterations per application.
   * Increasing the number of iterations greatly increases the rigidity of the constraint and avoids partial overlap of nodes, but also increases the runtime cost to evaluate the force.
   * Default: 1
   */
  iterations?: number;
}

type ForceNBody = {
  /**
   * "The many-body (or n-body) force applies mutually amongst all nodes.
   * It can be used to simulate gravity (attraction) if the strength is positive, or electrostatic charge (repulsion) if the strength is negative.
   * This implementation uses quadtrees and the Barnes–Hut approximation to greatly improve performance; the accuracy can be customized using the theta parameter.
   * Unlike links, which only affect two linked nodes, the charge force is global: every node affects every other node, even if they are on disconnected subgraphs."
   */
  force: "nbody";

  /**
   * "A positive value causes nodes to attract each other, similar to gravity, while a negative value causes nodes to repel each other, similar to electrostatic charge"
   * Default: -20
   */
  strength?: number;

  /**
   * "To accelerate computation, this force implements the Barnes–Hut approximation which takes O(n log n) per application where n is the number of nodes.
   * For each application, a quadtree stores the current node positions; then for each node, the combined force of all other nodes on the given node is computed.
   * For a cluster of nodes that is far away, the charge force can be approximated by treating the cluster as a single, larger node.
   * The theta parameter determines the accuracy of the approximation: if the ratio w / l of the width w of the quadtree cell to the distance l from the node to the cell’s center of mass is less than theta, all nodes in the given cell are treated as a single node rather than individually."
   * Default: 0.9
   */
  theta?: number;

  /**
   * "the minimum distance between nodes over which this force is considered.
   * A minimum distance establishes an upper bound on the strength of the force between two nearby nodes, avoiding instability.
   * In particular, it avoids an infinitely-strong force if two nodes are exactly coincident; in this case, the direction of the force is random"
   * Default: 1
   */
  distanceMin?: number;

  /**
   * "maximum distance between nodes over which this force is considered. If distance is not specified, returns the current maximum distance, which defaults to infinity.
   * Specifying a finite maximum distance improves performance and produces a more localized layout"
   * Default: infinity
   */
  distanceMax: number;
}

type ForceX = {
  /**
   * A positioning force along the x-axis.
   */
  force: "x";

  /**
   * The target x-coordinate.
   */
  x: LiteralOrFieldOrExpression;

  /**
   * "The strength determines how much to increment the node’s x-velocity: (x - node.x) × strength.
   * For example, a value of 0.1 indicates that the node should move a tenth of the way from its current x-position to the target x-position with each application.
   * Higher values moves nodes more quickly to the target position, often at the expense of other forces or constraints. A value outside the range [0,1] is not recommended."
   * Default: 0.1
   */
  strength?: number;
};

type ForceY = {
  /**
   * A positioning force along the y-axis.
   */
  force: "y";

  /**
   * The target y-coordinate.
   */
  y: LiteralOrFieldOrExpression;

  /**
   * "The strength determines how much to increment the node’s y-velocity: (y - node.y) × strength.
   * For example, a value of 0.1 indicates that the node should move a tenth of the way from its current x-position to the target x-position with each application.
   * Higher values moves nodes more quickly to the target position, often at the expense of other forces or constraints. A value outside the range [0,1] is not recommended."
   * Default: 0.1
   */
  strength?: number;
};


type ForceLink = {
  /**
   * "The link force pushes linked nodes together or apart according to the desired link distance.
   * The strength of the force is proportional to the difference between the linked nodes’ distance and the target distance, similar to a spring force."
   */
  force: "link";

  /**
   * "If links is specified, sets the array of links associated with this force, recomputes the distance and strength parameters for each link, and returns this force.
   *
   * Each link is an object with the following properties:
   *
   *     source - the link’s source node; see simulation.nodes
   *     target - the link’s target node; see simulation.nodes
   *     index - the zero-based index into links, assigned by this method"
   * Default: []
   */
  // links: NetworkLink[];

  // id: string; // field
  /**
   * "Target link distance"
   * Default: 30
   */
  distance?: LiteralOrFieldOrExpression;

  /**
   *  Strength of this force.
   * Default: 1 / Math.min(count(link.source), count(link.target))
   */
  strength?: LiteralOrFieldOrExpression;

  /**
   * "number of iterations per application.
   * Increasing the number of iterations greatly increases the rigidity of the constraint and is useful for complex structures such as lattices, but also increases the runtime cost to evaluate the force."
   * Default: 1
   */
  iterations?: number;
}


type ForceRadial = {
  /**
   * "A positioning force towards a circle of the specified radius"
   */
  force: "radial";

  /**
   * The circle radius
   */
  radius: LiteralOrFieldOrExpression;

  /**
   * "The strength determines how much to increment the node’s x- and y-velocity.
   * For example, a value of 0.1 indicates that the node should move a tenth of the way from its current position to the closest point on the circle with each application.
   * Higher values moves nodes more quickly to the target position, often at the expense of other forces or constraints. A value outside the range [0,1] is not recommended."
   * Default: 0.1
   */
  strength?: number;

  /**
   * The x-coordinate of the circle centre.
   * Default: 0
   */
  x?: number;

  /**
   * The y-coordinate of the circle centre.
   * Default: 0
   */
  y?: number;
}


type SingleForce = ForceCenter | ForceCollide | ForceNBody | ForceLink | ForceX | ForceY | ForceRadial;
//type SingleForce = any;

export  type LayoutSpecD3Force = LayoutSpecBase & {
  /**
   * Compute a force-directed layout using d3-force.
   */
  type: "d3-force";

  /**
   * The number of iterations to perform.
   */
  iterations?: number;

  /**
   * Initial value of the alpha value (default 1).
   */
  alpha?: number;

  /**
   * Value of alpha at which the simulation should stop (default 0.001).
   */
  alphaMin?: number;

  /**
   * Set the target alpha value (default 0).
   */
  alphaTarget?: number;

  /**
   * Set the velocity decay factor (default 0.4).
   */
  velocityDecay?: number;

  /**
   * List of forces to be applied during the simulation.
   */
  forces?: SingleForce[];
}