import type { Parameter } from "../types/params";

export const createAnimationControls = (controlDiv: HTMLDivElement, paramSpecs: Parameter[], setParam: (paramName: string, newVal: string | number) => void) => {

  for (const paramSpec of paramSpecs) {

    if (!("animate" in paramSpec) || !paramSpec.animate) {
      continue;
    }
    const animationSpec = paramSpec.animate;

    // handle alternative spec options
    let numSteps, stepSize;
    if (animationSpec.values) {
      numSteps = animationSpec.values.length;
      animationSpec.min = animationSpec.min ?? 0;
      animationSpec.max = animationSpec.max ?? (numSteps - 1);
      stepSize = 1;
    } else if (animationSpec.stepSize) {
      stepSize = animationSpec.stepSize;
      numSteps = Math.ceil((animationSpec.max - animationSpec.min) / stepSize);
    } else {
      numSteps = animationSpec.numSteps ?? 10;
      stepSize = (animationSpec.max - animationSpec.min) / numSteps;
    }

    let stepTime;
    if (animationSpec.totalTime) {
      stepTime = animationSpec.totalTime / animationSpec.numSteps;
    } else {
      stepTime = animationSpec.stepTime ?? 10;
    }

    // state management
    let val = paramSpec.value ?? animationSpec.min;
    if (animationSpec.values && animationSpec.values.includes(val)) {
      val = animationSpec.values.indexOf(val);
    }

    let isAnimating = animationSpec.autoStart ?? false;

    const nextVal = (oldVal) => {
      let newVal = oldVal + stepSize;

      if (newVal > animationSpec.max) {
        if (animationSpec?.loop) {
          newVal = animationSpec.min;
        } else {
          clearInterval(interval);
          newVal = animationSpec.max;
        }
      }
      return newVal;
    };

    const prevVal = (oldVal) => {
      let newVal = oldVal - stepSize;

      if (newVal < animationSpec.min) {
        if (animationSpec?.loop) {
          newVal = animationSpec.max - (animationSpec.min - newVal);
        } else {
          clearInterval(interval);
          newVal = animationSpec.min;
        }
      }
      return newVal;
    };

    const stopAnimating = () => {
      isAnimating = false;
      clearInterval(interval);
      playPauseButton.innerText = "▶";
    };

    const startAnimating = () => {
      isAnimating = true;
      interval = setInterval(function() {
        val = nextVal(val);
        updateVal(val);
      }, stepTime * 1000);
      playPauseButton.innerText = "⏸";
    };

    const updateVal = (newVal) => {
      // if animationSpec.values, the val is an array index, and we need to look up the correct value
      const valToUse = (animationSpec.values) ? animationSpec.values[newVal] : newVal;

      setParam(paramSpec.name, valToUse);
      if (valLabel) {
        valLabel.innerText = valToUse;
      }

      forwardsButton.disabled = (newVal + stepSize > animationSpec.max);
      jumpToEndButton.disabled = (newVal + stepSize > animationSpec.max);

      backButton.disabled = (newVal - stepSize < animationSpec.min);
      jumpToStartButton.disabled = (newVal - stepSize < animationSpec.min);
    };


    // draw UI elements

    // label
    const label = document.createElement("label");
    label.innerText = ("label" in animationSpec) ? animationSpec.label : paramSpec.name;

    label.style.display = "inline-flex";
    label.style.alignItems = "center";
    label.style.userSelect = "none";

    controlDiv.appendChild(label);


    // value
    let valLabel;
    if (animationSpec?.showValue) {
      valLabel = document.createElement("label");
      valLabel.innerText = (animationSpec.values) ? animationSpec.values[val] : val;

      valLabel.style.display = "inline-flex";
      valLabel.style.alignItems = "center";
      valLabel.style.userSelect = "none";
      valLabel.style.paddingLeft = "0.25em";

      controlDiv.appendChild(valLabel);
    }

    // jump to start button
    let jumpToStartButton;
    if (animationSpec?.showJumpControl) {
      jumpToStartButton = document.createElement("button");
      jumpToStartButton.innerText = "⏮";
      jumpToStartButton.style.height = "3em";
      controlDiv.appendChild(jumpToStartButton);

      jumpToStartButton.addEventListener("click", () => {
        val = animationSpec.min;
        updateVal(val);
      });
    }

    // back
    let backButton;
    if (animationSpec?.showStepControl) {
      backButton = document.createElement("button");
      backButton.innerText = "⏪︎";
      backButton.style.height = "3em";

      controlDiv.appendChild(backButton);

      backButton.addEventListener("click", () => {
        val = prevVal(val);
        updateVal(val);
      });
    }


    // play/pause
    let playPauseButton;
    if (animationSpec?.showPauseControl) {
      playPauseButton = document.createElement("button");
      playPauseButton.innerText = isAnimating ? "⏸" : "▶";
      playPauseButton.style.height = "3em";

      controlDiv.appendChild(playPauseButton);

      playPauseButton.addEventListener("click", () => {
        isAnimating ? stopAnimating() : startAnimating();
      });

    }

    // forwards
    let forwardsButton;
    if (animationSpec?.showStepControl) {
      forwardsButton = document.createElement("button");
      forwardsButton.innerText = "⏩︎";
      forwardsButton.style.height = "3em";

      controlDiv.appendChild(forwardsButton);

      forwardsButton.addEventListener("click", () => {
        val = nextVal(val);
        updateVal(val);
      });

    }

    // jumpToEnd
    let jumpToEndButton;
    if (animationSpec?.showJumpControl) {
      jumpToEndButton = document.createElement("button");
      jumpToEndButton.innerText = "⏭";
      jumpToEndButton.style.height = "3em";

      controlDiv.appendChild(jumpToEndButton);

      jumpToEndButton.addEventListener("click", () => {
        val = animationSpec.max;
        updateVal(val);
      });

    }

    // loop control
    /*
    if (animationSpec?.showRepeatToggle) {
      const loopButton = document.createElement("button");
      loopButton.innerText = "🔃";
      loopButton.style.height = "3em";

      controlDiv.appendChild(loopButton);
    }
   */

    // actually change the value
    let interval;
    if (isAnimating) {
      interval = setInterval(function() {
        val = nextVal(val);
        updateVal(val);
      }, stepTime * 1000);
    }
  }
};
