import { max, min } from "d3-array";
import pointInPolygon from "point-in-polygon";
import type { ModifierDefinitions } from "../types/interaction";
import { formatTooltip } from "./formatTooltip";

const createSelectionOverlay = (element: HTMLElement, width: number, height: number) => {

  element.style.position = "relative";

  const selectionSVG = document.createElementNS("http://www.w3.org/2000/svg", "svg");
  selectionSVG.style.position = "absolute";
  selectionSVG.style.left = "0";
  selectionSVG.style.width = `${width}px`;
  selectionSVG.style.height = `${height}px`;
  selectionSVG.style.zIndex = "1";

  element.appendChild(selectionSVG);

  const rectBrush = document.createElementNS("http://www.w3.org/2000/svg", "rect");
  rectBrush.style.fill = "none";
  rectBrush.style.stroke = "black";
  selectionSVG.appendChild(rectBrush);

  // main closed polygon
  const lasooPolygon = document.createElementNS("http://www.w3.org/2000/svg", "path");
  lasooPolygon.style.stroke = "grey";
  lasooPolygon.style.strokeDasharray = "2,2";
  lasooPolygon.style.fill = "lightgrey";
  lasooPolygon.style.strokeWidth = "2px";
  selectionSVG.appendChild(lasooPolygon);

  // non-filed overlay with solid outline
  const lasooPolygonOutline = document.createElementNS("http://www.w3.org/2000/svg", "path");
  lasooPolygonOutline.style.stroke = "grey";
  lasooPolygonOutline.style.fill = "none";
  lasooPolygonOutline.style.strokeWidth = "3px";
  selectionSVG.appendChild(lasooPolygonOutline);

  // circle for start of lasso
  const lassoCircle = document.createElementNS("http://www.w3.org/2000/svg", "circle");
  lassoCircle.style.fill = "none";
  lassoCircle.setAttribute("r", "6");
  selectionSVG.appendChild(lassoCircle);

  return { selectionSVG, lasooPolygon, lasooPolygonOutline, lassoCircle, rectBrush };

};

const isPressed = (ev, keyName) => {
  const attributeName = {
    "alt": "altKey",
    "ctrl": "ctrlKey",
    "meta": "metaKey",
    "shift": "shiftKey"
  };

  return ev[attributeName[keyName]];
};

const correctModifiersPresses = (ev, modifiers: ModifierDefinitions) => {
  let isInvalid = false;
  isInvalid = isInvalid || (modifiers.modifierKeyRequired || []).some(k => !isPressed(ev, k));
  isInvalid = isInvalid || (modifiers.modifierKeyAbsent || []).some(k => isPressed(ev, k));
  return !isInvalid;
};


const updateLasso = (lassoPoints, sceneJSON, ev, updateSelection2, zoom, origin, offset) => {
  const nodesToProcess: any[] = [sceneJSON];
  const leafNodesInside = [];

  const lassoPointsScaled = lassoPoints.map(p => ({
    x: (p.x / zoom) - origin[0]- offset.x,
    y: (p.y / zoom) - origin[1]- offset.y
  }));


  const box = {
    x1: min(lassoPointsScaled.map(p => p.x)),
    x2: max(lassoPointsScaled.map(p => p.x)),
    y1: min(lassoPointsScaled.map(p => p.y)),
    y2: max(lassoPointsScaled.map(p => p.y))
  };

  const polygon = lassoPointsScaled.map(p => [p.x, p.y]);

  while (nodesToProcess.length > 0) {
    const node = nodesToProcess.pop();


    if ("items" in node) {
      for (const child of node.items) {
        if ("bounds" in child) {
          // check if partiallyCovered
          // @ts-ignore
          if (!(child.bounds.x2 < box.x1 || child.bounds.x1 > box.x2 || child.bounds.y2 < box.y1 || child.bounds.y1 > box.y2)) {
            nodesToProcess.push(child);
          }
        } else if ("mark" in child) {
          // @ts-ignore
          if (pointInPolygon([child.x, child.y], polygon)) {
            leafNodesInside.push(child);
          }
        }
      }
    } else {
      // test for being inside polygon, rather than overlapping bounding box

      if (pointInPolygon([node.x, node.y], polygon)) {
        leafNodesInside.push(node);
      }
    }
  }

  updateSelection(leafNodesInside, ev, updateSelection2, "lasso");
};


const mousedown = (ev, target, interactionState, sceneJSON, updateSelection2, interactionModifiers, zoom, origin, offset) => {

  // if modifier keys match pan, assume we're panning
  if (!!interactionModifiers.pan && correctModifiersPresses(ev, interactionModifiers.pan)) {
    interactionState.panStartEv = ev;
    return;
  }


  if (target) {
    // Clicked on something, so prepare for a drag
    interactionState.dragStartEv = ev;

    // if currently active modifier keys match the drag operation for a selection, assume we're dragging the selection
    for (const [selectionName, modifiers] of interactionModifiers.dragSelection){
      if (correctModifiersPresses(ev, modifiers)){
        interactionState.draggedSelection = selectionName;
        interactionState.layoutName = modifiers.layout;
        return;
      }
    }

    // otherwise, check if we should be dragging whatever was clicked on
    // This might have false positives if currently active modifiers would enable dragging of som elements, but not this one
    // - such cases are filtered out by the check in mouseup()
    for (const modifiers of interactionModifiers.dragElement) {
      if (correctModifiersPresses(ev, modifiers)) {
        interactionState.draggedObj = target.mark;
      }
    }

    // handle labels for table
    if (target.role === "rowLabel" || target.role === "colLabel"){
      interactionState.draggedTableHeading = target;
    }

    return;
  }

  // If we've got here, the click was on the background rather than an object

  // check if modifier keys are consistent with dragging a selection
  for (const [selectionName, modifiers] of interactionModifiers.dragSelection) {
    if (correctModifiersPresses(ev, modifiers)) {
      interactionState.draggedSelection = selectionName;
      interactionState.layoutName = modifiers.layout;
      return;
    }
  }


  // check if currently active modifier keys match the lasso operation for a selection
    for (const modifiers of interactionModifiers.lasso){
      if (correctModifiersPresses(ev, modifiers)){
        interactionState.brush = { x1: null, x2: null, y1: null, y2: null };
        interactionState.lassoPoints.push({ x: ev.offsetX, y: ev.offsetY });

        const path = `M ${interactionState.lassoPoints[0].x} ${interactionState.lassoPoints[0].y}` + interactionState.lassoPoints.slice(1).map(p => `L ${p.x}, ${p.y}`).join(" ");
        interactionState.lasooPolygon.setAttribute("d", path + "Z");
        interactionState.lasooPolygonOutline.setAttribute("d", path);

        interactionState.lassoCircle.style.fill = "gray";
        interactionState.lassoCircle.setAttribute("cx", interactionState.lassoPoints[0].x);
        interactionState.lassoCircle.setAttribute("cy", interactionState.lassoPoints[0].y);

        console.log("Lasso:", path)

        if (interactionState.lassoPoints.length >= 3) {
          updateLasso(interactionState.lassoPoints, sceneJSON, ev, updateSelection2, zoom, origin, offset);
        }

        return;
      }
    }


  for (const modifiers of interactionModifiers.rect) {
    if (correctModifiersPresses(ev, modifiers)) {

      interactionState.lassoPoints = [];
      interactionState.lasooPolygon.setAttribute("d", "");
      interactionState.lasooPolygonOutline.setAttribute("d", "");

        interactionState.lassoCircle.style.fill = "none";
      interactionState.lassoCircle.setAttribute("cx", "");
      interactionState.lassoCircle.setAttribute("cy", "");

      interactionState.brush = { x1: ev.offsetX, x2: null, y1: ev.offsetY, y2: null };
      interactionState.rectBrush.setAttribute("x", ev.offsetX);
      interactionState.rectBrush.setAttribute("y", ev.offsetY);
      interactionState.rectBrush.setAttribute("width", "0");

      return;
    }
  }
};

const mousemove = (ev, interactionState, interactionModifiers) => {
  const { brush, rectBrush } = interactionState;
  if (brush.x1 && !brush.x2) {
    rectBrush.setAttribute("x", Math.min(ev.offsetX, brush.x1).toString());
    rectBrush.setAttribute("width", Math.abs(ev.offsetX - brush.x1).toString());

    rectBrush.setAttribute("y", Math.min(ev.offsetY, brush.y1).toString());
    rectBrush.setAttribute("height", Math.abs(ev.offsetY - brush.y1).toString());
  }
};



const updateSelection = (elements, ev, updateSelection2, interactionType) => {

  const newSelections = new Map();
  const actions = new Map();


  for (const element of elements) {
    for (const action of (element.actions || []) ) {
      let isInvalid = false;
      isInvalid = isInvalid || (action.modifierKeyAbsent || []).some(k => isPressed(ev, k));
      isInvalid = isInvalid || (action.modifierKeyRequired || []).some(k => !isPressed(ev, k));

      if (isInvalid) {
        continue;
      }

      if (action.interaction !== "select"){
        continue
      }

      if ( (action.using ?? 'rect') !== interactionType){
        continue;
      }

      // I don't think this should ever happen
      //if (!element.mark.dataset.startsWith(action.dataset)){
     //   continue;
     // }

      if (!newSelections.has(action.selection)){
        newSelections.set(action.selection, {nodes: [], links: []});
      }

      if (element.mark.dataset.endsWith(".nodes")){
        newSelections.get(action.selection).nodes.push(element.mark.id);
      } else if (element.mark.dataset.endsWith(".links")){
        newSelections.get(action.selection).links.push(element.mark.id);
      }

      actions.set(action.selection, action);
    }

  }
    updateSelection2(newSelections, actions)
};

const mouseup = (ev, interactionState, sceneJSON, updateLayout, updateSelection2, updatePan, updateOrder, interactionModifiers, state, zoom, origin, offset) => {

  const { brush } = interactionState;

     // first check if we're panning
    if (interactionState.panStartEv){
      updatePan(interactionState.panStartEv, ev);
      interactionState.panStartEv = null;
      return;
    }

  if (brush.x1) {
    // The mouseup ends drawing a rectangular brush
    brush.x2 = ev.offsetX;
    brush.y2 = ev.offsetY;

    console.log("New brush", brush); // TOOD: swap x1/x2 and y1/y2 if needed


    // find things in rect
    const nodesToProcess: any[] = [sceneJSON];
    const leafNodesInside = [];

    const box = {
      x1: Math.min(brush.x1, brush.x2),
      x2: Math.max(brush.x1, brush.x2),
      y1: Math.min(brush.y1, brush.y2),
      y2: Math.max(brush.y1, brush.y2)
    };

    const scaledBox = {
      x1: (box.x1 / zoom) - origin[0] - offset.x,
      x2: (box.x2 / zoom) - origin[0] - offset.x,
      y1: (box.y1 / zoom) - origin[1] - offset.y,
      y2: (box.y2 / zoom) - origin[1] - offset.y
    };

    while (nodesToProcess.length > 0) {
      const node = nodesToProcess.pop();


      if ("items" in node) {
        for (const child of node.items) {
          if ("bounds" in child) {
            // check if partiallyCovered
            // @ts-ignore
            if (!(child.bounds.x2 <= scaledBox.x1 || child.bounds.x1 >= scaledBox.x2 || child.bounds.y2 <= scaledBox.y1 || child.bounds.y1 >= scaledBox.y2)) {
              nodesToProcess.push(child);
            }
          } else if ("mark" in child) {
            // @ts-ignore
            if (!(child.mark.bounds.x2 <= scaledBox.x1 || child.mark.bounds.x1 >= scaledBox.x2 || child.mark.bounds.y2 <= scaledBox.y1 || child.mark.bounds.y1 >= scaledBox.y2)) {
              leafNodesInside.push(child);
            }
          }
        }
      } else {
        if (!(node.bounds.x2 <= scaledBox.x1 || node.bounds.x1 >= scaledBox.x2 || node.bounds.y2 <= scaledBox.y1 || node.bounds.y1 >= scaledBox.y2)) {
          leafNodesInside.push(node);
        }
      }
    }

    console.log("selected things:", { leafNodesInside });
    updateSelection(leafNodesInside, ev, updateSelection2, "rect");

    brush.x1 = undefined; // close the brush


    // clear the brush
    interactionState.rectBrush.setAttribute("width", "0");
    interactionState.rectBrush.setAttribute("height", "0");


  } else {

    const heading = interactionState.draggedTableHeading;
    if (heading) {
      interactionState.dragEndEv = ev;

      const shift = {
        dx: interactionState.dragEndEv.layerX - interactionState.dragStartEv.layerX,
        dy: interactionState.dragEndEv.layerY - interactionState.dragStartEv.layerY
      };

      const layout = state[heading.layoutName];
      const bounds = layout.get(heading.idForLayout);

      const tableName = heading.layoutName.split("#")[0]; // this is a bit of a hack

      let change, orderName;
      if (heading.role === "rowLabel"){
        change = Math.floor(shift.dy / bounds.height);
        orderName = state[tableName].yOrderName;
      } else {
        change = Math.floor(shift.dx / bounds.width);
        orderName = state[tableName].xOrderName;
      }

      const order = state[orderName];
      const currentPos = order.get(heading.idForLayout);

      const newPos = Math.max(0, Math.min(currentPos + change, order.size -1) );

      for (const id of order.keys()){
        const itemPos = order.get(id);

        if (change > 0){
          // we have moved selected item to the *right*,
          // so some items move one place to the left to make room
          if (itemPos > currentPos && itemPos <= newPos){
            order.set(id, itemPos - 1)
          }
        } else if (change < 0){
          // we have moved selected item to the *left*.
          // so some items move to the right to make room
          if (itemPos < currentPos && itemPos >= newPos){
            order.set(id, itemPos + 1)
          }
        }
      }

      order.set(heading.idForLayout, newPos);

      updateOrder(orderName);

      interactionState.draggedTableHeading = null;
      return;
    }

    if (interactionState.draggedObj) {
      interactionState.dragEndEv = ev;

      const target = interactionState.draggedObj;
      for (const action of (target.actions || [])) {

        if (action.interaction !== "drag") {
          continue;
        }

        if (!correctModifiersPresses(interactionState.dragStartEv, action)) {
          continue;
        }

        const shift = {
          dx: interactionState.dragEndEv.layerX - interactionState.dragStartEv.layerX,
          dy: interactionState.dragEndEv.layerY - interactionState.dragStartEv.layerY
        };
        updateLayout(target.layout_name, [target.id], shift);
      }
    }

    if (interactionState.draggedSelection) {
      interactionState.dragEndEv = ev;

      const shift = {
        dx: interactionState.dragEndEv.layerX - interactionState.dragStartEv.layerX,
        dy: interactionState.dragEndEv.layerY - interactionState.dragStartEv.layerY
      };
      const nodeIds = state[interactionState.draggedSelection].nodes.map(n => n.id);
      updateLayout(interactionState.layoutName, nodeIds, shift);
    }

    interactionState.draggedObj = null;
    interactionState.draggedSelection = null;
  }
};

const mouseover = (ev, target, tooltipDiv) => {
  if (!target) {
    return;
  }

  /*
  if (ev.target.childNodes.length === 0){
    const title = document.createElement("title");
    title.innerText = target.mark.items[0].tooltip;
    ev.target.appendChild(title);
  }
  */

//  element.title = target.mark.items[0].tooltip;
  console.log(target.mark.items[0].tooltip);

  tooltipDiv.style.opacity = target.mark.items[0].tooltip ? 1 : 0;

  tooltipDiv.style.left = `${ev.offsetX}px`;
  tooltipDiv.style.top = `${ev.offsetY}px`;
  tooltipDiv.innerHTML = formatTooltip(target.mark.items[0].tooltip);
};


const mouseout = (ev, target, tooltipDiv) => {
  tooltipDiv.style.opacity = 0;
}

// Handle clicking on a node
const cursorEventHandler = (ev, target, updateSelection2, interactionModifiers) => {

  // Clicked on background
  if (!target) {
    return;
  }

  const newSelections = new Map();
  const selectionActions = new Map();

  let actionTaken = false;

  for (const action of (target.actions || [])) {

    if (action.using !== "cursor" || action.event !== ev.type) {
      continue;
    }
    actionTaken = true;

    if (!correctModifiersPresses(ev, action)) {
      continue;
    }

    let newSelection;
    if (target.mark.dataset.endsWith(".links")) {
      newSelection = { nodes: [], links: [target.mark.id] };
    } else if (target.mark.dataset.endsWith(".nodes")) {
      newSelection = { nodes: [target.mark.id], links: [] };
    }

    newSelections.set(action.selection, newSelection);
    selectionActions.set(action.selection, action);

   // editSelection(action.selection, newSelection, action.action);
  }

  if (actionTaken){
      updateSelection2(newSelections, selectionActions);
      console.log("CURSOR INTERACTION TRIGGERED:", { ev, target });
  }
};
const backgroundClickHandler = (ev, target, parameterSpecs, updateSelection2) => {

  if (target){
    // clicked on something that wasn't the background
    return;
  }

  const newSelections = new Map();
  const selectionActions = new Map();

  let actionTaken = false;

  for (const paramSpec of (parameterSpecs || [])) {
    if (paramSpec.clearOnBackgroundClick) {

      const newSelection = { nodes: [], links: [] };
      selectionActions.set(paramSpec.name, "replace");
      newSelections.set(paramSpec.name, newSelection);
      actionTaken = true;
    }
  }

  if (actionTaken){
      updateSelection2(newSelections, selectionActions);
      console.log("Clearing selection(s) due to background click");
  }
};


export { createSelectionOverlay, cursorEventHandler, backgroundClickHandler, mousedown, mousemove, mouseup, mouseover, mouseout, updateLasso };