// We follow the Vega symbol convention of defining paths:
// "within a square bounding box with coordinates ranging from -1 to 1 along both the x and y dimensions"

// N.B. When editing path strings, this tool is helpful: https://yqnn.github.io/svg-path-editor/

// all pointing *rightwards"

import type { MarkerSpec, MarkerShape } from "../types/LinkMarks";
import { convertVal, setFieldsOnItem } from "../complete";


const shapes: Record<MarkerShape, string> = {
  triangle: "M -1 -1 L 1 0 L -1 1Z",
  "circle-triangle": "M 0.5 0 m -0.5 0 a 0.5 0.5 0 1 0 1 0 a 0.5 0.5 0 1 0 -1 0 M -1 1 L 0 0 L -1 -1Z",
  circle: "M 0 0 m -1 0 a 1 1 0 1 0 2 0 a 1 1 0 1 0 -2 0Z",

  tee: "M 0.75 1 L 0.75 -1 L 1 -1 L 1 1Z",

  vee: "M -1 1 L 1 0 L -1 -1 L -0.5 0Z",
  square: "M -1 -1 L -1 1 L 1 1 L 1 -1Z",

  diamond: "M -1 0 L 0 1 L 1 0 L 0 -1Z",

  "triangle-tee": "M0-1 .75 0 0 1Z M.75 1 .75-1 1-1 1 1Z",
  "triangle-cross": "M 0 1 L 0 -1 L 0.125 -1 L 0.125 1Z M 0.25 -1 L 1 0 L 0.25 1Z"
};

const computeMarkerAngle = (start, end) => {
  // computes clockwise rotation of marker
  const angleRadians = Math.atan2(end.x - start.x, end.y - start.y);
  return -angleRadians * 180 / Math.PI;
};

const getMarker = (markerSpec: MarkerSpec, startPos, endPos, markerPos, stepDirection, entry, state) => {
  const stepBack = (markerSpec.stepBack ?? 5) * stepDirection;
  const length = Math.sqrt(Math.pow(endPos.x - startPos.x, 2) + Math.pow(endPos.y - startPos.y, 2));

  const markerShape = convertVal(markerSpec.shape, entry, state, "none");

  const newItem = {
    fill: "black",
    stroke: "black",

    ...markerSpec, // will pass through properties like stroke, opacity, and fill
    //angle,
    // shape: path,
    angle: null,
    x: null,
    y: null
  };

  setFieldsOnItem(newItem, { mark: markerSpec }, entry, state);

  // N.B. we need to have done evaluation to determine shape before we are able to compute marker angle.


  let path;
  let angleOffset = 0;
  if (markerShape in shapes) {
    path = shapes[markerShape];
    angleOffset = +90;
  } else {
    path = markerShape;
  }

  newItem.shape = path;

  let angle;

  const x = markerPos.x + stepBack * (endPos.x - startPos.x) / length;
  const y = markerPos.y + stepBack * (endPos.y - startPos.y) / length;


  /*
  if (end === "start") {
    x = startPos.x + stepBack * (endPos.x - startPos.x) / length;
    y = startPos.y + stepBack * (endPos.y - startPos.y) / length;
  } else {
    x = endPos.x - stepBack * (endPos.x - startPos.x) / length;
    y = endPos.y - stepBack * (endPos.y - startPos.y) / length;
  }
   */

  // by default, arrow points from source towards target, regardless of which end of line it is drawn at
  angle = computeMarkerAngle(startPos, endPos) + angleOffset
  if (markerSpec.direction === "reverse"){
    angle = angle - 180;
  }
  newItem.angle = angle;

  newItem.x = x;
  newItem.y = y;


  return newItem;
};

export { shapes, getMarker };
