import { select } from "../data/select";
import { getMarker } from "./arrowheads";

import { convertVal, ensureInducedLayoutCreated, setFieldsOnItem } from "../complete";
import { constructScale } from "../scales/scale";
import { ScaleType } from "vega-typings/types/spec/scale";

import type {
  LinkPathSpec, LinkPathSpecArc,
  LinkPathSpecCurve,
  LinkPathSpecCurveSymmetric,
  LinkPathSpecGlyphs, LinkPathSpecLine, LinkPathSpecOrthogonal, LinkPathSpecSquiggle,
  LinkPathSpecWedge
} from "../types/LinkMarks";
import type { Layout } from "../types/layout";
import type { LinkPathSpecDiagonal } from "../types/LinkMarks";

function generateLine(spec: LinkPathSpec, mark: LinkPathSpecLine, entry, state, startPos, endPos) {
  const pointOne = { x: startPos.x, y: startPos.y, stroke: "lightgrey" };
  setFieldsOnItem(pointOne, spec, entry, state);

  const pointTwo = { x: endPos.x, y: endPos.y, stroke: "lightgrey" };
  setFieldsOnItem(pointTwo, spec, entry, state);

  const strokeDash = convertVal(mark.strokeDash, entry, state, [1, 0]);
  const strokeWidth = convertVal(mark.strokeWidth, entry, state, 2);
  const strokeOpacity = convertVal(mark.strokeOpacity, entry, state, 1);

  const tooltip = convertVal(mark.tooltip, entry, state, "");

  return {
    marktype: "line",
    dataset: spec.entries, // so we can update correct selection
    id: entry.id, // needed so we know which element was clicked on
    stroke: convertVal(spec.mark.stroke, entry, state, "grey"),
    items: [{ ...pointOne, strokeDash, strokeWidth, strokeOpacity, tooltip }, {
      ...pointTwo,
      strokeDash,
      strokeWidth,
      strokeOpacity,
      tooltip
    }]
  };
}

function generateLineWithColorGradient(spec: LinkPathSpec, mark: LinkPathSpecLine, entry, state, startPos, endPos) {

  if (!("colorGradient" in mark)) {
    // this function shouldn't have been called - call the normal line generator function instead
    return generateLine(spec, mark, entry, state, startPos, endPos);
  }

  if (typeof (mark.colorGradient) === "string") {
    mark.colorGradient = {
      colorScheme: mark.colorGradient
    };
  }

  const numSegments = mark.colorGradient.numSegments ?? 10;

  const domain = (mark.colorGradient.direction === "reverse") ? [1, 0] : [0, 1];

  const colorScale = constructScale({
    name: "color",
    type: "linear",
    domain: domain,
    scheme: convertVal(mark.colorGradient.colorScheme, entry, state)
  } as unknown as ScaleType);


  const segmentStep = {
    dx: (endPos.x - startPos.x) / numSegments,
    dy: (endPos.y - startPos.y) / numSegments
  };
  let segmentStart = startPos;
  let segmentEnd;

  const lineSegments = [];
  for (let i = 0; i < numSegments; i++) {
    segmentEnd = {
      x: segmentStart.x + segmentStep.dx,
      y: segmentStart.y + segmentStep.dy
    };

    lineSegments.push({
      marktype: "path",
      id: entry.id, // needed so we know which element was clicked on
      dataset: spec.entries, // so we can update correct selection
      path: `M ${segmentStart.x}, ${segmentStart.y} L ${segmentEnd.x} ,${segmentEnd.y}`,
      strokeOpacity: convertVal(mark.strokeOpacity, entry, state, 1),
      stroke: colorScale(i / numSegments),
      tooltip: convertVal(mark.tooltip, entry, state, "")
    });

    segmentStart = segmentEnd;
  }

  return {
    marktype: "path",
    id: entry.id, // needed so we know which element was clicked on
    dataset: spec.entries, // so we can update correct selection
    items: lineSegments
  };
}

function generateWedge(spec: LinkPathSpec, mark: LinkPathSpecWedge, entry, state, startPos, endPos) {
  const length = Math.sqrt(Math.pow(endPos.x - startPos.x, 2) + Math.pow(endPos.y - startPos.y, 2));

  const unitNormal = [
    -(endPos.y - startPos.y) / length,
    (endPos.x - startPos.x) / length
  ];

  const width = convertVal(mark.width, entry, state, 5);


  let a = startPos;
  let b = endPos;
  if (mark.direction === "reverse") {
    a = endPos;
    b = startPos;
  }

  const path = `M ${a.x + width * unitNormal[0]} ${a.y + width * unitNormal[1]}
        L ${b.x} ${b.y}
    L ${a.x - width * unitNormal[0]} ${a.y - width * unitNormal[1]} 
    Z
    `;


  return {
    marktype: "path",
    id: entry.id, // needed so we know which element was clicked on
    dataset: spec.entries, // so we can update correct selection
    items: [{
      fill: convertVal(mark.fill, entry, state, "grey"),
      opacity: convertVal(mark.opacity, entry, state, 1),
      stroke: convertVal(mark.stroke, entry, state, "grey"),
      strokeOpacity: convertVal(mark.strokeOpacity, entry, state, 1),
      tooltip: convertVal(mark.tooltip, entry, state, ""),
      path
    }]
  };
}

function generateAsymmetricCurve(spec: LinkPathSpec, mark: LinkPathSpecCurve, entry, state, startPos, endPos) {
  // see Overlaying Graph Links on Treemaps
  const length = Math.sqrt(Math.pow(endPos.x - startPos.x, 2) + Math.pow(endPos.y - startPos.y, 2));

  const unitNormal = [
    -(endPos.y - startPos.y) / length,
    (endPos.x - startPos.x) / length
  ];


  let h;
  let b;

  if ("angle" in mark) {
    const radians = mark.angle * Math.PI / 180;
    h = (length / 2) * Math.sin(radians);
    b = (length / 2) * Math.cos(radians);
  } else {
    // default is 60 degrees
    h = (length / 2) * Math.sqrt(3) / 2;
    b = (length / 2) / 2;
  }


  const A = [
    startPos.x + (endPos.x - startPos.x) * b / length,
    startPos.y + (endPos.y - startPos.y) * b / length
  ];

  const B = [
    A[0] + h * unitNormal[0],
    A[1] + h * unitNormal[1]
  ];


  const path = `M ${startPos.x} ${startPos.y} C ${startPos.x} ${startPos.y} ${B[0]} ${B[1]} ${endPos.x} ${endPos.y}`;

  return {
    marktype: "path",
    id: entry.id, // needed so we know which element was clicked on
    dataset: spec.entries, // so we can update correct selection
    items: [{
      fill: "none",
      stroke: convertVal(mark.stroke, entry, state, "grey"),
      strokeDash: convertVal(mark.strokeDash, entry, state, [1, 0]),
      strokeWidth: convertVal(mark.strokeWidth, entry, state, 2),
      strokeOpacity: convertVal(mark.strokeOpacity, entry, state, 1),
      tooltip: convertVal(mark.tooltip, entry, state, ""),
      path
    }]
  };
}

function generateSymmetricCurve(spec: LinkPathSpec, mark: LinkPathSpecCurveSymmetric, entry, state, startPos, endPos) {
  // symmetric curve, as implemented by Vega's `curve in linkTransform
  // This is based on curve() in vega-core.js, but with added positiveAngle/sweep option

  const sx = startPos.x;
  const sy = startPos.y;
  const tx = endPos.x;
  const ty = endPos.y;

  const dx = tx - sx;
  const dy = ty - sy;
  const ix = 0.2 * (dx + dy);
  const iy = 0.2 * (dy - dx);

  const path = `M${sx},${sy} C${sx + ix},${sy + iy} ${tx + iy},${ty - ix} ${tx},${ty}`;

  return {
    marktype: "path",
    id: entry.id, // needed so we know which element was clicked on
    dataset: spec.entries, // so we can update correct selection
    items: [{
      fill: "none",
      stroke: convertVal(mark.stroke, entry, state, "grey"),
      strokeDash: convertVal(mark.strokeDash, entry, state, [1, 0]),
      strokeWidth: convertVal(mark.strokeWidth, entry, state, 2),
      strokeOpacity: convertVal(mark.strokeOpacity, entry, state, 1),
      tooltip: convertVal(mark.tooltip, entry, state, ""),
      path
    }]
  };
}

function generateSigmoid(spec: LinkPathSpec, mark: LinkPathSpecDiagonal, entry, state, startPos, endPos) {
  let path;
  if (mark.orientation === "horizontal") {
    const m = (startPos.x + endPos.x) / 2;
    path = `M ${startPos.x},${startPos.y} C${m},${startPos.y} ${m},${endPos.y} ${endPos.x},${endPos.y}`;
  } else {
    const m = (startPos.y + endPos.y) / 2;
    path = `M ${startPos.x},${startPos.y} C${startPos.x},${m} ${endPos.x},${m} ${endPos.x},${endPos.y}`;
  }

  return {
    marktype: "path",
    id: entry.id, // needed so we know which element was clicked on
    dataset: spec.entries, // so we can update correct selection
    items: [{
      fill: "none",
      stroke: convertVal(mark.stroke, entry, state, "grey"),
      strokeDash: convertVal(mark.strokeDash, entry, state, [1, 0]),
      strokeWidth: convertVal(mark.strokeWidth, entry, state, 2),
      strokeOpacity: convertVal(mark.strokeOpacity, entry, state, 1),
      tooltip: convertVal(mark.tooltip, entry, state, ""),
      path
    }]
  };
}

function generateOrthogonal(spec: LinkPathSpec, mark: LinkPathSpecOrthogonal, entry, state, startPos, endPos) {
  let path;
  if (mark.orientation === "vertical") {
    path = `M${startPos.x},${startPos.y} V${endPos.y} H${endPos.x}`;
  } else {
    path = `M${startPos.x},${startPos.y} H${endPos.x} V${endPos.y}`;
  }

  return {
    marktype: "path",
    id: entry.id, // needed so we know which element was clicked on
    dataset: spec.entries, // so we can update correct selection
    items: [{
      fill: "none",
      stroke: convertVal(mark.stroke, entry, state, "grey"),
      strokeDash: convertVal(mark.strokeDash, entry, state, [1, 0]),
      strokeWidth: convertVal(mark.strokeWidth, entry, state, 2),
      strokeOpacity: convertVal(mark.strokeOpacity, entry, state, 1),
      tooltip: convertVal(mark.tooltip, entry, state, ""),
      path
    }]
  };
}

function generateGlyphs(spec: LinkPathSpec, mark: LinkPathSpecGlyphs, entry, state, startPos, endPos) {
  let fill;
  let stroke;
  if (!mark.colorGradient) {

    fill = () => convertVal((mark as LinkPathSpecGlyphs).fill, entry, state, "black");
    stroke = () => convertVal(mark.stroke, entry, state, "black");

  } else {

    if (typeof mark.colorGradient === "string") {
      mark.colorGradient = {
        colorScheme: mark.colorGradient
      };
    }
    const domain = (mark.colorGradient.direction === "reverse") ? [1, 0] : [0, 1];

    const colorScale = constructScale({
      name: "color",
      type: "linear",
      domain: domain,
      scheme: convertVal(mark.colorGradient.colorScheme, entry, state)
    } as unknown as ScaleType);

    fill = colorScale;
    stroke = colorScale;
  }


  const separation = mark.separation ?? 5;
  const length = Math.sqrt(Math.pow(endPos.x - startPos.x, 2) + Math.pow(endPos.y - startPos.y, 2));

  const markerSpec = {
    shape: convertVal(mark.glyphShape, entry, state),
    stepBack: 0,
    opacity: convertVal(mark.opacity, entry, state, 1),
    strokeOpacity: convertVal(mark.strokeOpacity, entry, state, 1),
    strokeWidth: convertVal(mark.strokeWidth, entry, state, 1)
  };

  const unitStep = {
    dx: (endPos.x - startPos.x) / length,
    dy: (endPos.y - startPos.y) / length
  };

  const markers = [];
  for (let i = 0; i * separation < length; i++) {

    const markerPos = {
      x: startPos.x + i * separation * unitStep.dx,
      y: startPos.y + i * separation * unitStep.dy
    };

    const coloredMarkerSpec = {
      ...markerSpec,
      fill: fill((i * separation) / length),
      stroke: stroke((i * separation) / length)
    };
    const tooltip = convertVal(mark.tooltip, entry, state, "");

    markers.push({ ...getMarker(coloredMarkerSpec, startPos, endPos, markerPos, 0, entry, state), tooltip });
  }

  return {
    marktype: "symbol",
    id: entry.id, // needed so we know which element was clicked on
    dataset: spec.entries, // so we can update correct selection
    items: markers
  };
}

function generateSquiggles(spec: LinkPathSpec, mark: LinkPathSpecSquiggle, entry, state, startPos, endPos) {
  // AbBySS-explorer style squiggles
  // This implementation is based on https://github.com/bcgsc/ABySS-explorer/blob/master/ca/bcgsc/abyssexplorer/visualization/decorators/SquiggleGenerator.java#L25
  // however, we do some trigonometry here, rather than applying an affine transformation to rotate the squiggle after drawing it

  let path = `M ${startPos.x} ${startPos.y} `;

  let n;
  if ("spacing" in mark) {
    const length = Math.sqrt(Math.pow(endPos.y - startPos.y, 2) + Math.pow(endPos.x - startPos.x, 2));
    n = length / mark.spacing;
  } else {
    n = mark.n ?? 10;
  }

  const amp = mark.amplitude ?? 20;

  if (n == 0) {
    // no squiggles, so just return a straight line
    path += `L ${endPos.x}, ${endPos.y}`;
  } else {
    const d = (endPos.x - startPos.x);
    const curveWidth = d / n;

    const curveHeightStep = (endPos.y - startPos.y) / n;


    let start_x = startPos.x;
    let start_y = startPos.y;

    const lineAngleRadians = Math.atan2(endPos.y - startPos.y, endPos.x - startPos.x);

    for (let i = 0; i < n; i++) {

      // determine the amplitude for this curve
      const mx = (start_x + curveWidth / 2) - startPos.x; // how far along the line this point is
      const T = 2 * d; // twice the total line length
      const x = (mx / T) * Math.PI;

      let radians = 2 * x;
      if (mx / T <= 1 / 6) {
        radians += x;
      } else if (mx / T <= 1 / 2) {
        radians += -x / 2 + Math.PI / 4;
      }

      const normAmp = amp * Math.sin(radians);

      // construct a segment of Bezier curve
      let mid_x;
      let mid_y;
      if (i % 2 == 0) {
        mid_x = start_x + (curveWidth / 2) + normAmp * Math.sin(Math.PI - lineAngleRadians);
        mid_y = start_y + (curveHeightStep / 2) + normAmp * Math.cos(Math.PI - lineAngleRadians);
      } else {
        mid_x = start_x + (curveWidth / 2) - normAmp * Math.sin(Math.PI - lineAngleRadians);
        mid_y = start_y + (curveHeightStep / 2) - normAmp * Math.cos(Math.PI - lineAngleRadians);
      }

      const end_x = start_x + curveWidth;
      const end_y = start_y + curveHeightStep;

      path += ` Q ${mid_x}, ${mid_y}, ${end_x}, ${end_y}`;

      start_x += curveWidth;
      start_y += curveHeightStep;
    }
  }

  return {
    marktype: "path",
    id: entry.id, // needed so we know which element was clicked on
    dataset: spec.entries, // so we can update correct selection
    items: [{
      fill: "none",
      stroke: convertVal(mark.stroke, entry, state, "grey"),
      strokeDash: convertVal(mark.strokeDash, entry, state, [1, 0]),
      strokeWidth: convertVal(mark.strokeWidth, entry, state, 2),
      strokeOpacity: convertVal(mark.strokeOpacity, entry, state, 1),
      tooltip: convertVal(mark.tooltip, entry, state, ""),
      path
    }]
  };
}

function generateArc(spec: LinkPathSpec, mark: LinkPathSpecArc, entry, state, startPos, endPos) {
  // This is based on arc$2 in vega-core.js, but with added positiveAngle/sweep option
  const dx = endPos.x - startPos.x;
  const dy = endPos.y - startPos.y;
  const rr = Math.sqrt(dx * dx + dy * dy) / 2;
  const ra = 180 * Math.atan2(dy, dx) / Math.PI;


  // See https://svgwg.org/svg2-draft/paths.html#PathDataEllipticalArcCommands
  // " A rx ry x-axis-rotation large-arc-flag sweep-flag x y"
  const sweep = ("positiveAngle" in mark && !mark.positiveAngle) ? 0 : 1;
  const path = `M ${startPos.x},${startPos.y} A${rr},${rr} ${ra} 0 ${sweep} ${endPos.x},${endPos.y}`;

  return {
    marktype: "path",
    id: entry.id, // needed so we know which element was clicked on
    dataset: spec.entries, // so we can update correct selection
    items: [{
      fill: "none",
      stroke: convertVal(mark.stroke, entry, state, "grey"),
      strokeDash: convertVal(mark.strokeDash, entry, state, [1, 0]),
      strokeWidth: convertVal(mark.strokeWidth, entry, state, 2),
      strokeOpacity: convertVal(mark.strokeOpacity, entry, state, 1),
      tooltip: convertVal(mark.tooltip, entry, state, ""),
      path
    }]
  };
}


export const convertLinkPath = (spec: LinkPathSpec, entry, state, layout?: Layout, extra?, bounds?) => {
  const startObj = select(entry, spec.mark.start ?? "source");
  let startPos = layout.get(startObj.id);

  const endObj = select(entry, spec.mark.end ?? "target");
  let endPos = layout.get(endObj.id);

  // handle self-loops
  if (startObj === endObj) {
    const r = 2;
    const path = `
  M ${startPos.x} ${startPos.y}
          m ${-r}, 0
        a ${r},${r} 0 1,0 ${r * 2},0
        a ${r},${r} 0 1,0 ${-r * 2},0
  `;

    return [{
      marktype: "path",
      items: [{
        fill: "none",
        stroke: convertVal(spec.mark.stroke, entry, state, "grey"),
        tooltip: convertVal(spec.mark.tooltip, entry, state, ""),
        path
      }],
      id: entry.id, // needed so we know which element was clicked on
      dataset: spec.entries // so we can update correct selection
    }];
  }


  // N.B. We could use a `marker`, but this would only work for SVG rendering, not canvas rendering
  const endMarkers = [];
  if (spec.mark.startMarker) {
    endMarkers.push(getMarker(spec.mark.startMarker, startPos, endPos, startPos, +1, entry, state));
  }

  if (spec.mark.endMarker) {
    endMarkers.push(getMarker(spec.mark.endMarker, startPos, endPos, endPos, -1, entry, state));
  }

  const endMarker = {
    marktype: "symbol",
    items: endMarkers,
    id: entry.id, // needed so we know which element was clicked on
    dataset: spec.entries // so we can update correct selection
  };

  let sceneItem;

 // const shape = spec.mark.shape; // tODO: lookup

  // if necessary, swap source/target positions
  // N.B., this is done after start/end markers constructed, so doesn't affect them
  if (spec.mark.directionForShape) {
    const direction = spec.mark.directionForShape;

    const ordering = state[direction.ordering];
    const p1 = ordering.get(startObj.id);
    const p2 = ordering.get(endObj.id);

    const orderingMax = Math.max(...ordering.values());

    if (direction.reverse) {

      const wrapAround = (p1 === 0 && p2 === orderingMax && direction.cyclic);
      if (wrapAround || (p1 > p2)) {
        const tmp = startPos;
        startPos = endPos;
        endPos = tmp;
      }

    } else {
      const wrapAround = (p2 === 0 && p1 === orderingMax && direction.cyclic);
      if (wrapAround || (p2 > p1)) {
        const tmp = startPos;
        startPos = endPos;
        endPos = tmp;
      }
    }
  }
  //

  const shape = convertVal(spec.mark.shape, entry, state)

  if (shape === "wedge") {
    sceneItem = generateWedge(spec, spec.mark as LinkPathSpecWedge, entry, state, startPos, endPos);
  } else if (shape === "line" && (spec.mark as LinkPathSpecLine).colorGradient) {
    sceneItem = generateLineWithColorGradient(spec, spec.mark as LinkPathSpecLine, entry, state, startPos, endPos);
  } else if (shape === "curve") {
    sceneItem = generateAsymmetricCurve(spec, spec.mark as LinkPathSpecCurve, entry, state, startPos, endPos);
  } else if (shape === "curveSymmetric") {
    sceneItem = generateSymmetricCurve(spec, spec.mark as LinkPathSpecCurveSymmetric, entry, state, startPos, endPos);
  } else if (shape === "arc") {
    sceneItem = generateArc(spec, spec.mark as LinkPathSpecArc, entry, state, startPos, endPos);
  } else if (shape === "sigmoid") {
    sceneItem = generateSigmoid(spec, spec.mark as LinkPathSpecDiagonal, entry, state, startPos, endPos);
  } else if (shape === "orthogonal") {
    sceneItem = generateOrthogonal(spec, spec.mark as LinkPathSpecOrthogonal, entry, state, startPos, endPos);
  } else if (shape === "glyphs") {
    sceneItem = generateGlyphs(spec, spec.mark as LinkPathSpecGlyphs, entry, state, startPos, endPos);
  } else if (shape === "squiggle") {
    sceneItem = generateSquiggles(spec, spec.mark as LinkPathSpecSquiggle, entry, state, startPos, endPos);
  } else {
    sceneItem = generateLine(spec, spec.mark as LinkPathSpecLine, entry, state, startPos, endPos);
  }

  if (spec.inducedLayout) {

    const inducedLayoutName = spec.inducedLayout.name;
    ensureInducedLayoutCreated(extra, inducedLayoutName);

    let svgPath;
    if (!shape || shape === "line") {
      // for this shape, scene graph includes a line rather than a SVG Path
      svgPath = `M${sceneItem.items[0].x},${sceneItem.items[0].y}L${sceneItem.items[1].x},${sceneItem.items[1].y}`;
    } else {
      svgPath = sceneItem.items[0].path;
    }

    const pos = getCurvePosition(svgPath, spec.inducedLayout.position ?? 0.5);

    const xOffset = spec.inducedLayout.x ? convertVal(spec.inducedLayout.x, entry, state, 0, bounds) : 0;
    const yOffset = spec.inducedLayout.y ? convertVal(spec.inducedLayout.y, entry, state, 0, bounds) : 0;

    extra[inducedLayoutName].set(entry.id, {
      x: pos.x + xOffset, // + bounds.x
      y: pos.y + yOffset, // + bounds.y
      width: spec.inducedLayout.width ?? 100,
      height: spec.inducedLayout.height ?? 100
    });
  }

  return [sceneItem, endMarker];
};

const getCurvePosition = (svgPathString: string, fractionalLength: number) => {
  const xmlns = "http://www.w3.org/2000/svg";
  const svgEl = document.createElementNS(xmlns, "svg");

  const path = document.createElementNS(xmlns, "path");
  path.setAttribute("d", svgPathString);
  svgEl.appendChild(path);

  return path.getPointAtLength(path.getTotalLength() * fractionalLength);
};